"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const BlockDTO_1 = require("../dto/BlockDTO");
class Switcher {
    constructor(dao, invalidForks, avgGenTime, forkWindowSize, switchOnHeadAdvance, logger = undefined) {
        this.dao = dao;
        this.invalidForks = invalidForks;
        this.avgGenTime = avgGenTime;
        this.forkWindowSize = forkWindowSize;
        this.switchOnHeadAdvance = switchOnHeadAdvance;
        this.logger = logger;
    }
    /**
     * Looks at known blocks in the sandbox and try to follow the longest resulting chain that has at least both 3 blocks of
     * advance and 3 * avgGenTime of medianTime advancce.
     */
    tryToFork() {
        return __awaiter(this, void 0, void 0, function* () {
            const current = yield this.dao.getCurrent();
            if (current) {
                const numberStart = current.number + this.switchOnHeadAdvance;
                const timeStart = current.medianTime + this.switchOnHeadAdvance * this.avgGenTime;
                // Phase 1: find potential chains
                const suites = yield this.findPotentialSuites(numberStart, timeStart);
                if (suites.length) {
                    this.logger && this.logger.info("Fork resolution: %s potential suite(s) found...", suites.length);
                }
                // Phase 2: select the best chain
                let longestChain = yield this.findLongestChain(current, suites);
                // Phase 3: a best exist?
                if (longestChain) {
                    const chainHEAD = longestChain[longestChain.length - 1];
                    // apply it if it respects the 3-3 rule
                    if (chainHEAD.number >= numberStart && chainHEAD.medianTime >= timeStart) {
                        yield this.switchOnChain(longestChain);
                        return yield this.dao.getCurrent();
                    }
                }
            }
            return null;
        });
    }
    /**
     * Find all the suites' HEAD that we could potentially fork on, in the current fork window.
     * @param current
     */
    findPotentialSuitesHeads(current) {
        return __awaiter(this, void 0, void 0, function* () {
            const numberStart = current.number - this.forkWindowSize;
            const timeStart = current.medianTime - this.forkWindowSize * this.avgGenTime;
            const suites = yield this.findPotentialSuites(numberStart, timeStart);
            return suites.map(suite => suite[suite.length - 1]);
        });
    }
    /**
     * Looks at the potential blocks that could form fork chains in the sandbox, and sort them to have a maximum of unique
     * chains.
     * @param numberStart The minimum number of a fork block.
     * @param timeStart The minimum medianTime of a fork block.
     * @returns {SwitchBlock[][]} The suites found.
     */
    findPotentialSuites(numberStart, timeStart) {
        return __awaiter(this, void 0, void 0, function* () {
            const suites = [];
            const potentials = yield this.dao.getPotentials(numberStart, timeStart, numberStart + this.forkWindowSize);
            const knownForkBlocks = {};
            for (const candidate of potentials) {
                knownForkBlocks[BlockDTO_1.BlockDTO.fromJSONObject(candidate).blockstamp] = true;
            }
            const invalids = {};
            if (potentials.length) {
                this.logger && this.logger.info("Fork resolution: %s potential block(s) found...", potentials.length);
            }
            for (const candidate of potentials) {
                const suite = [];
                // Do not process the block if it is already known as invalid (has no fork point with current blockchain or misses
                // some blocks) or is already contained in a valid chain.
                if (!invalids[candidate.hash] && !Switcher.suitesContains(suites, candidate)) {
                    // Tries to build up a full chain that is linked to current chain by a fork point.
                    let previous = candidate, commonRootFound = false;
                    let previousNumber = previous.number - 1;
                    let previousHash = previous.previousHash;
                    while (previous && previous.number > candidate.number - this.forkWindowSize) {
                        suite.push(previous);
                        previousNumber = previous.number - 1;
                        previousHash = previous.previousHash;
                        previous = null;
                        const previousBlockstamp = [previousNumber, previousHash].join('-');
                        // We try to look at blockchain if, of course, it is not already known as a fork block
                        // Otherwise it cost a useless DB access
                        if (!knownForkBlocks[previousBlockstamp]) {
                            previous = yield this.dao.getBlockchainBlock(previousNumber, previousHash);
                        }
                        if (previous) {
                            // Stop the loop: common block has been found
                            previous = null;
                            suites.push(suite);
                            commonRootFound = true;
                        }
                        else {
                            // Have a look in sandboxes
                            previous = yield this.dao.getSandboxBlock(previousNumber, previousHash);
                            if (previous) {
                                knownForkBlocks[BlockDTO_1.BlockDTO.fromJSONObject(previous).blockstamp] = true;
                                const alreadyKnownInvalidBlock = this.invalidForks.indexOf([previous.number, previous.hash].join('-')) !== -1;
                                if (alreadyKnownInvalidBlock) {
                                    // Incorrect = not found
                                    this.logger && this.logger.info("Fork resolution: block #%s-%s is known as incorrect. Skipping.", previous.number, previous.hash.substr(0, 8));
                                    previous = null;
                                }
                            }
                        }
                    }
                    // Forget about invalid blocks
                    if (!commonRootFound) {
                        if (!previous) {
                            this.logger && this.logger.debug("Suite -> %s-%s missing block#%s-%s", candidate.number, candidate.hash.substr(0, 8), previousNumber, previousHash.substr(0, 8));
                            for (const b of suite) {
                                invalids[b.hash] = b;
                            }
                        }
                        else {
                            // The chain would be too long, we could not revert correctly the chain.
                            this.logger && this.logger.debug("Suite #%s-%s -> %s-%s out of fork window", previousNumber, previousHash.substr(0, 8), candidate.number, candidate.hash.substr(0, 8));
                        }
                    }
                }
            }
            return suites;
        });
    }
    /**
     * Find the longest chain among a suite of chains. Tests the validity of each block against the current blockchain.
     * The length of a chain is the number of blocks successfuly added to current blockchain.
     * @param {SwitchBlock} current
     * @param {SwitchBlock[][]} suites
     * @returns {SwitchBlock[]}
     */
    findLongestChain(current, suites) {
        return __awaiter(this, void 0, void 0, function* () {
            if (suites.length) {
                this.logger && this.logger.info("Fork resolution: HEAD = block#%s", current.number);
            }
            let longestChain = null;
            let j = 0;
            for (const s of suites) {
                j++;
                s.reverse();
                // Revert current blockchain to fork point
                const forkPoint = s[0].number - 1;
                const forkHead = s[s.length - 1];
                this.logger && this.logger.info("Fork resolution: suite %s/%s (-> #%s-%s) revert to fork point block#%s", j, suites.length, forkHead.number, forkHead.hash.substr(0, 6), forkPoint);
                const reverted = yield this.dao.revertTo(s[0].number - 1);
                // Try to add a maximum of blocks
                let added = true, i = 0, successfulBlocks = [];
                while (added && i < s.length) {
                    try {
                        yield this.dao.addBlock(s[i]);
                        this.logger && this.logger.info("Fork resolution: suite %s/%s added block#%s-%s", j, suites.length, s[i].number, s[i].hash);
                        successfulBlocks.push(s[i]);
                    }
                    catch (e) {
                        this.invalidForks.push([s[i].number, s[i].hash].join('-'));
                        this.logger && this.logger.info("Fork resolution: suite %s/%s REFUSED block#%s: %s", j, suites.length, s[0].number + i, e && e.message);
                        added = false;
                    }
                    i++;
                }
                // Pop the successfuly added blocks
                if (successfulBlocks.length) {
                    for (const b of successfulBlocks) {
                        this.invalidForks.push([b.number, b.hash].join('-'));
                    }
                    const addedToHeadLevel = successfulBlocks[successfulBlocks.length - 1].number - current.number;
                    this.logger && this.logger.info("Fork resolution: suite %s/%s reached HEAD + %s. Now rolling back.", j, suites.length, addedToHeadLevel);
                    yield this.dao.revertTo(forkPoint);
                }
                // Push back the initial blocks that were temporarily reverted
                reverted.reverse();
                for (const b of reverted) {
                    yield this.dao.addBlock(b);
                }
                // Remember the chain if it is the longest (highest HEAD) among tested chains
                const longestHEAD = longestChain && longestChain[longestChain.length - 1];
                const successHEAD = successfulBlocks && successfulBlocks[successfulBlocks.length - 1];
                if ((!longestHEAD && successHEAD) || (longestHEAD && successHEAD && longestHEAD.number < successHEAD.number)) {
                    longestChain = successfulBlocks;
                }
            }
            return longestChain;
        });
    }
    /**
     * Switch current blockchain on another chain, by poping top blocks and replacing them by new ones.
     * @param {SwitchBlock[]} chain
     */
    switchOnChain(chain) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.dao.revertTo(chain[0].number - 1);
            for (const b of chain) {
                yield this.dao.addBlock(b);
            }
        });
    }
    /**
     * Checks if a suite of chains contains a particular block in one of its chains.
     * @param {SwitchBlock[][]} suites
     * @param {SwitchBlock} block
     */
    static suitesContains(suites, block) {
        for (const suite of suites) {
            for (const b of suite) {
                if (b.number === block.number && b.hash === block.hash) {
                    return true;
                }
            }
        }
        return false;
    }
}
exports.Switcher = Switcher;
//# sourceMappingURL=Switcher.js.map