"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const BlockDTO_1 = require("../dto/BlockDTO");
const DuniterBlockchain_1 = require("../blockchain/DuniterBlockchain");
const _ = require('underscore');
const indexer = require('../indexer').Indexer;
const constants = require('../constants');
class BlockchainContext {
    constructor() {
        this.HEADrefreshed = Promise.resolve();
    }
    /**
     * Refresh the virtual HEAD value for determined variables of the next coming block, avoiding to recompute them
     * each time a new block arrives to check if the values are correct. We can know and store them early on, in vHEAD.
     */
    refreshHead() {
        this.HEADrefreshed = (() => __awaiter(this, void 0, void 0, function* () {
            this.vHEAD_1 = yield this.dal.head(1);
            // We suppose next block will have same version #, and no particular data in the block (empty index)
            let block;
            // But if no HEAD_1 exist, we must initialize a block with default values
            if (!this.vHEAD_1) {
                block = {
                    version: constants.BLOCK_GENERATED_VERSION,
                    time: Math.round(Date.now() / 1000),
                    powMin: this.conf.powMin || 0,
                    powZeros: 0,
                    powRemainder: 0,
                    avgBlockSize: 0
                };
            }
            else {
                block = { version: this.vHEAD_1.version };
            }
            this.vHEAD = yield indexer.completeGlobalScope(BlockDTO_1.BlockDTO.fromJSONObject(block), this.conf, [], this.dal);
        }))();
        return this.HEADrefreshed;
    }
    /**
     * Gets a copy of vHEAD, extended with some extra properties.
     * @param props The extra properties to add.
     */
    getvHeadCopy(props = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.vHEAD) {
                yield this.refreshHead();
            }
            const copy = {};
            const keys = Object.keys(this.vHEAD);
            for (const k of keys) {
                copy[k] = this.vHEAD[k];
            }
            _.extend(copy, props);
            return copy;
        });
    }
    /**
     * Get currently written HEAD.
     */
    getvHEAD_1() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.vHEAD) {
                yield this.refreshHead();
            }
            return this.vHEAD_1;
        });
    }
    /**
     * Utility method: gives the personalized difficulty level of a given issuer for next block.
     * @param issuer The issuer we want to get the difficulty level.
     */
    getIssuerPersonalizedDifficulty(issuer) {
        return __awaiter(this, void 0, void 0, function* () {
            const local_vHEAD = yield this.getvHeadCopy({ issuer });
            yield indexer.preparePersonalizedPoW(local_vHEAD, this.vHEAD_1, (n, m, p = "") => this.dal.range(n, m, p), this.conf);
            return local_vHEAD.issuerDiff;
        });
    }
    setConfDAL(newConf, newDAL, theBlockchain, theQuickSynchronizer) {
        this.dal = newDAL;
        this.conf = newConf;
        this.blockchain = theBlockchain;
        this.quickSynchronizer = theQuickSynchronizer;
        this.logger = require('../logger').NewLogger(this.dal.profile);
    }
    checkBlock(block, withPoWAndSignature) {
        return __awaiter(this, void 0, void 0, function* () {
            return DuniterBlockchain_1.DuniterBlockchain.checkBlock(block, withPoWAndSignature, this.conf, this.dal);
        });
    }
    addBlock(obj, index = null, HEAD = null) {
        return __awaiter(this, void 0, void 0, function* () {
            const block = yield this.blockchain.pushTheBlock(obj, index, HEAD, this.conf, this.dal, this.logger);
            this.vHEAD_1 = this.vHEAD = this.HEADrefreshed = null;
            return block;
        });
    }
    addSideBlock(obj) {
        return __awaiter(this, void 0, void 0, function* () {
            const dbb = yield this.blockchain.pushSideBlock(obj, this.dal, this.logger);
            return dbb.toBlockDTO();
        });
    }
    revertCurrentBlock() {
        return __awaiter(this, void 0, void 0, function* () {
            const head_1 = yield this.dal.bindexDAL.head(1);
            this.logger.debug('Reverting block #%s...', head_1.number);
            const res = yield this.blockchain.revertBlock(head_1.number, head_1.hash, this.dal);
            // Invalidates the head, since it has changed.
            yield this.refreshHead();
            return res;
        });
    }
    applyNextAvailableFork() {
        return __awaiter(this, void 0, void 0, function* () {
            const current = yield this.current();
            this.logger.debug('Find next potential block #%s...', current.number + 1);
            const forks = yield this.dal.getForkBlocksFollowing(current);
            if (!forks.length) {
                throw constants.ERRORS.NO_POTENTIAL_FORK_AS_NEXT;
            }
            const block = forks[0];
            yield this.checkAndAddBlock(block);
            this.logger.debug('Applied block #%s', block.number);
        });
    }
    checkAndAddBlock(block) {
        return __awaiter(this, void 0, void 0, function* () {
            const { index, HEAD } = yield this.checkBlock(block, constants.WITH_SIGNATURES_AND_POW);
            return yield this.addBlock(block, index, HEAD);
        });
    }
    current() {
        return this.dal.getCurrentBlockOrNull();
    }
    checkHaveEnoughLinks(target, newLinks) {
        return __awaiter(this, void 0, void 0, function* () {
            const links = yield this.dal.getValidLinksTo(target);
            let count = links.length;
            if (newLinks[target] && newLinks[target].length) {
                count += newLinks[target].length;
            }
            if (count < this.conf.sigQty) {
                throw 'Key ' + target + ' does not have enough links (' + count + '/' + this.conf.sigQty + ')';
            }
        });
    }
    quickApplyBlocks(blocks, to) {
        return this.quickSynchronizer.quickApplyBlocks(blocks, to);
    }
}
exports.BlockchainContext = BlockchainContext;
//# sourceMappingURL=BlockchainContext.js.map