"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const qfs = require('q-io/fs');
const sqlite3 = require("sqlite3").verbose();
const MEMORY_PATH = ':memory:';
class SQLiteDriver {
    constructor(path) {
        this.path = path;
        this.dbPromise = null;
        this.logger = require('../../logger').NewLogger('driver');
    }
    getDB() {
        if (!this.dbPromise) {
            this.dbPromise = (() => __awaiter(this, void 0, void 0, function* () {
                this.logger.debug('Opening SQLite database "%s"...', this.path);
                let sqlite = new sqlite3.Database(this.path);
                yield new Promise((resolve) => sqlite.once('open', resolve));
                // Database is opened
                // Force case sensitiveness on LIKE operator
                const sql = 'PRAGMA case_sensitive_like=ON';
                yield new Promise((resolve, reject) => sqlite.exec(sql, (err) => {
                    if (err)
                        return reject(Error('SQL error "' + err.message + '" on INIT queries "' + sql + '"'));
                    return resolve();
                }));
                // Database is ready
                return sqlite;
            }))();
        }
        return this.dbPromise;
    }
    executeAll(sql, params) {
        return __awaiter(this, void 0, void 0, function* () {
            const db = yield this.getDB();
            return new Promise((resolve, reject) => db.all(sql, params, (err, rows) => {
                if (err) {
                    return reject(Error('SQL error "' + err.message + '" on query "' + sql + '"'));
                }
                else {
                    return resolve(rows);
                }
            }));
        });
    }
    executeSql(sql) {
        return __awaiter(this, void 0, void 0, function* () {
            const db = yield this.getDB();
            return new Promise((resolve, reject) => db.exec(sql, (err) => {
                if (err) {
                    return reject(Error('SQL error "' + err.message + '" on query "' + sql + '"'));
                }
                else {
                    return resolve();
                }
            }));
        });
    }
    destroyDatabase() {
        return __awaiter(this, void 0, void 0, function* () {
            this.logger.debug('Removing SQLite database...');
            yield this.closeConnection();
            if (this.path !== MEMORY_PATH) {
                yield qfs.remove(this.path);
            }
            this.logger.debug('Database removed');
        });
    }
    closeConnection() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.dbPromise) {
                return;
            }
            const db = yield this.getDB();
            if (process.platform === 'win32') {
                db.open; // For an unknown reason, we need this line.
            }
            yield new Promise((resolve, reject) => {
                this.logger.debug('Trying to close SQLite...');
                db.on('close', () => {
                    this.logger.info('Database closed.');
                    this.dbPromise = null;
                    resolve();
                });
                db.on('error', (err) => {
                    if (err && err.message === 'SQLITE_MISUSE: Database is closed') {
                        this.dbPromise = null;
                        return resolve();
                    }
                    reject(err);
                });
                try {
                    db.close();
                }
                catch (e) {
                    this.logger.error(e);
                    throw e;
                }
            });
        });
    }
}
exports.SQLiteDriver = SQLiteDriver;
//# sourceMappingURL=SQLiteDriver.js.map