"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const ConfDAL_1 = require("./fileDALs/ConfDAL");
const StatDAL_1 = require("./fileDALs/StatDAL");
const ConfDTO_1 = require("../dto/ConfDTO");
const TransactionDTO_1 = require("../dto/TransactionDTO");
const DBBlock_1 = require("../db/DBBlock");
const MerkleDTO_1 = require("../dto/MerkleDTO");
const constants_1 = require("../common-libs/constants");
const fs = require('fs');
const path = require('path');
const readline = require('readline');
const _ = require('underscore');
const indexer = require('../indexer').Indexer;
const logger = require('../logger').NewLogger('filedal');
const constants = require('../constants');
class FileDAL {
    constructor(params) {
        this.setRevocating = (existing, revocation_sig) => {
            existing.revocation_sig = revocation_sig;
            existing.revoked = false;
            return this.idtyDAL.saveIdentity(existing);
        };
        this.rootPath = params.home;
        this.myFS = params.fs;
        this.sqliteDriver = params.dbf();
        this.wotb = params.wotb;
        this.profile = 'DAL';
        // DALs
        this.confDAL = new ConfDAL_1.ConfDAL(this.rootPath, this.myFS);
        this.metaDAL = new (require('./sqliteDAL/MetaDAL').MetaDAL)(this.sqliteDriver);
        this.peerDAL = new (require('./sqliteDAL/PeerDAL').PeerDAL)(this.sqliteDriver);
        this.blockDAL = new (require('./sqliteDAL/BlockDAL').BlockDAL)(this.sqliteDriver);
        this.txsDAL = new (require('./sqliteDAL/TxsDAL').TxsDAL)(this.sqliteDriver);
        this.statDAL = new StatDAL_1.StatDAL(this.rootPath, this.myFS);
        this.idtyDAL = new (require('./sqliteDAL/IdentityDAL').IdentityDAL)(this.sqliteDriver);
        this.certDAL = new (require('./sqliteDAL/CertDAL').CertDAL)(this.sqliteDriver);
        this.msDAL = new (require('./sqliteDAL/MembershipDAL').MembershipDAL)(this.sqliteDriver);
        this.walletDAL = new (require('./sqliteDAL/WalletDAL').WalletDAL)(this.sqliteDriver);
        this.bindexDAL = new (require('./sqliteDAL/index/BIndexDAL').BIndexDAL)(this.sqliteDriver);
        this.mindexDAL = new (require('./sqliteDAL/index/MIndexDAL').MIndexDAL)(this.sqliteDriver);
        this.iindexDAL = new (require('./sqliteDAL/index/IIndexDAL').IIndexDAL)(this.sqliteDriver);
        this.sindexDAL = new (require('./sqliteDAL/index/SIndexDAL').SIndexDAL)(this.sqliteDriver);
        this.cindexDAL = new (require('./sqliteDAL/index/CIndexDAL').CIndexDAL)(this.sqliteDriver);
        this.newDals = {
            'metaDAL': this.metaDAL,
            'blockDAL': this.blockDAL,
            'certDAL': this.certDAL,
            'msDAL': this.msDAL,
            'idtyDAL': this.idtyDAL,
            'txsDAL': this.txsDAL,
            'peerDAL': this.peerDAL,
            'confDAL': this.confDAL,
            'statDAL': this.statDAL,
            'walletDAL': this.walletDAL,
            'bindexDAL': this.bindexDAL,
            'mindexDAL': this.mindexDAL,
            'iindexDAL': this.iindexDAL,
            'sindexDAL': this.sindexDAL,
            'cindexDAL': this.cindexDAL
        };
    }
    init(conf) {
        return __awaiter(this, void 0, void 0, function* () {
            const dalNames = _.keys(this.newDals);
            for (const dalName of dalNames) {
                const dal = this.newDals[dalName];
                yield dal.init();
            }
            logger.debug("Upgrade database...");
            yield this.metaDAL.upgradeDatabase(conf);
            // Update the maximum certifications count a member can issue into the C++ addon
            const currencyParams = yield this.getParameters();
            if (currencyParams && currencyParams.sigStock !== undefined && currencyParams.sigStock !== null) {
                this.wotb.setMaxCert(currencyParams.sigStock);
            }
        });
    }
    getDBVersion() {
        return this.metaDAL.getVersion();
    }
    writeFileOfBlock(block) {
        return this.blockDAL.saveBlock(block);
    }
    writeSideFileOfBlock(block) {
        return this.blockDAL.saveSideBlock(block);
    }
    listAllPeers() {
        return this.peerDAL.listAll();
    }
    getPeer(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                return this.peerDAL.getPeer(pubkey);
            }
            catch (err) {
                throw Error('Unknown peer ' + pubkey);
            }
        });
    }
    getWS2Peers() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.peerDAL.getPeersWithEndpointsLike('WS2P');
        });
    }
    getBlock(number) {
        return __awaiter(this, void 0, void 0, function* () {
            const block = yield this.blockDAL.getBlock(number);
            return block || null;
        });
    }
    getAbsoluteBlockByNumberAndHash(number, hash) {
        return this.blockDAL.getAbsoluteBlock(number, hash);
    }
    getAbsoluteBlockByBlockstamp(blockstamp) {
        if (!blockstamp)
            throw "Blockstamp is required to find the block";
        const sp = blockstamp.split('-');
        const number = parseInt(sp[0]);
        const hash = sp[1];
        return this.getAbsoluteBlockByNumberAndHash(number, hash);
    }
    getBlockByBlockstampOrNull(blockstamp) {
        if (!blockstamp)
            throw "Blockstamp is required to find the block";
        const sp = blockstamp.split('-');
        const number = parseInt(sp[0]);
        const hash = sp[1];
        return this.getBlockByNumberAndHashOrNull(number, hash);
    }
    getBlockByBlockstamp(blockstamp) {
        if (!blockstamp)
            throw "Blockstamp is required to find the block";
        const sp = blockstamp.split('-');
        const number = parseInt(sp[0]);
        const hash = sp[1];
        return this.getBlockByNumberAndHash(number, hash);
    }
    getBlockByNumberAndHash(number, hash) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const block = yield this.getBlock(number);
                if (!block || block.hash != hash)
                    throw "Not found";
                else
                    return block;
            }
            catch (err) {
                throw 'Block ' + [number, hash].join('-') + ' not found';
            }
        });
    }
    getBlockByNumberAndHashOrNull(number, hash) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                return yield this.getBlockByNumberAndHash(number, hash);
            }
            catch (e) {
                return null;
            }
        });
    }
    existsNonChainableLink(from, vHEAD_1, sigStock) {
        return __awaiter(this, void 0, void 0, function* () {
            // Cert period rule
            const medianTime = vHEAD_1 ? vHEAD_1.medianTime : 0;
            const linksFrom = yield this.cindexDAL.reducablesFrom(from);
            const unchainables = _.filter(linksFrom, (link) => link.chainable_on > medianTime);
            if (unchainables.length > 0)
                return true;
            // Max stock rule
            let activeLinks = _.filter(linksFrom, (link) => !link.expired_on);
            return activeLinks.length >= sigStock;
        });
    }
    getCurrentBlockOrNull() {
        return __awaiter(this, void 0, void 0, function* () {
            let current = null;
            try {
                current = yield this.getBlockCurrent();
            }
            catch (e) {
                if (e != constants.ERROR.BLOCK.NO_CURRENT_BLOCK) {
                    throw e;
                }
            }
            return current;
        });
    }
    getPromoted(number) {
        return this.getBlock(number);
    }
    // Block
    lastUDBlock() {
        return this.blockDAL.lastBlockWithDividend();
    }
    getRootBlock() {
        return this.getBlock(0);
    }
    getPotentialRootBlocks() {
        return this.blockDAL.getPotentialRoots();
    }
    lastBlockOfIssuer(issuer) {
        return this.blockDAL.lastBlockOfIssuer(issuer);
    }
    getCountOfPoW(issuer) {
        return this.blockDAL.getCountOfBlocksIssuedBy(issuer);
    }
    getBlocksBetween(start, end) {
        return Promise.resolve(this.blockDAL.getBlocks(Math.max(0, start), end));
    }
    getForkBlocksFollowing(current) {
        return this.blockDAL.getNextForkBlocks(current.number, current.hash);
    }
    getPotentialForkBlocks(numberStart, medianTimeStart, maxNumber) {
        return this.blockDAL.getPotentialForkBlocks(numberStart, medianTimeStart, maxNumber);
    }
    getBlockCurrent() {
        return __awaiter(this, void 0, void 0, function* () {
            const current = yield this.blockDAL.getCurrent();
            if (!current)
                throw 'No current block';
            return current;
        });
    }
    getValidLinksTo(to) {
        return this.cindexDAL.getValidLinksTo(to);
    }
    getAvailableSourcesByPubkey(pubkey) {
        return this.sindexDAL.getAvailableForPubkey(pubkey);
    }
    getIdentityByHashOrNull(hash) {
        return __awaiter(this, void 0, void 0, function* () {
            const pending = yield this.idtyDAL.getByHash(hash);
            if (!pending) {
                return this.iindexDAL.getFromHash(hash);
            }
            return pending;
        });
    }
    getMembers() {
        return this.iindexDAL.getMembers();
    }
    // TODO: this should definitely be reduced by removing fillInMembershipsOfIdentity
    getWritten(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                return yield this.fillInMembershipsOfIdentity(this.iindexDAL.getFromPubkey(pubkey));
            }
            catch (err) {
                logger.error(err);
                return null;
            }
        });
    }
    getWrittenIdtyByPubkey(pubkey) {
        return this.iindexDAL.getFromPubkey(pubkey);
    }
    getWrittenIdtyByUID(uid) {
        return this.iindexDAL.getFromUID(uid);
    }
    fillInMembershipsOfIdentity(queryPromise) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const idty = yield Promise.resolve(queryPromise);
                if (idty) {
                    const mss = yield this.msDAL.getMembershipsOfIssuer(idty.pubkey);
                    const mssFromMindex = yield this.mindexDAL.reducable(idty.pubkey);
                    idty.memberships = mss.concat(mssFromMindex.map((ms) => {
                        const sp = ms.created_on.split('-');
                        return {
                            blockstamp: ms.created_on,
                            membership: ms.leaving ? 'OUT' : 'IN',
                            number: sp[0],
                            fpr: sp[1],
                            written_number: parseInt(ms.written_on)
                        };
                    }));
                    return idty;
                }
            }
            catch (err) {
                logger.error(err);
            }
            return null;
        });
    }
    findPeersWhoseHashIsIn(hashes) {
        return __awaiter(this, void 0, void 0, function* () {
            const peers = yield this.peerDAL.listAll();
            return _.chain(peers).filter((p) => hashes.indexOf(p.hash) !== -1).value();
        });
    }
    getTxByHash(hash) {
        return this.txsDAL.getTX(hash);
    }
    removeTxByHash(hash) {
        return this.txsDAL.removeTX(hash);
    }
    getTransactionsPending(versionMin = 0) {
        return this.txsDAL.getAllPending(versionMin);
    }
    getNonWritten(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            const pending = yield this.idtyDAL.getPendingIdentities();
            return _.chain(pending).where({ pubkey: pubkey }).value();
        });
    }
    getRevocatingMembers() {
        return __awaiter(this, void 0, void 0, function* () {
            const revoking = yield this.idtyDAL.getToRevoke();
            const toRevoke = [];
            for (const pending of revoking) {
                const idty = yield this.getWrittenIdtyByPubkey(pending.pubkey);
                if (!idty.revoked_on) {
                    toRevoke.push(pending);
                }
            }
            return toRevoke;
        });
    }
    getToBeKickedPubkeys() {
        return this.iindexDAL.getToBeKickedPubkeys();
    }
    getRevokedPubkeys() {
        return this.mindexDAL.getRevokedPubkeys();
    }
    searchJustIdentities(search) {
        return __awaiter(this, void 0, void 0, function* () {
            const pendings = yield this.idtyDAL.searchThoseMatching(search);
            const writtens = yield this.iindexDAL.searchThoseMatching(search);
            const nonPendings = _.filter(writtens, (w) => {
                return _.where(pendings, { pubkey: w.pub }).length == 0;
            });
            const found = pendings.concat(nonPendings.map((i) => {
                // Use the correct field
                i.pubkey = i.pub;
                return i;
            }));
            return yield Promise.all(found.map((f) => __awaiter(this, void 0, void 0, function* () {
                const ms = yield this.mindexDAL.getReducedMS(f.pub);
                if (ms) {
                    f.revoked_on = ms.revoked_on ? parseInt(ms.revoked_on) : null;
                    f.revoked = !!f.revoked_on;
                    f.revocation_sig = ms.revocation || null;
                }
                return f;
            })));
        });
    }
    certsToTarget(pub, hash) {
        return __awaiter(this, void 0, void 0, function* () {
            const certs = yield this.certDAL.getToTarget(hash);
            const links = yield this.cindexDAL.getValidLinksTo(pub);
            let matching = certs;
            yield Promise.all(links.map((entry) => __awaiter(this, void 0, void 0, function* () {
                entry.from = entry.issuer;
                const wbt = entry.written_on.split('-');
                const blockNumber = parseInt(entry.created_on); // created_on field of `c_index` does not have the full blockstamp
                const basedBlock = yield this.getBlock(blockNumber);
                entry.block = blockNumber;
                entry.block_number = blockNumber;
                entry.block_hash = basedBlock ? basedBlock.hash : null;
                entry.linked = true;
                entry.written_block = parseInt(wbt[0]);
                entry.written_hash = wbt[1];
                matching.push(entry);
            })));
            matching = _.sortBy(matching, (c) => -c.block);
            matching.reverse();
            return matching;
        });
    }
    certsFrom(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            const certs = yield this.certDAL.getFromPubkeyCerts(pubkey);
            const links = yield this.cindexDAL.getValidLinksFrom(pubkey);
            let matching = certs;
            yield Promise.all(links.map((entry) => __awaiter(this, void 0, void 0, function* () {
                const idty = yield this.getWrittenIdtyByPubkey(entry.receiver);
                entry.from = entry.issuer;
                entry.to = entry.receiver;
                const cbt = entry.created_on.split('-');
                const wbt = entry.written_on.split('-');
                entry.block = parseInt(cbt[0]);
                entry.block_number = parseInt(cbt[0]);
                entry.block_hash = cbt[1];
                entry.target = idty.hash;
                entry.linked = true;
                entry.written_block = parseInt(wbt[0]);
                entry.written_hash = wbt[1];
                matching.push(entry);
            })));
            matching = _.sortBy(matching, (c) => -c.block);
            matching.reverse();
            return matching;
        });
    }
    isSentry(pubkey, conf) {
        return __awaiter(this, void 0, void 0, function* () {
            const current = yield this.getCurrentBlockOrNull();
            if (current) {
                const dSen = Math.ceil(Math.pow(current.membersCount, 1 / conf.stepMax));
                const linksFrom = yield this.cindexDAL.getValidLinksFrom(pubkey);
                const linksTo = yield this.cindexDAL.getValidLinksTo(pubkey);
                return linksFrom.length >= dSen && linksTo.length >= dSen;
            }
            return false;
        });
    }
    certsFindNew() {
        return __awaiter(this, void 0, void 0, function* () {
            const certs = yield this.certDAL.getNotLinked();
            return _.chain(certs).where({ linked: false }).sortBy((c) => -c.block).value();
        });
    }
    certsNotLinkedToTarget(hash) {
        return __awaiter(this, void 0, void 0, function* () {
            const certs = yield this.certDAL.getNotLinkedToTarget(hash);
            return _.chain(certs).sortBy((c) => -c.block).value();
        });
    }
    getMostRecentMembershipNumberForIssuer(issuer) {
        return __awaiter(this, void 0, void 0, function* () {
            const mss = yield this.msDAL.getMembershipsOfIssuer(issuer);
            const reduced = yield this.mindexDAL.getReducedMS(issuer);
            let max = reduced ? parseInt(reduced.created_on) : -1;
            for (const ms of mss) {
                max = Math.max(ms.number, max);
            }
            return max;
        });
    }
    lastJoinOfIdentity(target) {
        return __awaiter(this, void 0, void 0, function* () {
            let pending = yield this.msDAL.getPendingINOfTarget(target);
            return _(pending).sortBy((ms) => -ms.number)[0];
        });
    }
    findNewcomers(blockMedianTime = 0) {
        return __awaiter(this, void 0, void 0, function* () {
            const pending = yield this.msDAL.getPendingIN();
            const mss = yield Promise.all(pending.map((p) => __awaiter(this, void 0, void 0, function* () {
                const reduced = yield this.mindexDAL.getReducedMS(p.issuer);
                if (!reduced || !reduced.chainable_on || blockMedianTime >= reduced.chainable_on || blockMedianTime < constants.TIME_TO_TURN_ON_BRG_107) {
                    return p;
                }
                return null;
            })));
            return _.chain(mss)
                .filter((ms) => ms)
                .sortBy((ms) => -ms.sigDate)
                .value();
        });
    }
    findLeavers() {
        return __awaiter(this, void 0, void 0, function* () {
            const mss = yield this.msDAL.getPendingOUT();
            return _.chain(mss).sortBy((ms) => -ms.sigDate).value();
        });
    }
    existsNonReplayableLink(from, to) {
        return this.cindexDAL.existsNonReplayableLink(from, to);
    }
    getSource(identifier, pos) {
        return this.sindexDAL.getSource(identifier, pos);
    }
    isMember(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const idty = yield this.iindexDAL.getFromPubkey(pubkey);
                if (!idty) {
                    return false;
                }
                return idty.member;
            }
            catch (err) {
                return false;
            }
        });
    }
    isMemberAndNonLeaver(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const idty = yield this.iindexDAL.getFromPubkey(pubkey);
                if (idty && idty.member) {
                    return !(yield this.isLeaving(pubkey));
                }
                return false;
            }
            catch (err) {
                return false;
            }
        });
    }
    isLeaving(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            const ms = yield this.mindexDAL.getReducedMS(pubkey);
            return (ms && ms.leaving) || false;
        });
    }
    existsCert(cert) {
        return __awaiter(this, void 0, void 0, function* () {
            const existing = yield this.certDAL.existsGivenCert(cert);
            if (existing)
                return existing;
            const existsLink = yield this.cindexDAL.existsNonReplayableLink(cert.from, cert.to);
            return !!existsLink;
        });
    }
    deleteCert(cert) {
        return this.certDAL.deleteCert(cert);
    }
    deleteMS(ms) {
        return this.msDAL.deleteMS(ms);
    }
    setRevoked(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            const idty = yield this.getWrittenIdtyByPubkey(pubkey);
            idty.revoked = true;
            return yield this.idtyDAL.saveIdentity(idty);
        });
    }
    getPeerOrNull(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            let peer = null;
            try {
                peer = yield this.getPeer(pubkey);
            }
            catch (e) {
                if (e != constants.ERROR.BLOCK.NO_CURRENT_BLOCK) {
                    throw e;
                }
            }
            return peer;
        });
    }
    removePeerByPubkey(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.peerDAL.removePeerByPubkey(pubkey);
        });
    }
    findAllPeersNEWUPBut(pubkeys) {
        return __awaiter(this, void 0, void 0, function* () {
            const peers = yield this.listAllPeers();
            return peers.filter((peer) => pubkeys.indexOf(peer.pubkey) == -1
                && ['UP'].indexOf(peer.status) !== -1);
        });
    }
    listAllPeersWithStatusNewUP() {
        return __awaiter(this, void 0, void 0, function* () {
            const peers = yield this.peerDAL.listAll();
            return _.chain(peers)
                .filter((p) => ['UP']
                .indexOf(p.status) !== -1).value();
        });
    }
    listAllPeersWithStatusNewUPWithtout(pub) {
        return __awaiter(this, void 0, void 0, function* () {
            const peers = yield this.peerDAL.listAll();
            return _.chain(peers).filter((p) => p.status == 'UP').filter((p) => p.pubkey !== pub).value();
        });
    }
    findPeers(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const peer = yield this.getPeer(pubkey);
                return [peer];
            }
            catch (err) {
                return [];
            }
        });
    }
    getRandomlyUPsWithout(pubkeys) {
        return __awaiter(this, void 0, void 0, function* () {
            const peers = yield this.listAllPeersWithStatusNewUP();
            return peers.filter((peer) => pubkeys.indexOf(peer.pubkey) == -1);
        });
    }
    setPeerUP(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const p = yield this.getPeer(pubkey);
                p.status = 'UP';
                p.first_down = null;
                p.last_try = null;
                return this.peerDAL.savePeer(p);
            }
            catch (err) {
                return null;
            }
        });
    }
    setPeerDown(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                // We do not set mirror peers as down (ex. of mirror: 'M1_HnFcSms8jzwngtVomTTnzudZx7SHUQY8sVE1y8yBmULk')
                if (!pubkey.match(/_/)) {
                    const p = yield this.getPeer(pubkey);
                    if (p) {
                        const now = (new Date()).getTime();
                        p.status = 'DOWN';
                        if (!p.first_down) {
                            p.first_down = now;
                        }
                        p.last_try = now;
                        yield this.peerDAL.savePeer(p);
                    }
                }
            }
            catch (err) {
                throw err;
            }
        });
    }
    saveBlock(block) {
        return __awaiter(this, void 0, void 0, function* () {
            const dbb = DBBlock_1.DBBlock.fromBlockDTO(block);
            dbb.wrong = false;
            yield Promise.all([
                this.saveBlockInFile(dbb),
                this.saveTxsInFiles(block.transactions, block.number, block.medianTime)
            ]);
        });
    }
    generateIndexes(block, conf, index, HEAD) {
        return __awaiter(this, void 0, void 0, function* () {
            // We need to recompute the indexes for block#0
            if (!index || !HEAD || HEAD.number == 0) {
                index = indexer.localIndex(block, conf);
                HEAD = yield indexer.completeGlobalScope(block, conf, index, this);
            }
            let mindex = indexer.mindex(index);
            let iindex = indexer.iindex(index);
            let sindex = indexer.sindex(index);
            let cindex = indexer.cindex(index);
            sindex = sindex.concat(yield indexer.ruleIndexGenDividend(HEAD, this));
            sindex = sindex.concat(yield indexer.ruleIndexGarbageSmallAccounts(HEAD, sindex, this));
            cindex = cindex.concat(yield indexer.ruleIndexGenCertificationExpiry(HEAD, this));
            mindex = mindex.concat(yield indexer.ruleIndexGenMembershipExpiry(HEAD, this));
            iindex = iindex.concat(yield indexer.ruleIndexGenExclusionByMembership(HEAD, mindex, this));
            iindex = iindex.concat(yield indexer.ruleIndexGenExclusionByCertificatons(HEAD, cindex, iindex, conf, this));
            mindex = mindex.concat(yield indexer.ruleIndexGenImplicitRevocation(HEAD, this));
            yield indexer.ruleIndexCorrectMembershipExpiryDate(HEAD, mindex, this);
            yield indexer.ruleIndexCorrectCertificationExpiryDate(HEAD, cindex, this);
            return { HEAD, mindex, iindex, sindex, cindex };
        });
    }
    updateWotbLinks(cindex) {
        return __awaiter(this, void 0, void 0, function* () {
            for (const entry of cindex) {
                const from = yield this.getWrittenIdtyByPubkey(entry.issuer);
                const to = yield this.getWrittenIdtyByPubkey(entry.receiver);
                if (entry.op == constants_1.CommonConstants.IDX_CREATE) {
                    this.wotb.addLink(from.wotb_id, to.wotb_id);
                }
                else {
                    // Update = removal
                    this.wotb.removeLink(from.wotb_id, to.wotb_id);
                }
            }
        });
    }
    trimIndexes(maxNumber) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.bindexDAL.trimBlocks(maxNumber);
            yield this.iindexDAL.trimRecords(maxNumber);
            yield this.mindexDAL.trimRecords(maxNumber);
            yield this.cindexDAL.trimExpiredCerts(maxNumber);
            yield this.sindexDAL.trimConsumedSource(maxNumber);
            return true;
        });
    }
    trimSandboxes(block) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.certDAL.trimExpiredCerts(block.medianTime);
            yield this.msDAL.trimExpiredMemberships(block.medianTime);
            yield this.idtyDAL.trimExpiredIdentities(block.medianTime);
            yield this.txsDAL.trimExpiredNonWrittenTxs(block.medianTime - constants_1.CommonConstants.TX_WINDOW);
            return true;
        });
    }
    savePendingMembership(ms) {
        return this.msDAL.savePendingMembership(ms);
    }
    saveBlockInFile(block) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.writeFileOfBlock(block);
        });
    }
    saveSideBlockInFile(block) {
        return this.writeSideFileOfBlock(block);
    }
    saveTxsInFiles(txs, block_number, medianTime) {
        return __awaiter(this, void 0, void 0, function* () {
            return Promise.all(txs.map((tx) => __awaiter(this, void 0, void 0, function* () {
                const sp = tx.blockstamp.split('-');
                tx.blockstampTime = (yield this.getBlockByNumberAndHash(parseInt(sp[0]), sp[1])).medianTime;
                const txEntity = TransactionDTO_1.TransactionDTO.fromJSONObject(tx);
                txEntity.computeAllHashes();
                return this.txsDAL.addLinked(TransactionDTO_1.TransactionDTO.fromJSONObject(txEntity), block_number, medianTime);
            })));
        });
    }
    merkleForPeers() {
        return __awaiter(this, void 0, void 0, function* () {
            let peers = yield this.listAllPeersWithStatusNewUP();
            const leaves = peers.map((peer) => peer.hash);
            const merkle = new MerkleDTO_1.MerkleDTO();
            merkle.initialize(leaves);
            return merkle;
        });
    }
    removeAllSourcesOfBlock(blockstamp) {
        return this.sindexDAL.removeBlock(blockstamp);
    }
    updateTransactions(txs) {
        return this.txsDAL.insertBatchOfTxs(txs);
    }
    savePendingIdentity(idty) {
        return this.idtyDAL.saveIdentity(idty);
    }
    revokeIdentity(pubkey) {
        return this.idtyDAL.revokeIdentity(pubkey);
    }
    removeUnWrittenWithPubkey(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.idtyDAL.removeUnWrittenWithPubkey(pubkey);
        });
    }
    removeUnWrittenWithUID(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.idtyDAL.removeUnWrittenWithUID(pubkey);
        });
    }
    registerNewCertification(cert) {
        return this.certDAL.saveNewCertification(cert);
    }
    saveTransaction(tx) {
        return this.txsDAL.addPending(TransactionDTO_1.TransactionDTO.fromJSONObject(tx));
    }
    getTransactionsHistory(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            const history = {
                sent: [],
                received: [],
                sending: [],
                receiving: [],
                pending: []
            };
            const res = yield Promise.all([
                this.txsDAL.getLinkedWithIssuer(pubkey),
                this.txsDAL.getLinkedWithRecipient(pubkey),
                this.txsDAL.getPendingWithIssuer(pubkey),
                this.txsDAL.getPendingWithRecipient(pubkey)
            ]);
            history.sent = res[0] || [];
            history.received = res[1] || [];
            history.sending = res[2] || [];
            history.pending = res[3] || [];
            return history;
        });
    }
    getUDHistory(pubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            const sources = yield this.sindexDAL.getUDSources(pubkey);
            return {
                history: sources.map((src) => _.extend({
                    block_number: src.pos,
                    time: src.written_time
                }, src))
            };
        });
    }
    savePeer(peer) {
        return this.peerDAL.savePeer(peer);
    }
    getUniqueIssuersBetween(start, end) {
        return __awaiter(this, void 0, void 0, function* () {
            const current = yield this.blockDAL.getCurrent();
            const firstBlock = Math.max(0, start);
            const lastBlock = Math.max(0, Math.min(current.number, end));
            const blocks = yield this.blockDAL.getBlocks(firstBlock, lastBlock);
            return _.chain(blocks).pluck('issuer').uniq().value();
        });
    }
    /**
     * Gets a range of entries for the last `start`th to the last `end`th HEAD entry.
     * @param start The starting entry number (min. 1)
     * @param end The ending entry (max. BINDEX length)
     * @param property If provided, transforms the range of entries into an array of the asked property.
     */
    range(start, end, property) {
        return __awaiter(this, void 0, void 0, function* () {
            const range = yield this.bindexDAL.range(start, end);
            if (property) {
                // Filter on a particular property
                return range.map((b) => b[property]);
            }
            else {
                return range;
            }
        });
    }
    /**
     * Get the last `n`th entry from the BINDEX.
     * @param n The entry number (min. 1).
     */
    head(n) {
        return this.bindexDAL.head(n);
    }
    /***********************
     *    CONFIGURATION
     **********************/
    getParameters() {
        return this.confDAL.getParameters();
    }
    loadConf(overrideConf, defaultConf = false) {
        return __awaiter(this, void 0, void 0, function* () {
            let conf = ConfDTO_1.ConfDTO.complete(overrideConf || {});
            if (!defaultConf) {
                const savedConf = yield this.confDAL.loadConf();
                const savedProxyConf = _(savedConf.proxyConf).extend({});
                conf = _(savedConf).extend(overrideConf || {});
                if (overrideConf.proxiesConf !== undefined) { }
                else {
                    conf.proxyConf = _(savedProxyConf).extend({});
                }
            }
            if (this.loadConfHook) {
                yield this.loadConfHook(conf);
            }
            return conf;
        });
    }
    saveConf(confToSave) {
        return __awaiter(this, void 0, void 0, function* () {
            // Save the conf in file
            let theConf = confToSave;
            if (this.saveConfHook) {
                theConf = yield this.saveConfHook(theConf);
            }
            return this.confDAL.saveConf(theConf);
        });
    }
    /***********************
     *     WALLETS
     **********************/
    getWallet(conditions) {
        return __awaiter(this, void 0, void 0, function* () {
            let wallet = yield this.walletDAL.getWallet(conditions);
            if (!wallet) {
                wallet = { conditions, balance: 0 };
            }
            return wallet;
        });
    }
    saveWallet(wallet) {
        return this.walletDAL.saveWallet(wallet);
    }
    /***********************
     *     STATISTICS
     **********************/
    loadStats() {
        return this.statDAL.loadStats();
    }
    getStat(name) {
        return this.statDAL.getStat(name);
    }
    pushStats(stats) {
        return this.statDAL.pushStats(stats);
    }
    cleanCaches() {
        return __awaiter(this, void 0, void 0, function* () {
            yield _.values(this.newDals).map((dal) => dal.cleanCache && dal.cleanCache());
        });
    }
    close() {
        return __awaiter(this, void 0, void 0, function* () {
            yield _.values(this.newDals).map((dal) => dal.cleanCache && dal.cleanCache());
            return this.sqliteDriver.closeConnection();
        });
    }
    resetPeers() {
        return __awaiter(this, void 0, void 0, function* () {
            this.peerDAL.removeAll();
            return yield this.close();
        });
    }
    getLogContent(linesQuantity) {
        return new Promise((resolve, reject) => {
            try {
                let lines = [], i = 0;
                const logPath = path.join(this.rootPath, 'duniter.log');
                const readStream = fs.createReadStream(logPath);
                readStream.on('error', (err) => reject(err));
                const lineReader = readline.createInterface({
                    input: readStream
                });
                lineReader.on('line', (line) => {
                    line = "\n" + line;
                    lines.push(line);
                    i++;
                    if (i >= linesQuantity)
                        lines.shift();
                });
                lineReader.on('close', () => resolve(lines));
                lineReader.on('error', (err) => reject(err));
            }
            catch (e) {
                reject(e);
            }
        });
    }
}
exports.FileDAL = FileDAL;
//# sourceMappingURL=fileDAL.js.map