"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require('underscore');
const path = require('path');
const DEEP_WRITE = true;
class CFSCore {
    constructor(rootPath, qfs) {
        this.rootPath = rootPath;
        this.qfs = qfs;
        this.deletedFolder = path.join(rootPath, '.deleted');
        this.deletionFolderPromise = null;
        /**
         * Creates the deletion folder before effective deletion.
         * @returns {*|any|Promise<void>} Promise of creation.
         */
        this.createDeletionFolder = () => this.deletionFolderPromise || (this.deletionFolderPromise = this.makeTree('.deleted'));
    }
    /**
     * READ operation of CFS. Reads given file. May lead to tree traversal if file is not found.
     * @param filePath Path to the file.
     * @returns {*} Promise for file content.
     */
    read(filePath) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const isDeleted = yield this.qfs.exists(path.join(this.deletedFolder, this.toRemoveFileName(filePath)));
                if (isDeleted) {
                    // A deleted file must be considered non-existant
                    return null;
                }
                return yield this.qfs.read(path.join(this.rootPath, filePath));
            }
            catch (e) {
                return null;
            }
        });
    }
    /**
     * READ operation of CFS. Reads given file. May lead to tree traversal if file is not found.
     * @param filePath Path to the file.
     * @returns {*} Promise for file content.
     */
    exists(filePath) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const isDeleted = yield this.qfs.exists(path.join(this.deletedFolder, this.toRemoveFileName(filePath)));
                if (isDeleted) {
                    // A deleted file must be considered non-existant
                    return false;
                }
                return yield this.qfs.exists(path.join(this.rootPath, filePath));
            }
            catch (e) {
                return null;
            }
        });
    }
    /**
     * LIST operation of CFS. List files at given location. Tree traversal.
     * @param ofPath Location folder to list files.
     * @param localLevel Limit listing to local level.
     * @returns {*} Promise of file names.
     */
    list(ofPath) {
        return __awaiter(this, void 0, void 0, function* () {
            const dirPath = path.normalize(ofPath);
            let files = [], folder = path.join(this.rootPath, dirPath);
            const hasDir = yield this.qfs.exists(folder);
            if (hasDir) {
                files = files.concat(yield this.qfs.list(folder));
            }
            const hasDeletedFiles = yield this.qfs.exists(this.deletedFolder);
            if (hasDeletedFiles) {
                const deletedFiles = yield this.qfs.list(this.deletedFolder);
                const deletedOfThisPath = deletedFiles.filter((f) => f.match(new RegExp('^' + this.toRemoveDirName(dirPath))));
                const locallyDeletedFiles = deletedOfThisPath.map((f) => f.replace(this.toRemoveDirName(dirPath), '')
                    .replace(/^__/, ''));
                files = _.difference(files, locallyDeletedFiles);
            }
            return _.uniq(files);
        });
    }
    ;
    /**
     * WRITE operation of CFS. Writes the file in local Core.
     * @param filePath Path to the file to write.
     * @param content String content to write.
     * @param deep Wether to make a deep write or not.
     */
    write(filePath, content, deep) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.qfs.write(path.join(this.rootPath, filePath), content);
        });
    }
    ;
    /**
     * REMOVE operation of CFS. Set given file as removed. Logical deletion since physical won't work due to the algorithm of CFS.
     * @param filePath File to set as removed.
     * @param deep Wether to remove the file in the root core or not.
     * @returns {*} Promise of removal.
     */
    remove(filePath, deep) {
        return __awaiter(this, void 0, void 0, function* () {
            // Make a deep physical deletion
            // Root core: physical deletion
            return this.qfs.remove(path.join(this.rootPath, filePath));
        });
    }
    /**
     * REMOVE operation of CFS. Set given file as removed. Logical deletion since physical won't work due to the algorithm of CFS.
     * @param filePath File to set as removed.
     * @returns {*} Promise of removal.
     */
    removeDeep(filePath) {
        return this.remove(filePath, DEEP_WRITE);
    }
    /**
     * Create a directory tree.
     * @param treePath Tree path to create.
     */
    makeTree(treePath) {
        return __awaiter(this, void 0, void 0, function* () {
            // Note: qfs.makeTree does not work on windows, so we implement it manually
            try {
                let normalized = path.normalize(treePath);
                let folders = normalized.split(path.sep);
                let folder = this.rootPath;
                for (let i = 0, len = folders.length; i < len; i++) {
                    folder = folder ? path.join(folder, folders[i]) : folders[i];
                    let exists = yield this.qfs.exists(folder);
                    if (!exists) {
                        yield this.qfs.makeDirectory(folder);
                    }
                }
            }
            catch (e) {
                if (e && e.code !== "EISDIR" && e.code !== "EEXIST")
                    throw e;
            }
        });
    }
    /**
     * Write JSON object to given file.
     * @param filePath File path.
     * @param content JSON content to stringify and write.
     * @param deep Wether to make a deep write or not.
     */
    writeJSON(filePath, content, deep = false) {
        return this.write(filePath, JSON.stringify(content, null, ' '), deep);
    }
    /**
     * Write JSON object to given file deeply in the core structure.
     * @param filePath File path.
     * @param content JSON content to stringify and write.
     */
    writeJSONDeep(filePath, content) {
        return this.writeJSON(filePath, content, DEEP_WRITE);
    }
    /**
     * Read a file and parse its content as JSON.
     * @param filePath File to read.
     */
    readJSON(filePath) {
        return __awaiter(this, void 0, void 0, function* () {
            let data;
            try {
                data = yield this.read(filePath);
                return JSON.parse(data);
            }
            catch (err) {
                if (data && err.message.match(/^Unexpected token {/)) {
                    // This is a bug thrown during Unit Tests with MEMORY_MODE true...
                    return JSON.parse(data.match(/^(.*)}{.*/)[1] + '}');
                }
                else if (err.message.match(/^Unexpected end of input/)) {
                    // Could not read, return empty object
                    return {};
                }
                throw err;
            }
        });
    }
    /**
     * Read contents of files at given path and parse it as JSON.
     * @param dirPath Path to get the files' contents.
     * @param localLevel Wether to read only local level or not.
     */
    listJSON(dirPath) {
        return this.list(dirPath).then((files) => __awaiter(this, void 0, void 0, function* () { return Promise.all(files.map((f) => this.readJSON(path.join(dirPath, f)))); }));
    }
    /**
     * Read contents of files at given LOCAL path and parse it as JSON.
     * @param dirPath Path to get the files' contents.
     */
    listJSONLocal(dirPath) {
        return this.listJSON(dirPath);
    }
    /**
     * Normalize the path of a dir to be used for file deletion matching.
     * @param dirPath Directory path to normalize.
     * @returns {string|ng.ILocationService|XML} Normalized dir path.
     */
    toRemoveDirName(dirPath) {
        if (!dirPath.match(/\/$/)) {
            dirPath += '/';
        }
        return path.normalize(dirPath).replace(/\//g, '__').replace(/\\/g, '__');
    }
    /**
     * Normalize the name of the deleted file.
     * @param filePath Full path of the file, included file name.
     * @returns {string|ng.ILocationService|XML} Normalized file name.
     */
    toRemoveFileName(filePath) {
        return path.normalize(filePath).replace(/\//g, '__').replace(/\\/g, '__');
    }
}
exports.CFSCore = CFSCore;
//# sourceMappingURL=CFSCore.js.map