"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Created by cgeek on 22/08/15.
 */
const _ = require('underscore');
const co = require('co');
const colors = require('colors');
const logger = require('../../logger').NewLogger('sqlite');
class AbstractSQLite {
    constructor(driver, table, pkFields = [], fields = [], arrays = [], booleans = [], bigintegers = [], transientFields = [], beforeSaveHook = null) {
        this.driver = driver;
        this.table = table;
        this.pkFields = pkFields;
        this.fields = fields;
        this.arrays = arrays;
        this.booleans = booleans;
        this.bigintegers = bigintegers;
        this.transientFields = transientFields;
        this.beforeSaveHook = beforeSaveHook;
    }
    query(sql, params = []) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                //logger.trace(sql, JSON.stringify(params || []));
                const start = Date.now();
                const res = yield this.driver.executeAll(sql, params || []);
                const duration = Date.now() - start;
                const entities = res.map((t) => this.toEntity(t));
                // Display result
                let msg = sql + ' | %s\t==> %s rows in %s ms';
                if (duration <= 2) {
                    msg = colors.green(msg);
                }
                else if (duration <= 5) {
                    msg = colors.yellow(msg);
                }
                else if (duration <= 10) {
                    msg = colors.magenta(msg);
                }
                else if (duration <= 100) {
                    msg = colors.red(msg);
                }
                logger.query(msg, JSON.stringify(params || []), entities.length, duration);
                return entities;
            }
            catch (e) {
                logger.error('ERROR >> %s', sql, JSON.stringify(params || []), e.stack || e.message || e);
                throw e;
            }
        });
    }
    cleanData() {
        return this.exec("DELETE FROM " + this.table);
    }
    sqlListAll() {
        return this.query("SELECT * FROM " + this.table);
    }
    sqlDeleteAll() {
        return this.cleanData();
    }
    sqlFind(obj, sortObj = {}) {
        const conditions = this.toConditionsArray(obj).join(' and ');
        const values = this.toParams(obj);
        const sortKeys = _.keys(sortObj);
        const sort = sortKeys.length ? ' ORDER BY ' + sortKeys.map((k) => "`" + k + "` " + (sortObj[k] ? 'DESC' : 'ASC')).join(',') : '';
        return this.query('SELECT * FROM ' + this.table + ' WHERE ' + conditions + sort, values);
    }
    sqlFindOne(obj, sortObj = null) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.sqlFind(obj, sortObj);
            return res[0];
        });
    }
    sqlFindLikeAny(obj) {
        const keys = _.keys(obj);
        return this.query('SELECT * FROM ' + this.table + ' WHERE ' + keys.map((k) => 'UPPER(`' + k + '`) like ?').join(' or '), keys.map((k) => obj[k].toUpperCase()));
    }
    sqlRemoveWhere(obj) {
        return __awaiter(this, void 0, void 0, function* () {
            const keys = _.keys(obj);
            yield this.query('DELETE FROM ' + this.table + ' WHERE ' + keys.map((k) => '`' + k + '` = ?').join(' and '), keys.map((k) => obj[k]));
        });
    }
    sqlExisting(entity) {
        return this.getEntity(entity);
    }
    saveEntity(entity) {
        return __awaiter(this, void 0, void 0, function* () {
            let toSave = entity;
            if (this.beforeSaveHook) {
                this.beforeSaveHook(toSave);
            }
            const existing = yield this.getEntity(toSave);
            if (existing) {
                toSave = this.toRow(toSave);
                const valorizations = this.fields.map((field) => '`' + field + '` = ?').join(', ');
                const conditions = this.getPKFields().map((field) => '`' + field + '` = ?').join(' and ');
                const setValues = this.fields.map((field) => toSave[field]);
                const condValues = this.getPKFields().map((k) => toSave[k]);
                yield this.query('UPDATE ' + this.table + ' SET ' + valorizations + ' WHERE ' + conditions, setValues.concat(condValues));
                return;
            }
            yield this.insert(toSave);
        });
    }
    insert(entity) {
        return __awaiter(this, void 0, void 0, function* () {
            const row = this.toRow(entity);
            const values = this.fields.map((f) => row[f]);
            yield this.query(this.getInsertQuery(), values);
        });
    }
    getEntity(entity) {
        return __awaiter(this, void 0, void 0, function* () {
            const conditions = this.getPKFields().map((field) => '`' + field + '` = ?').join(' and ');
            const params = this.toParams(entity, this.getPKFields());
            return (yield this.query('SELECT * FROM ' + this.table + ' WHERE ' + conditions, params))[0];
        });
    }
    deleteEntity(entity) {
        return __awaiter(this, void 0, void 0, function* () {
            const toSave = this.toRow(entity);
            if (this.beforeSaveHook) {
                this.beforeSaveHook(toSave);
            }
            const conditions = this.getPKFields().map((field) => '`' + field + '` = ?').join(' and ');
            const condValues = this.getPKFields().map((k) => toSave[k]);
            yield this.query('DELETE FROM ' + this.table + ' WHERE ' + conditions, condValues);
        });
    }
    exec(sql) {
        try {
            //console.warn(sql);
            return this.driver.executeSql(sql);
        }
        catch (e) {
            //console.error('ERROR >> %s', sql);
            throw e;
        }
    }
    getInsertQuery() {
        return "INSERT INTO " + this.table + " (" + this.fields.map(f => '`' + f + '`').join(',') + ") VALUES (" + "?,".repeat(this.fields.length - 1) + "?);";
    }
    getInsertHead() {
        const valuesKeys = this.fields;
        return 'INSERT INTO ' + this.table + " (" + valuesKeys.map(f => '`' + f + '`').join(',') + ") VALUES ";
    }
    getInsertValue(toSave) {
        if (this.beforeSaveHook) {
            this.beforeSaveHook(toSave);
        }
        const row = this.toRow(toSave);
        const valuesKeys = this.fields;
        const values = valuesKeys.map((field) => this.escapeToSQLite(row[field]));
        return "(" + values.join(',') + ")";
    }
    toInsertValues(entity) {
        const row = this.toRow(entity);
        const values = this.fields.map((f) => row[f]);
        const formatted = values.map((s) => this.escapeToSQLite(s));
        return "(" + formatted.join(',') + ")";
    }
    /**
     * Make a batch insert.
     * @param records The records to insert as a batch.
     */
    insertBatch(records) {
        return __awaiter(this, void 0, void 0, function* () {
            const queries = [];
            if (records.length) {
                const insert = this.getInsertHead();
                const values = records.map((src) => this.getInsertValue(src));
                queries.push(insert + '\n' + values.join(',\n') + ';');
            }
            if (queries.length) {
                yield this.exec(queries.join('\n'));
            }
        });
    }
    toConditionsArray(obj) {
        return _.keys(obj).map((k) => {
            if (obj[k].$lte !== undefined) {
                return '`' + k + '` <= ?';
            }
            else if (obj[k].$gte !== undefined) {
                return '`' + k + '` >= ?';
            }
            else if (obj[k].$gt !== undefined) {
                return '`' + k + '` > ?';
            }
            else if (obj[k].$lt !== undefined) {
                return '`' + k + '` < ?';
            }
            else if (obj[k].$null !== undefined) {
                return '`' + k + '` IS ' + (!obj[k].$null ? 'NOT' : '') + ' NULL';
            }
            else if (obj[k].$contains !== undefined) {
                return '`' + k + '` LIKE ?';
            }
            else {
                return '`' + k + '` = ?';
            }
        });
    }
    toParams(obj, fields = null) {
        let params = [];
        (fields || _.keys(obj)).forEach((f) => {
            if (obj[f].$null === undefined) {
                let pValue;
                if (obj[f].$lte !== undefined) {
                    pValue = obj[f].$lte;
                }
                else if (obj[f].$gte !== undefined) {
                    pValue = obj[f].$gte;
                }
                else if (obj[f].$gt !== undefined) {
                    pValue = obj[f].$gt;
                }
                else if (obj[f].$lt !== undefined) {
                    pValue = obj[f].$lt;
                }
                else if (obj[f].$null !== undefined) {
                    pValue = obj[f].$null;
                }
                else if (obj[f].$contains !== undefined) {
                    pValue = "%" + obj[f].$contains + "%";
                }
                else if (~this.bigintegers.indexOf(f) && typeof obj[f] !== "string") {
                    pValue = String(obj[f]);
                }
                else {
                    pValue = obj[f];
                }
                params.push(pValue);
            }
        });
        return params;
    }
    escapeToSQLite(val) {
        if (typeof val == "boolean") {
            // SQLite specific: true => 1, false => 0
            if (val !== null && val !== undefined) {
                return val ? 1 : 0;
            }
            else {
                return null;
            }
        }
        else if (typeof val == "string") {
            return "'" + val.replace(/'/g, "\\'") + "'";
        }
        else if (val === undefined) {
            return "null";
        }
        else {
            return JSON.stringify(val);
        }
    }
    getPKFields() {
        return this.pkFields;
    }
    toEntity(row) {
        for (const arr of this.arrays) {
            row[arr] = row[arr] ? JSON.parse(row[arr]) : [];
        }
        // Big integers are stored as strings to avoid data loss
        for (const bigint of this.bigintegers) {
            if (row[bigint] !== null && row[bigint] !== undefined) {
                row[bigint] = parseInt(row[bigint]);
            }
        }
        // Booleans
        for (const f of this.booleans) {
            row[f] = row[f] !== null ? Boolean(row[f]) : null;
        }
        // Transient
        for (const f of (this.transientFields || [])) {
            row[f] = row[f];
        }
        return row;
    }
    toRow(entity) {
        let row = {};
        for (const f of this.fields) {
            row[f] = entity[f];
        }
        for (const arr of this.arrays) {
            row[arr] = JSON.stringify(row[arr] || []);
        }
        // Big integers are stored as strings to avoid data loss
        for (const bigint of this.bigintegers) {
            if (entity[bigint] === null || entity[bigint] === undefined) {
                row[bigint] = null;
            }
            else {
                row[bigint] = String(entity[bigint]);
            }
        }
        return row;
    }
}
exports.AbstractSQLite = AbstractSQLite;
//# sourceMappingURL=AbstractSQLite.js.map