"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const AbstractSQLite_1 = require("./AbstractSQLite");
const constants = require('../../constants');
const IS_FORK = true;
const IS_NOT_FORK = false;
class BlockDAL extends AbstractSQLite_1.AbstractSQLite {
    constructor(driver) {
        super(driver, 'block', 
        // PK fields
        ['number', 'hash'], 
        // Fields
        ['fork', 'hash', 'inner_hash', 'signature', 'currency', 'issuer', 'issuersCount', 'issuersFrame', 'issuersFrameVar', 'parameters', 'previousHash', 'previousIssuer', 'version', 'membersCount', 'monetaryMass', 'UDTime', 'medianTime', 'dividend', 'unitbase', 'time', 'powMin', 'number', 'nonce', 'transactions', 'certifications', 'identities', 'joiners', 'actives', 'leavers', 'revoked', 'excluded', 'len'], 
        // Arrays
        ['identities', 'certifications', 'actives', 'revoked', 'excluded', 'leavers', 'joiners', 'transactions'], 
        // Booleans
        ['wrong'], 
        // BigIntegers
        ['monetaryMass'], 
        // Transient
        []);
        /**
         * Periodically cleans the current block cache.
         * It seems the cache is not always correct and may stuck the node, so it is preferable to reset it periodically.
         */
        setInterval(this.cleanCache, constants.CURRENT_BLOCK_CACHE_DURATION);
    }
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.exec('BEGIN;' +
                'CREATE TABLE IF NOT EXISTS ' + this.table + ' (' +
                'fork BOOLEAN NOT NULL,' +
                'hash VARCHAR(64) NOT NULL,' +
                'inner_hash VARCHAR(64) NOT NULL,' +
                'signature VARCHAR(100) NOT NULL,' +
                'currency VARCHAR(50) NOT NULL,' +
                'issuer VARCHAR(50) NOT NULL,' +
                'parameters VARCHAR(255),' +
                'previousHash VARCHAR(64),' +
                'previousIssuer VARCHAR(50),' +
                'version INTEGER NOT NULL,' +
                'membersCount INTEGER NOT NULL,' +
                'monetaryMass VARCHAR(100) DEFAULT \'0\',' +
                'UDTime DATETIME,' +
                'medianTime DATETIME NOT NULL,' +
                'dividend INTEGER DEFAULT \'0\',' +
                'unitbase INTEGER NULL,' +
                'time DATETIME NOT NULL,' +
                'powMin INTEGER NOT NULL,' +
                'number INTEGER NOT NULL,' +
                'nonce INTEGER NOT NULL,' +
                'transactions TEXT,' +
                'certifications TEXT,' +
                'identities TEXT,' +
                'joiners TEXT,' +
                'actives TEXT,' +
                'leavers TEXT,' +
                'revoked TEXT,' +
                'excluded TEXT,' +
                'created DATETIME DEFAULT NULL,' +
                'updated DATETIME DEFAULT NULL,' +
                'PRIMARY KEY (number,hash)' +
                ');' +
                'CREATE INDEX IF NOT EXISTS idx_block_hash ON block (hash);' +
                'CREATE INDEX IF NOT EXISTS idx_block_fork ON block (fork);' +
                'COMMIT;');
        });
    }
    cleanCache() {
        this.current = null;
    }
    getCurrent() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.current) {
                this.current = (yield this.query('SELECT * FROM block WHERE NOT fork ORDER BY number DESC LIMIT 1'))[0];
            }
            return Promise.resolve(this.current);
        });
    }
    getBlock(number) {
        return __awaiter(this, void 0, void 0, function* () {
            return (yield this.query('SELECT * FROM block WHERE number = ? and NOT fork', [parseInt(String(number))]))[0];
        });
    }
    getAbsoluteBlock(number, hash) {
        return __awaiter(this, void 0, void 0, function* () {
            return (yield this.query('SELECT * FROM block WHERE number = ? and hash = ?', [number, hash]))[0];
        });
    }
    getBlocks(start, end) {
        return this.query('SELECT * FROM block WHERE number BETWEEN ? and ? and NOT fork ORDER BY number ASC', [start, end]);
    }
    lastBlockWithDividend() {
        return __awaiter(this, void 0, void 0, function* () {
            return (yield this.query('SELECT * FROM block WHERE dividend > 0 and NOT fork ORDER BY number DESC LIMIT 1'))[0];
        });
    }
    lastBlockOfIssuer(issuer) {
        return __awaiter(this, void 0, void 0, function* () {
            return (yield this.query('SELECT * FROM block WHERE issuer = ? and NOT fork ORDER BY number DESC LIMIT 1', [issuer]))[0];
        });
    }
    getCountOfBlocksIssuedBy(issuer) {
        return __awaiter(this, void 0, void 0, function* () {
            let res = yield this.query('SELECT COUNT(*) as quantity FROM block WHERE issuer = ? and NOT fork', [issuer]);
            return res[0].quantity;
        });
    }
    getForkBlocks() {
        return this.query('SELECT * FROM block WHERE fork ORDER BY number');
    }
    getPotentialForkBlocks(numberStart, medianTimeStart, maxNumber) {
        return this.query('SELECT * FROM block WHERE fork AND number >= ? AND number <= ? AND medianTime >= ? ORDER BY number DESC', [numberStart, maxNumber, medianTimeStart]);
    }
    getPotentialRoots() {
        return this.query('SELECT * FROM block WHERE fork AND number = ?', [0]);
    }
    getDividendBlocks() {
        return this.query('SELECT * FROM block WHERE dividend IS NOT NULL ORDER BY number');
    }
    saveBunch(blocks) {
        return __awaiter(this, void 0, void 0, function* () {
            let queries = "INSERT INTO block (" + this.fields.join(',') + ") VALUES ";
            for (let i = 0, len = blocks.length; i < len; i++) {
                let block = blocks[i];
                queries += this.toInsertValues(block);
                if (i + 1 < len) {
                    queries += ",\n";
                }
            }
            yield this.exec(queries);
            this.cleanCache();
        });
    }
    saveBlock(block) {
        return __awaiter(this, void 0, void 0, function* () {
            let saved = yield this.saveBlockAs(block, IS_NOT_FORK);
            if (!this.current || this.current.number < block.number) {
                this.current = block;
            }
            return saved;
        });
    }
    saveSideBlock(block) {
        return this.saveBlockAs(block, IS_FORK);
    }
    saveBlockAs(block, fork) {
        return __awaiter(this, void 0, void 0, function* () {
            block.fork = fork;
            return yield this.saveEntity(block);
        });
    }
    setSideBlock(number, previousBlock) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.query('UPDATE block SET fork = ? WHERE number = ?', [true, number]);
            this.current = previousBlock;
        });
    }
    getNextForkBlocks(number, hash) {
        return this.query('SELECT * FROM block WHERE fork AND number = ? AND previousHash like ? ORDER BY number', [number + 1, hash]);
    }
}
exports.BlockDAL = BlockDAL;
//# sourceMappingURL=BlockDAL.js.map