"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const AbstractIndex_1 = require("../AbstractIndex");
const indexer_1 = require("../../../indexer");
const _ = require('underscore');
class IIndexDAL extends AbstractIndex_1.AbstractIndex {
    constructor(driver) {
        super(driver, 'i_index', 
        // PK fields
        ['op', 'pub', 'created_on', 'written_on'], 
        // Fields
        [
            'op',
            'uid',
            'pub',
            'hash',
            'sig',
            'created_on',
            'written_on',
            'writtenOn',
            'member',
            'wasMember',
            'kick',
            'wotb_id'
        ], 
        // Arrays
        [], 
        // Booleans
        ['member', 'wasMember', 'kick'], 
        // BigIntegers
        [], 
        // Transient
        []);
    }
    init() {
        return this.exec('BEGIN;' +
            'CREATE TABLE IF NOT EXISTS ' + this.table + ' (' +
            'op VARCHAR(10) NOT NULL,' +
            'uid VARCHAR(100) NULL,' +
            'pub VARCHAR(50) NOT NULL,' +
            'hash VARCHAR(80) NULL,' +
            'sig VARCHAR(80) NULL,' +
            'created_on VARCHAR(80) NULL,' +
            'written_on VARCHAR(80) NOT NULL,' +
            'member BOOLEAN NULL,' +
            'wasMember BOOLEAN NULL,' +
            'kick BOOLEAN NULL,' +
            'wotb_id INTEGER NULL,' +
            'PRIMARY KEY (op,pub,created_on,written_on)' +
            ');' +
            'CREATE INDEX IF NOT EXISTS idx_iindex_pub ON i_index (pub);' +
            'COMMIT;');
    }
    getMembers() {
        return __awaiter(this, void 0, void 0, function* () {
            // All those who has been subject to, or who are currently subject to kicking. Make one result per pubkey.
            const pubkeys = yield this.query('SELECT DISTINCT(pub) FROM ' + this.table);
            // We get the full representation for each member
            const reduced = yield Promise.all(pubkeys.map((entry) => __awaiter(this, void 0, void 0, function* () {
                const reducable = yield this.reducable(entry.pub);
                return indexer_1.Indexer.DUP_HELPERS.reduce(reducable);
            })));
            // Filter on those to be kicked, return their pubkey
            const filtered = _.filter(reduced, (entry) => entry.member);
            return filtered.map((t) => this.toCorrectEntity(t));
        });
    }
    getMembersPubkeys() {
        return this.query('SELECT i1.pub ' +
            'FROM i_index i1 ' +
            'WHERE i1.member ' +
            'AND CAST(i1.written_on as int) = (' +
            ' SELECT MAX(CAST(i2.written_on as int)) ' +
            ' FROM i_index i2 ' +
            ' WHERE i1.pub = i2.pub ' +
            ' AND i2.member IS NOT NULL' +
            ')');
    }
    getToBeKickedPubkeys() {
        return __awaiter(this, void 0, void 0, function* () {
            // All those who has been subject to, or who are currently subject to kicking. Make one result per pubkey.
            const reducables = indexer_1.Indexer.DUP_HELPERS.reduceBy(yield this.sqlFind({ kick: true }), ['pub']);
            // We get the full representation for each member
            const reduced = yield Promise.all(reducables.map((entry) => __awaiter(this, void 0, void 0, function* () {
                const reducable = yield this.reducable(entry.pub);
                return indexer_1.Indexer.DUP_HELPERS.reduce(reducable);
            })));
            // Filter on those to be kicked, return their pubkey
            return _.filter(reduced, (entry) => entry.kick).map((entry) => entry.pub);
        });
    }
    searchThoseMatching(search) {
        return __awaiter(this, void 0, void 0, function* () {
            const reducables = indexer_1.Indexer.DUP_HELPERS.reduceBy(yield this.sqlFindLikeAny({
                pub: "%" + search + "%",
                uid: "%" + search + "%"
            }), ['pub']);
            // We get the full representation for each member
            return yield Promise.all(reducables.map((entry) => __awaiter(this, void 0, void 0, function* () {
                return this.toCorrectEntity(indexer_1.Indexer.DUP_HELPERS.reduce(yield this.reducable(entry.pub)));
            })));
        });
    }
    getFromPubkey(pubkey) {
        return this.entityOrNull('pub', pubkey);
    }
    getFromUID(uid) {
        return this.entityOrNull('uid', uid);
    }
    getFromHash(hash) {
        return this.entityOrNull('hash', hash, true);
    }
    reducable(pub) {
        return this.query('SELECT * FROM ' + this.table + ' WHERE pub = ? ORDER BY CAST(written_on as integer) ASC', [pub]);
    }
    removeBlock(blockstamp) {
        return this.exec('DELETE FROM ' + this.table + ' WHERE written_on = \'' + blockstamp + '\'');
    }
    entityOrNull(field, value, retrieveOnField = false) {
        return __awaiter(this, void 0, void 0, function* () {
            let reducable = yield this.query('SELECT * FROM ' + this.table + ' WHERE ' + field + ' = ?', [value]);
            if (reducable.length) {
                if (retrieveOnField) {
                    // Force full retrieval on `pub` field
                    reducable = yield this.query('SELECT * FROM ' + this.table + ' WHERE pub = ? ORDER BY CAST(written_on as int) ASC', [reducable[0].pub]);
                }
                return this.toCorrectEntity(indexer_1.Indexer.DUP_HELPERS.reduce(reducable));
            }
            return null;
        });
    }
    toCorrectEntity(row) {
        // Old field
        return {
            pubkey: row.pub,
            pub: row.pub,
            buid: row.created_on,
            revocation_sig: null,
            uid: row.uid,
            hash: row.hash,
            sig: row.sig,
            created_on: row.created_on,
            member: row.member,
            wasMember: row.wasMember,
            kick: row.kick,
            wotb_id: row.wotb_id,
            age: row.age,
            index: row.index,
            op: row.op,
            writtenOn: row.writtenOn,
            written_on: row.written_on
        };
    }
}
exports.IIndexDAL = IIndexDAL;
//# sourceMappingURL=IIndexDAL.js.map