"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const PeerDAL_1 = require("../dal/sqliteDAL/PeerDAL");
const common_1 = require("../common");
const constants_1 = require("../common-libs/constants");
class PeerDTO {
    constructor(version, currency, pubkey, blockstamp, endpoints, signature, status, statusTS, member = false) {
        this.version = version;
        this.currency = currency;
        this.pubkey = pubkey;
        this.blockstamp = blockstamp;
        this.endpoints = endpoints;
        this.signature = signature;
        this.status = status;
        this.statusTS = statusTS;
        this.member = false;
        this.member = member;
    }
    clone() {
        return PeerDTO.fromJSONObject(this);
    }
    get block() {
        return this.blockstamp;
    }
    blockNumber() {
        return parseInt(this.blockstamp);
    }
    keyID() {
        return this.pubkey && this.pubkey.length > 10 ? this.pubkey.substring(0, 10) : "Unknown";
    }
    getRawUnsigned() {
        return this.getRaw();
    }
    getRaw() {
        let raw = "";
        raw += "Version: " + this.version + "\n";
        raw += "Type: Peer\n";
        raw += "Currency: " + this.currency + "\n";
        raw += "PublicKey: " + this.pubkey + "\n";
        raw += "Block: " + this.blockstamp + "\n";
        raw += "Endpoints:" + "\n";
        for (const ep of this.endpoints) {
            raw += ep + "\n";
        }
        return raw;
    }
    getRawSigned() {
        return this.getRaw() + this.signature + "\n";
    }
    json() {
        return {
            version: this.version,
            currency: this.currency,
            endpoints: this.endpoints,
            status: this.status,
            block: this.block,
            signature: this.signature,
            raw: this.getRawSigned(),
            pubkey: this.pubkey
        };
    }
    getBMA() {
        let bma = null;
        this.endpoints.forEach((ep) => {
            const matches = !bma && ep.match(constants_1.CommonConstants.BMA_REGEXP);
            if (matches) {
                bma = {
                    "dns": matches[2] || '',
                    "ipv4": matches[4] || '',
                    "ipv6": matches[6] || '',
                    "port": matches[8] || 9101
                };
            }
        });
        return bma || {};
    }
    getWS2P(canReachTorEp, canReachClearEp) {
        let api = null;
        const endpointRegexp = (canReachTorEp) ? constants_1.CommonConstants.WS2PTOR_REGEXP : constants_1.CommonConstants.WS2P_REGEXP;
        for (const ep of this.endpoints) {
            if (canReachTorEp) {
                const matches = ep.match(constants_1.CommonConstants.WS2PTOR_REGEXP);
                if (matches) {
                    return {
                        uuid: matches[1],
                        host: matches[2] || '',
                        port: parseInt(matches[3]) || 0,
                        path: matches[4]
                    };
                }
            }
            if (canReachClearEp) {
                const matches = !api && ep.match(constants_1.CommonConstants.WS2P_REGEXP);
                if (matches) {
                    api = {
                        uuid: matches[1],
                        host: matches[2] || '',
                        port: parseInt(matches[3]) || 0,
                        path: matches[4]
                    };
                }
            }
        }
        return api || null;
    }
    getDns() {
        const bma = this.getBMA();
        return bma.dns ? bma.dns : null;
    }
    getIPv4() {
        const bma = this.getBMA();
        return bma.ipv4 ? bma.ipv4 : null;
    }
    getIPv6() {
        const bma = this.getBMA();
        return bma.ipv6 ? bma.ipv6 : null;
    }
    getPort() {
        const bma = this.getBMA();
        return bma.port ? bma.port : null;
    }
    getHostPreferDNS() {
        const bma = this.getBMA();
        return (bma.dns ? bma.dns :
            (bma.ipv4 ? bma.ipv4 :
                (bma.ipv6 ? bma.ipv6 : '')));
    }
    getURL() {
        const bma = this.getBMA();
        let base = this.getHostPreferDNS();
        if (bma.port)
            base += ':' + bma.port;
        return base;
    }
    hasValid4(bma) {
        return !!(bma.ipv4 && !bma.ipv4.match(/^127.0/) && !bma.ipv4.match(/^192.168/));
    }
    getNamedURL() {
        return this.getURL();
    }
    isReachable() {
        return !!(this.getURL());
    }
    containsEndpoint(ep) {
        return this.endpoints.reduce((found, endpoint) => found || endpoint == ep, false);
    }
    containsAllEndpoints(endpoints) {
        for (const ep of endpoints) {
            if (!this.containsEndpoint(ep)) {
                return false;
            }
        }
        return true;
    }
    endpointSum() {
        return this.endpoints.join('_');
    }
    getHash() {
        return common_1.hashf(this.getRawSigned());
    }
    toDBPeer() {
        const p = new PeerDAL_1.DBPeer();
        p.version = this.version;
        p.currency = this.currency;
        p.status = this.status || "DOWN";
        p.statusTS = this.statusTS || 0;
        p.hash = this.getHash();
        p.first_down = 0;
        p.last_try = 0;
        p.pubkey = this.pubkey;
        p.block = this.block;
        p.signature = this.signature;
        p.endpoints = this.endpoints;
        p.raw = this.getRawSigned();
        return p;
    }
    static blockNumber(blockstamp) {
        return parseInt(blockstamp);
    }
    static fromDBPeer(p) {
        return new PeerDTO(p.version, p.currency, p.pubkey, p.block, p.endpoints, p.signature, p.status, p.statusTS, false);
    }
    static fromJSONObject(obj) {
        return new PeerDTO(parseInt(obj.version), obj.currency || "", obj.pubkey || obj.pub || obj.issuer || "", obj.blockstamp || obj.block, obj.endpoints || [], obj.signature || obj.sig, obj.status || "DOWN", obj.statusTS || 0, obj.member);
    }
    static endpoint2host(endpoint) {
        return PeerDTO.fromJSONObject({ endpoints: [endpoint] }).getURL();
    }
    static indexOfFirst(endpoints, intoEndpoints) {
        for (let i = 0; i < intoEndpoints.length; i++) {
            const index = endpoints.indexOf(intoEndpoints[i]);
            if (index !== -1) {
                return index;
            }
        }
        return 0;
    }
}
exports.PeerDTO = PeerDTO;
//# sourceMappingURL=PeerDTO.js.map