"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const TransactionDTO_1 = require("../dto/TransactionDTO");
const keyring_1 = require("../common-libs/crypto/keyring");
const index_1 = require("../common-libs/index");
const constants_1 = require("../common-libs/constants");
const IdentityDTO_1 = require("../dto/IdentityDTO");
const common_1 = require("../common");
const indexer_1 = require("../indexer");
const _ = require('underscore');
const constants = constants_1.CommonConstants;
// Empty logger by default
let logger = {
    debug: (...args) => { },
    warn: (...args) => { }
};
function evalParams(params, conditions = '', sigResult) {
    const res = [];
    const issuers = sigResult.sigs.map(s => s.k);
    for (const func of params) {
        if (func.match(/^SIG/)) {
            const param = func.match(/^SIG\((.*)\)$/)[1];
            const index = parseInt(param);
            const sigEntry = !isNaN(index) && index < issuers.length && sigResult.sigs[index];
            const signatory = sigEntry || { k: '', ok: false };
            res.push({
                funcName: 'SIG',
                parameter: signatory.k,
                successful: signatory.ok
            });
        }
        else if (func.match(/^XHX/)) {
            const password = func.match(/^XHX\((.*)\)$/)[1];
            const hash = common_1.hashf(password);
            res.push({
                funcName: 'XHX',
                parameter: password,
                successful: conditions.indexOf('XHX(' + hash + ')') !== -1
            });
        }
    }
    return res;
}
exports.evalParams = evalParams;
exports.GLOBAL_RULES_FUNCTIONS = {
    checkIdentitiesAreWritable: (block, conf, dal) => __awaiter(this, void 0, void 0, function* () {
        let current = yield dal.getCurrentBlockOrNull();
        for (const obj of block.identities) {
            let idty = IdentityDTO_1.IdentityDTO.fromInline(obj);
            let found = yield dal.getWrittenIdtyByUID(idty.uid);
            if (found) {
                throw Error('Identity already used');
            }
            // Because the window rule does not apply on initial certifications
            if (current && idty.buid != constants.SPECIAL_BLOCK) {
                // From DUP 0.5: we fully check the blockstamp
                const basedBlock = yield dal.getBlockByBlockstamp(idty.buid);
                // Check if writable
                let duration = current.medianTime - parseInt(basedBlock.medianTime);
                if (duration > conf.idtyWindow) {
                    throw Error('Identity is too old and cannot be written');
                }
            }
        }
        return true;
    }),
    checkSourcesAvailability: (block, conf, dal, alsoCheckPendingTransactions) => __awaiter(this, void 0, void 0, function* () {
        const txs = block.transactions;
        const current = yield dal.getCurrentBlockOrNull();
        for (const tx of txs) {
            const inputs = tx.inputsAsObjects();
            const outputs = tx.outputsAsObjects();
            let unlocks = {};
            let sumOfInputs = 0;
            let maxOutputBase = current.unitbase;
            for (const theUnlock of tx.unlocks) {
                let sp = theUnlock.split(':');
                let index = parseInt(sp[0]);
                unlocks[index] = sp[1];
            }
            for (let k = 0, len2 = inputs.length; k < len2; k++) {
                let src = inputs[k];
                let dbSrc = yield dal.getSource(src.identifier, src.pos);
                logger.debug('Source %s:%s:%s:%s = %s', src.amount, src.base, src.identifier, src.pos, dbSrc && dbSrc.consumed);
                if (!dbSrc && alsoCheckPendingTransactions) {
                    // For chained transactions which are checked on sandbox submission, we accept them if there is already
                    // a previous transaction of the chain already recorded in the pool
                    dbSrc = yield (() => __awaiter(this, void 0, void 0, function* () {
                        let hypotheticSrc = null;
                        let targetTX = yield dal.getTxByHash(src.identifier);
                        if (targetTX) {
                            let outputStr = targetTX.outputs[src.pos];
                            if (outputStr) {
                                hypotheticSrc = TransactionDTO_1.TransactionDTO.outputStr2Obj(outputStr);
                                hypotheticSrc.consumed = false;
                                hypotheticSrc.time = 0;
                            }
                        }
                        return hypotheticSrc;
                    }))();
                }
                if (!dbSrc || dbSrc.consumed) {
                    logger.warn('Source ' + [src.type, src.identifier, src.pos].join(':') + ' is not available');
                    throw constants.ERRORS.SOURCE_ALREADY_CONSUMED;
                }
                sumOfInputs += dbSrc.amount * Math.pow(10, dbSrc.base);
                if (block.medianTime - dbSrc.written_time < tx.locktime) {
                    throw constants.ERRORS.LOCKTIME_PREVENT;
                }
                let unlockValues = unlocks[k];
                let unlocksForCondition = (unlockValues || '').split(' ');
                let unlocksMetadata = {};
                if (dbSrc.conditions) {
                    if (dbSrc.conditions.match(/CLTV/)) {
                        unlocksMetadata.currentTime = block.medianTime;
                    }
                    if (dbSrc.conditions.match(/CSV/)) {
                        unlocksMetadata.elapsedTime = block.medianTime - dbSrc.written_time;
                    }
                    const sigs = tx.getTransactionSigResult();
                    try {
                        if (!index_1.txunlock(dbSrc.conditions, unlocksForCondition, sigs, unlocksMetadata)) {
                            throw Error('Locked');
                        }
                    }
                    catch (e) {
                        logger.warn('Source ' + [src.amount, src.base, src.type, src.identifier, src.pos].join(':') + ' unlock fail');
                        throw constants.ERRORS.WRONG_UNLOCKER;
                    }
                }
                else {
                    throw Error("Source with no conditions");
                }
            }
            let sumOfOutputs = outputs.reduce(function (p, output) {
                if (output.base > maxOutputBase) {
                    throw constants.ERRORS.WRONG_OUTPUT_BASE;
                }
                return p + output.amount * Math.pow(10, output.base);
            }, 0);
            if (sumOfInputs !== sumOfOutputs) {
                logger.warn('Inputs/Outputs != 1 (%s/%s)', sumOfInputs, sumOfOutputs);
                throw constants.ERRORS.WRONG_AMOUNTS;
            }
        }
        return true;
    })
};
exports.GLOBAL_RULES_HELPERS = {
    // Functions used in an external context too
    checkMembershipBlock: (ms, current, conf, dal) => checkMSTarget(ms, current ? { number: current.number + 1 } : { number: 0 }, conf, dal),
    checkCertificationIsValid: (cert, current, findIdtyFunc, conf, dal) => {
        return checkCertificationIsValid(current ? current : { number: 0, currency: '' }, cert, findIdtyFunc, conf, dal);
    },
    checkCertificationIsValidForBlock: (cert, block, findIdtyFunc, conf, dal) => {
        return checkCertificationIsValid(block, cert, findIdtyFunc, conf, dal);
    },
    isOver3Hops: (member, newLinks, newcomers, current, conf, dal) => __awaiter(this, void 0, void 0, function* () {
        if (!current) {
            return Promise.resolve(false);
        }
        try {
            return indexer_1.Indexer.DUP_HELPERS.checkPeopleAreNotOudistanced([member], newLinks, newcomers, conf, dal);
        }
        catch (e) {
            return true;
        }
    }),
    checkExistsUserID: (uid, dal) => dal.getWrittenIdtyByUID(uid),
    checkExistsPubkey: (pub, dal) => dal.getWrittenIdtyByPubkey(pub),
    checkSingleTransaction: (tx, block, conf, dal, alsoCheckPendingTransactions = false) => exports.GLOBAL_RULES_FUNCTIONS.checkSourcesAvailability({
        transactions: [tx],
        medianTime: block.medianTime
    }, conf, dal, alsoCheckPendingTransactions),
    checkTxBlockStamp: (tx, dal) => __awaiter(this, void 0, void 0, function* () {
        const number = parseInt(tx.blockstamp.split('-')[0]);
        const hash = tx.blockstamp.split('-')[1];
        const basedBlock = yield dal.getBlockByNumberAndHashOrNull(number, hash);
        if (!basedBlock) {
            throw "Wrong blockstamp for transaction";
        }
        // Valuates the blockstampTime field
        tx.blockstampTime = basedBlock.medianTime;
        const current = yield dal.getCurrentBlockOrNull();
        if (current && current.medianTime > basedBlock.medianTime + constants.TX_WINDOW) {
            throw "Transaction has expired";
        }
    })
};
/*****************************
 *
 *      UTILITY FUNCTIONS
 *
 *****************************/
function checkMSTarget(ms, block, conf, dal) {
    return __awaiter(this, void 0, void 0, function* () {
        if (block.number == 0 && ms.number != 0) {
            throw Error('Number must be 0 for root block\'s memberships');
        }
        else if (block.number == 0 && ms.fpr != 'E3B0C44298FC1C149AFBF4C8996FB92427AE41E4649B934CA495991B7852B855') {
            throw Error('Hash must be E3B0C44298FC1C149AFBF4C8996FB92427AE41E4649B934CA495991B7852B855 for root block\'s memberships');
        }
        else if (block.number == 0) {
            return null; // Valid for root block
        }
        else {
            let basedBlock;
            try {
                basedBlock = yield dal.getBlockByNumberAndHash(ms.number, ms.fpr);
            }
            catch (e) {
                throw Error('Membership based on an unexisting block');
            }
            let current = yield dal.getCurrentBlockOrNull();
            if (current && current.medianTime > basedBlock.medianTime + conf.msValidity) {
                throw Error('Membership has expired');
            }
            return basedBlock;
        }
    });
}
function checkCertificationIsValid(block, cert, findIdtyFunc, conf, dal) {
    return __awaiter(this, void 0, void 0, function* () {
        if (block.number == 0 && cert.block_number != 0) {
            throw Error('Number must be 0 for root block\'s certifications');
        }
        else {
            let basedBlock = {
                hash: constants.SPECIAL_HASH
            };
            if (block.number != 0) {
                try {
                    basedBlock = yield dal.getBlock(cert.block_number);
                }
                catch (e) {
                    throw Error('Certification based on an unexisting block');
                }
                try {
                    const issuer = yield dal.getWrittenIdtyByPubkey(cert.from);
                    if (!issuer || !issuer.member) {
                        throw Error('Issuer is not a member');
                    }
                }
                catch (e) {
                    throw Error('Certifier must be a member');
                }
            }
            let idty = yield findIdtyFunc(block, cert.to, dal);
            let current = block.number == 0 ? null : yield dal.getCurrentBlockOrNull();
            if (!idty) {
                throw Error('Identity does not exist for certified');
            }
            else if (current && current.medianTime > basedBlock.medianTime + conf.sigValidity) {
                throw Error('Certification has expired');
            }
            else if (cert.from == idty.pubkey)
                throw Error('Rejected certification: certifying its own self-certification has no meaning');
            else {
                const buid = [cert.block_number, basedBlock.hash].join('-');
                if (cert.block_hash && buid != [cert.block_number, cert.block_hash].join('-'))
                    throw Error('Certification based on an unexisting block buid. from ' + cert.from.substring(0, 8) + ' to ' + idty.pubkey.substring(0, 8));
                idty.currency = conf.currency;
                const raw = index_1.rawer.getOfficialCertification(_.extend(idty, {
                    idty_issuer: idty.pubkey,
                    idty_uid: idty.uid,
                    idty_buid: idty.buid,
                    idty_sig: idty.sig,
                    issuer: cert.from,
                    buid: buid,
                    sig: ''
                }));
                const verified = keyring_1.verify(raw, cert.sig, cert.from);
                if (!verified) {
                    throw constants.ERRORS.WRONG_SIGNATURE_FOR_CERT;
                }
                return true;
            }
        }
    });
}
function setLogger(newLogger) {
    logger = newLogger;
}
exports.setLogger = setLogger;
//# sourceMappingURL=global_rules.js.map