"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const BlockDTO_1 = require("../dto/BlockDTO");
const indexer_1 = require("../indexer");
const TransactionDTO_1 = require("../dto/TransactionDTO");
const keyring_1 = require("../common-libs/crypto/keyring");
const common_1 = require("../common");
const constants_1 = require("../common-libs/constants");
const IdentityDTO_1 = require("../dto/IdentityDTO");
const MembershipDTO_1 = require("../dto/MembershipDTO");
const _ = require('underscore');
const constants = constants_1.CommonConstants;
const maxAcceleration = require('./helpers').maxAcceleration;
exports.LOCAL_RULES_FUNCTIONS = {
    checkParameters: (block) => __awaiter(this, void 0, void 0, function* () {
        if (block.number == 0 && !block.parameters) {
            throw Error('Parameters must be provided for root block');
        }
        else if (block.number > 0 && block.parameters) {
            throw Error('Parameters must not be provided for non-root block');
        }
        return true;
    }),
    isProofOfWorkCorrect: (block) => {
        let remainder = block.powMin % 16;
        let nb_zeros = (block.powMin - remainder) / 16;
        const powRegexp = new RegExp('^0{' + nb_zeros + '}');
        return !!block.hash.match(powRegexp);
    },
    checkProofOfWork: (block) => __awaiter(this, void 0, void 0, function* () {
        if (!exports.LOCAL_RULES_FUNCTIONS.isProofOfWorkCorrect(block)) {
            throw Error('Not a proof-of-work');
        }
        return true;
    }),
    checkInnerHash: (block) => __awaiter(this, void 0, void 0, function* () {
        let inner_hash = common_1.hashf(block.getRawInnerPart()).toUpperCase();
        if (block.inner_hash != inner_hash) {
            throw Error('Wrong inner hash');
        }
        return true;
    }),
    checkPreviousHash: (block) => __awaiter(this, void 0, void 0, function* () {
        if (block.number == 0 && block.previousHash) {
            throw Error('PreviousHash must not be provided for root block');
        }
        else if (block.number > 0 && !block.previousHash) {
            throw Error('PreviousHash must be provided for non-root block');
        }
        return true;
    }),
    checkPreviousIssuer: (block) => __awaiter(this, void 0, void 0, function* () {
        if (block.number == 0 && block.previousIssuer)
            throw Error('PreviousIssuer must not be provided for root block');
        else if (block.number > 0 && !block.previousIssuer)
            throw Error('PreviousIssuer must be provided for non-root block');
        return true;
    }),
    checkUnitBase: (block) => __awaiter(this, void 0, void 0, function* () {
        if (block.number == 0 && block.unitbase != 0) {
            throw Error('UnitBase must equal 0 for root block');
        }
        return true;
    }),
    checkBlockSignature: (block) => __awaiter(this, void 0, void 0, function* () {
        if (!keyring_1.verify(block.getSignedPart(), block.signature, block.issuer))
            throw Error('Block\'s signature must match');
        return true;
    }),
    checkBlockTimes: (block, conf) => __awaiter(this, void 0, void 0, function* () {
        const time = block.time;
        const medianTime = block.medianTime;
        if (block.number > 0 && (time < medianTime || time > medianTime + maxAcceleration(conf)))
            throw Error('A block must have its Time between MedianTime and MedianTime + ' + maxAcceleration(conf));
        else if (block.number == 0 && time != medianTime)
            throw Error('Root block must have Time equal MedianTime');
        return true;
    }),
    checkIdentitiesSignature: (block) => __awaiter(this, void 0, void 0, function* () {
        let i = 0;
        let wrongSig = false;
        while (!wrongSig && i < block.identities.length) {
            const idty = IdentityDTO_1.IdentityDTO.fromInline(block.identities[i]);
            idty.currency = block.currency;
            wrongSig = !keyring_1.verify(idty.rawWithoutSig(), idty.sig, idty.pubkey);
            if (wrongSig) {
                throw Error('Identity\'s signature must match');
            }
            i++;
        }
        return true;
    }),
    checkIdentitiesUserIDConflict: (block, conf, index) => __awaiter(this, void 0, void 0, function* () {
        const creates = indexer_1.Indexer.iindexCreate(index);
        const uids = _.chain(creates).pluck('uid').uniq().value();
        if (creates.length !== uids.length) {
            throw Error('Block must not contain twice same identity uid');
        }
        return true;
    }),
    checkIdentitiesPubkeyConflict: (block, conf, index) => __awaiter(this, void 0, void 0, function* () {
        const creates = indexer_1.Indexer.iindexCreate(index);
        const pubkeys = _.chain(creates).pluck('pub').uniq().value();
        if (creates.length !== pubkeys.length) {
            throw Error('Block must not contain twice same identity pubkey');
        }
        return true;
    }),
    checkIdentitiesMatchJoin: (block, conf, index) => __awaiter(this, void 0, void 0, function* () {
        const icreates = indexer_1.Indexer.iindexCreate(index);
        const mcreates = indexer_1.Indexer.mindexCreate(index);
        for (const icreate of icreates) {
            const matching = _(mcreates).filter({ pub: icreate.pub });
            if (matching.length == 0) {
                throw Error('Each identity must match a newcomer line with same userid and certts');
            }
        }
        return true;
    }),
    checkRevokedAreExcluded: (block, conf, index) => __awaiter(this, void 0, void 0, function* () {
        const iindex = indexer_1.Indexer.iindex(index);
        const mindex = indexer_1.Indexer.mindex(index);
        const revocations = _.chain(mindex)
            .filter((row) => row.op == constants.IDX_UPDATE && row.revoked_on !== null)
            .pluck('pub')
            .value();
        for (const pub of revocations) {
            const exclusions = _(iindex).where({ op: constants.IDX_UPDATE, member: false, pub });
            if (exclusions.length == 0) {
                throw Error('A revoked member must be excluded');
            }
        }
        return true;
    }),
    checkRevokedUnicity: (block, conf, index) => __awaiter(this, void 0, void 0, function* () {
        try {
            yield exports.LOCAL_RULES_FUNCTIONS.checkMembershipUnicity(block, conf, index);
        }
        catch (e) {
            throw Error('A single revocation per member is allowed');
        }
        return true;
    }),
    checkMembershipUnicity: (block, conf, index) => __awaiter(this, void 0, void 0, function* () {
        const mindex = indexer_1.Indexer.mindex(index);
        const pubkeys = _.chain(mindex).pluck('pub').uniq().value();
        if (pubkeys.length !== mindex.length) {
            throw Error('Unicity constraint PUBLIC_KEY on MINDEX is not respected');
        }
        return true;
    }),
    checkMembershipsSignature: (block) => __awaiter(this, void 0, void 0, function* () {
        let i = 0;
        let wrongSig = false, ms;
        // Joiners
        while (!wrongSig && i < block.joiners.length) {
            ms = MembershipDTO_1.MembershipDTO.fromInline(block.joiners[i], 'IN', block.currency);
            wrongSig = !checkSingleMembershipSignature(ms);
            i++;
        }
        // Actives
        i = 0;
        while (!wrongSig && i < block.actives.length) {
            ms = MembershipDTO_1.MembershipDTO.fromInline(block.actives[i], 'IN', block.currency);
            wrongSig = !checkSingleMembershipSignature(ms);
            i++;
        }
        // Leavers
        i = 0;
        while (!wrongSig && i < block.leavers.length) {
            ms = MembershipDTO_1.MembershipDTO.fromInline(block.leavers[i], 'OUT', block.currency);
            wrongSig = !checkSingleMembershipSignature(ms);
            i++;
        }
        if (wrongSig) {
            throw Error('Membership\'s signature must match');
        }
        return true;
    }),
    checkPubkeyUnicity: (block) => __awaiter(this, void 0, void 0, function* () {
        const pubkeys = [];
        let conflict = false;
        let pubk;
        // Joiners
        let i = 0;
        while (!conflict && i < block.joiners.length) {
            pubk = block.joiners[i].split(':')[0];
            conflict = !!(~pubkeys.indexOf(pubk));
            pubkeys.push(pubk);
            i++;
        }
        // Actives
        i = 0;
        while (!conflict && i < block.actives.length) {
            pubk = block.actives[i].split(':')[0];
            conflict = !!(~pubkeys.indexOf(pubk));
            pubkeys.push(pubk);
            i++;
        }
        // Leavers
        i = 0;
        while (!conflict && i < block.leavers.length) {
            pubk = block.leavers[i].split(':')[0];
            conflict = !!(~pubkeys.indexOf(pubk));
            pubkeys.push(pubk);
            i++;
        }
        // Excluded
        i = 0;
        while (!conflict && i < block.excluded.length) {
            pubk = block.excluded[i].split(':')[0];
            conflict = !!(~pubkeys.indexOf(pubk));
            pubkeys.push(pubk);
            i++;
        }
        if (conflict) {
            throw Error('Block cannot contain a same pubkey more than once in joiners, actives, leavers and excluded');
        }
        return true;
    }),
    checkCertificationOneByIssuer: (block, conf, index) => __awaiter(this, void 0, void 0, function* () {
        if (block.number > 0) {
            const cindex = indexer_1.Indexer.cindex(index);
            const certFromA = _.uniq(cindex.map((row) => row.issuer));
            if (certFromA.length !== cindex.length) {
                throw Error('Block cannot contain two certifications from same issuer');
            }
        }
        return true;
    }),
    checkCertificationUnicity: (block, conf, index) => __awaiter(this, void 0, void 0, function* () {
        const cindex = indexer_1.Indexer.cindex(index);
        const certAtoB = _.uniq(cindex.map((row) => row.issuer + row.receiver));
        if (certAtoB.length !== cindex.length) {
            throw Error('Block cannot contain identical certifications (A -> B)');
        }
        return true;
    }),
    checkCertificationIsntForLeaverOrExcluded: (block, conf, index) => __awaiter(this, void 0, void 0, function* () {
        const cindex = indexer_1.Indexer.cindex(index);
        const iindex = indexer_1.Indexer.iindex(index);
        const mindex = indexer_1.Indexer.mindex(index);
        const certified = cindex.map((row) => row.receiver);
        for (const pub of certified) {
            const exclusions = _(iindex).where({ op: constants.IDX_UPDATE, member: false, pub: pub });
            const leavers = _(mindex).where({ op: constants.IDX_UPDATE, leaving: true, pub: pub });
            if (exclusions.length > 0 || leavers.length > 0) {
                throw Error('Block cannot contain certifications concerning leavers or excluded members');
            }
        }
        return true;
    }),
    checkTxVersion: (block) => __awaiter(this, void 0, void 0, function* () {
        const txs = block.transactions;
        // Check rule against each transaction
        for (const tx of txs) {
            if (tx.version != 10) {
                throw Error('A transaction must have the version 10');
            }
        }
        return true;
    }),
    checkTxLen: (block) => __awaiter(this, void 0, void 0, function* () {
        const txs = block.transactions;
        // Check rule against each transaction
        for (const tx of txs) {
            const txLen = TransactionDTO_1.TransactionDTO.fromJSONObject(tx).getLen();
            if (txLen > constants.MAXIMUM_LEN_OF_COMPACT_TX) {
                throw constants.ERRORS.A_TRANSACTION_HAS_A_MAX_SIZE;
            }
        }
        // Check rule against each output of each transaction
        for (const tx of txs) {
            for (const output of tx.outputs) {
                const out = typeof output === 'string' ? output : TransactionDTO_1.TransactionDTO.outputObj2Str(output);
                if (out.length > constants.MAXIMUM_LEN_OF_OUTPUT) {
                    throw constants.ERRORS.MAXIMUM_LEN_OF_OUTPUT;
                }
            }
        }
        // Check rule against each unlock of each transaction
        for (const tx of txs) {
            for (const unlock of tx.unlocks) {
                if (unlock.length > constants.MAXIMUM_LEN_OF_UNLOCK) {
                    throw constants.ERRORS.MAXIMUM_LEN_OF_UNLOCK;
                }
            }
        }
        return true;
    }),
    checkTxIssuers: (block) => __awaiter(this, void 0, void 0, function* () {
        const txs = block.transactions;
        // Check rule against each transaction
        for (const tx of txs) {
            if (tx.issuers.length == 0) {
                throw Error('A transaction must have at least 1 issuer');
            }
        }
        return true;
    }),
    checkTxSources: (block) => __awaiter(this, void 0, void 0, function* () {
        const dto = BlockDTO_1.BlockDTO.fromJSONObject(block);
        for (const tx of dto.transactions) {
            if (!tx.inputs || tx.inputs.length == 0) {
                throw Error('A transaction must have at least 1 source');
            }
        }
        const sindex = indexer_1.Indexer.localSIndex(dto);
        const inputs = _.filter(sindex, (row) => row.op == constants.IDX_UPDATE).map((row) => [row.op, row.identifier, row.pos].join('-'));
        if (inputs.length !== _.uniq(inputs).length) {
            throw Error('It cannot exist 2 identical sources for transactions inside a given block');
        }
        const outputs = _.filter(sindex, (row) => row.op == constants.IDX_CREATE).map((row) => [row.op, row.identifier, row.pos].join('-'));
        if (outputs.length !== _.uniq(outputs).length) {
            throw Error('It cannot exist 2 identical sources for transactions inside a given block');
        }
        return true;
    }),
    checkTxAmounts: (block) => __awaiter(this, void 0, void 0, function* () {
        for (const tx of block.transactions) {
            exports.LOCAL_RULES_HELPERS.checkTxAmountsValidity(tx);
        }
    }),
    checkTxRecipients: (block) => __awaiter(this, void 0, void 0, function* () {
        const txs = block.transactions;
        // Check rule against each transaction
        for (const tx of txs) {
            if (!tx.outputs || tx.outputs.length == 0) {
                throw Error('A transaction must have at least 1 recipient');
            }
            else {
                // Cannot have empty output condition
                for (const output of tx.outputsAsObjects()) {
                    if (!output.conditions.match(/(SIG|XHX)/)) {
                        throw Error('Empty conditions are forbidden');
                    }
                }
            }
        }
        return true;
    }),
    checkTxSignature: (block) => __awaiter(this, void 0, void 0, function* () {
        const txs = block.transactions;
        // Check rule against each transaction
        for (const tx of txs) {
            if (!tx.checkSignatures()) {
                throw Error('Signature from a transaction must match');
            }
        }
        return true;
    })
};
function checkSingleMembershipSignature(ms) {
    return keyring_1.verify(ms.getRaw(), ms.signature, ms.issuer);
}
function checkBunchOfTransactions(transactions, done = undefined) {
    const block = { transactions };
    return (() => __awaiter(this, void 0, void 0, function* () {
        try {
            let local_rule = exports.LOCAL_RULES_FUNCTIONS;
            yield local_rule.checkTxLen(block);
            yield local_rule.checkTxIssuers(block);
            yield local_rule.checkTxSources(block);
            yield local_rule.checkTxRecipients(block);
            yield local_rule.checkTxAmounts(block);
            yield local_rule.checkTxSignature(block);
            done && done();
        }
        catch (err) {
            if (done)
                return done(err);
            throw err;
        }
    }))();
}
exports.LOCAL_RULES_HELPERS = {
    maxAcceleration: (conf) => maxAcceleration(conf),
    checkSingleMembershipSignature: checkSingleMembershipSignature,
    checkBunchOfTransactions: checkBunchOfTransactions,
    checkSingleTransactionLocally: (tx, done = undefined) => checkBunchOfTransactions([tx], done),
    checkTxAmountsValidity: (tx) => {
        const inputs = tx.inputsAsObjects();
        const outputs = tx.outputsAsObjects();
        // Rule of money conservation
        const commonBase = inputs.concat(outputs).reduce((min, input) => {
            if (min === null)
                return input.base;
            return Math.min(min, input.base);
        }, 0);
        const inputSumCommonBase = inputs.reduce((sum, input) => {
            return sum + input.amount * Math.pow(10, input.base - commonBase);
        }, 0);
        const outputSumCommonBase = outputs.reduce((sum, output) => {
            return sum + output.amount * Math.pow(10, output.base - commonBase);
        }, 0);
        if (inputSumCommonBase !== outputSumCommonBase) {
            throw constants.ERRORS.TX_INPUTS_OUTPUTS_NOT_EQUAL;
        }
        // Rule of unit base transformation
        const maxOutputBase = outputs.reduce((max, output) => {
            return Math.max(max, output.base);
        }, 0);
        // Compute deltas
        const deltas = {};
        for (let i = commonBase; i <= maxOutputBase; i++) {
            const inputBaseSum = inputs.reduce((sum, input) => {
                if (input.base == i) {
                    return sum + input.amount * Math.pow(10, input.base - commonBase);
                }
                else {
                    return sum;
                }
            }, 0);
            const outputBaseSum = outputs.reduce((sum, output) => {
                if (output.base == i) {
                    return sum + output.amount * Math.pow(10, output.base - commonBase);
                }
                else {
                    return sum;
                }
            }, 0);
            const delta = outputBaseSum - inputBaseSum;
            let sumUpToBase = 0;
            for (let j = commonBase; j < i; j++) {
                sumUpToBase -= deltas[j];
            }
            if (delta > 0 && delta > sumUpToBase) {
                throw constants.ERRORS.TX_OUTPUT_SUM_NOT_EQUALS_PREV_DELTAS;
            }
            deltas[i] = delta;
        }
    },
    getMaxPossibleVersionNumber: (current) => __awaiter(this, void 0, void 0, function* () {
        // Looking at current blockchain, find what is the next maximum version we can produce
        // 1. We follow previous block's version
        let version = current ? current.version : constants.BLOCK_GENERATED_VERSION;
        // 2. If we can, we go to the next version
        return version;
    })
};
//# sourceMappingURL=local_rules.js.map