"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const AbstractController_1 = require("./AbstractController");
const parameters_1 = require("../parameters");
const constants_1 = require("../constants");
const MembershipDTO_1 = require("../../../../lib/dto/MembershipDTO");
const dtos_1 = require("../dtos");
const _ = require('underscore');
const http2raw = require('../http2raw');
const toJson = require('../tojson');
class BlockchainBinding extends AbstractController_1.AbstractController {
    constructor(server) {
        super(server);
        this.parameters = () => __awaiter(this, void 0, void 0, function* () {
            const params = yield this.server.dal.getParameters();
            return {
                "currency": params.currency,
                "c": params.c,
                "dt": params.dt,
                "ud0": params.ud0,
                "sigPeriod": params.sigPeriod,
                "sigStock": params.sigStock,
                "sigWindow": params.sigWindow,
                "sigValidity": params.sigValidity,
                "sigQty": params.sigQty,
                "idtyWindow": params.idtyWindow,
                "msWindow": params.msWindow,
                "xpercent": params.xpercent,
                "msValidity": params.msValidity,
                "stepMax": params.stepMax,
                "medianTimeBlocks": params.medianTimeBlocks,
                "avgGenTime": params.avgGenTime,
                "dtDiffEval": params.dtDiffEval,
                "percentRot": params.percentRot,
                "udTime0": params.udTime0,
                "udReevalTime0": params.udReevalTime0,
                "dtReeval": params.dtReeval
            };
        });
        this.with = {
            newcomers: this.getStat('newcomers'),
            certs: this.getStat('certs'),
            joiners: this.getStat('joiners'),
            actives: this.getStat('actives'),
            leavers: this.getStat('leavers'),
            revoked: this.getStat('revoked'),
            excluded: this.getStat('excluded'),
            ud: this.getStat('ud'),
            tx: this.getStat('tx')
        };
    }
    parseMembership(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.pushEntity(req, http2raw.membership, (raw) => this.server.writeRawMembership(raw));
            return {
                signature: res.signature,
                membership: {
                    version: res.version,
                    currency: res.currency,
                    issuer: res.issuer,
                    membership: res.membership,
                    date: res.date || 0,
                    sigDate: res.sigDate || 0,
                    raw: res.getRaw()
                }
            };
        });
    }
    parseBlock(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.pushEntity(req, http2raw.block, (raw) => this.server.writeRawBlock(raw));
            return dtos_1.block2HttpBlock(res);
        });
    }
    getStat(statName) {
        return () => __awaiter(this, void 0, void 0, function* () {
            let stat = yield this.server.dal.getStat(statName);
            return { result: toJson.stat(stat) };
        });
    }
    promoted(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const number = yield parameters_1.ParametersService.getNumberP(req);
            const promoted = yield this.BlockchainService.promoted(number);
            return toJson.block(promoted);
        });
    }
    blocks(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const params = parameters_1.ParametersService.getCountAndFrom(req);
            const count = parseInt(params.count);
            const from = parseInt(params.from);
            let blocks = yield this.BlockchainService.blocksBetween(from, count);
            blocks = blocks.map((b) => toJson.block(b));
            return blocks;
        });
    }
    current() {
        return __awaiter(this, void 0, void 0, function* () {
            const current = yield this.server.dal.getCurrentBlockOrNull();
            if (!current)
                throw constants_1.BMAConstants.ERRORS.NO_CURRENT_BLOCK;
            return toJson.block(current);
        });
    }
    hardship(req) {
        return __awaiter(this, void 0, void 0, function* () {
            let nextBlockNumber = 0;
            const search = yield parameters_1.ParametersService.getSearchP(req);
            const idty = yield this.IdentityService.findMemberWithoutMemberships(search);
            if (!idty) {
                throw constants_1.BMAConstants.ERRORS.NO_MATCHING_IDENTITY;
            }
            if (!idty.member) {
                throw constants_1.BMAConstants.ERRORS.NOT_A_MEMBER;
            }
            const current = yield this.BlockchainService.current();
            if (current) {
                nextBlockNumber = current ? current.number + 1 : 0;
            }
            const difficulty = yield this.server.getBcContext().getIssuerPersonalizedDifficulty(idty.pubkey);
            return {
                "block": nextBlockNumber,
                "level": difficulty
            };
        });
    }
    difficulties() {
        return __awaiter(this, void 0, void 0, function* () {
            const current = yield this.server.dal.getCurrentBlockOrNull();
            const number = (current && current.number) || 0;
            const issuers = yield this.server.dal.getUniqueIssuersBetween(number - 1 - current.issuersFrame, number - 1);
            const difficulties = [];
            for (const issuer of issuers) {
                const member = yield this.server.dal.getWrittenIdtyByPubkey(issuer);
                const difficulty = yield this.server.getBcContext().getIssuerPersonalizedDifficulty(member.pubkey);
                difficulties.push({
                    uid: member.uid,
                    level: difficulty
                });
            }
            return {
                "block": number + 1,
                "levels": _.sortBy(difficulties, (diff) => diff.level)
            };
        });
    }
    memberships(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const search = yield parameters_1.ParametersService.getSearchP(req);
            const idty = yield this.IdentityService.findMember(search);
            const json = {
                pubkey: idty.pubkey,
                uid: idty.uid,
                sigDate: idty.buid,
                memberships: idty.memberships.map((msObj) => {
                    const ms = MembershipDTO_1.MembershipDTO.fromJSONObject(msObj);
                    return {
                        version: ms.version,
                        currency: this.conf.currency,
                        membership: ms.membership,
                        blockNumber: ms.block_number,
                        blockHash: ms.block_hash,
                        written: (!msObj.written_number && msObj.written_number !== 0) ? null : msObj.written_number
                    };
                })
            };
            json.memberships = _.sortBy(json.memberships, 'blockNumber');
            json.memberships.reverse();
            return json;
        });
    }
    branches() {
        return __awaiter(this, void 0, void 0, function* () {
            const branches = yield this.BlockchainService.branches();
            const blocks = branches.map((b) => toJson.block(b));
            return {
                blocks: blocks
            };
        });
    }
}
exports.BlockchainBinding = BlockchainBinding;
//# sourceMappingURL=blockchain.js.map