"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const AbstractController_1 = require("./AbstractController");
const parameters_1 = require("../parameters");
const source_1 = require("../entity/source");
const constants_1 = require("../constants");
const TransactionDTO_1 = require("../../../../lib/dto/TransactionDTO");
const _ = require('underscore');
const http2raw = require('../http2raw');
class TransactionBinding extends AbstractController_1.AbstractController {
    parseTransaction(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.pushEntity(req, http2raw.transaction, (raw) => this.server.writeRawTransaction(raw));
            return {
                version: res.version,
                currency: res.currency,
                issuers: res.issuers,
                inputs: res.inputs,
                outputs: res.outputs,
                unlocks: res.unlocks,
                signatures: res.signatures,
                comment: res.comment,
                locktime: res.locktime,
                hash: res.hash,
                written_block: res.blockNumber,
                raw: res.getRaw()
            };
        });
    }
    getSources(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const pubkey = yield parameters_1.ParametersService.getPubkeyP(req);
            const sources = yield this.server.dal.getAvailableSourcesByPubkey(pubkey);
            const result = {
                "currency": this.conf.currency,
                "pubkey": pubkey,
                "sources": []
            };
            sources.forEach(function (src) {
                result.sources.push(new source_1.Source(src).json());
            });
            return result;
        });
    }
    getByHash(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const hash = parameters_1.ParametersService.getHash(req);
            const tx = yield this.server.dal.getTxByHash(hash);
            if (!tx) {
                throw constants_1.BMAConstants.ERRORS.TX_NOT_FOUND;
            }
            tx.inputs = tx.inputs.map((i) => i.raw || i);
            tx.outputs = tx.outputs.map((o) => o.raw || o);
            return {
                version: tx.version,
                currency: tx.currency,
                locktime: tx.locktime,
                // blockstamp: tx.blockstamp,
                // blockstampTime: tx.blockstampTime,
                issuers: tx.issuers,
                inputs: tx.inputs,
                outputs: tx.outputs,
                unlocks: tx.unlocks,
                signatures: tx.signatures,
                comment: tx.comment,
                hash: tx.hash,
                // time: tx.time,
                // block_number: tx.block_number,
                written_block: tx.block_number,
                // received: tx.received,
                raw: ""
            };
        });
    }
    getHistory(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const pubkey = yield parameters_1.ParametersService.getPubkeyP(req);
            return this.getFilteredHistory(pubkey, (results) => results);
        });
    }
    getHistoryBetweenBlocks(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const pubkey = yield parameters_1.ParametersService.getPubkeyP(req);
            const from = yield parameters_1.ParametersService.getFromP(req);
            const to = yield parameters_1.ParametersService.getToP(req);
            return this.getFilteredHistory(pubkey, (res) => {
                const histo = res.history;
                histo.sent = _.filter(histo.sent, function (tx) { return tx && tx.block_number >= from && tx.block_number <= to; });
                histo.received = _.filter(histo.received, function (tx) { return tx && tx.block_number >= from && tx.block_number <= to; });
                _.extend(histo, { sending: [], receiving: [] });
                return res;
            });
        });
    }
    getHistoryBetweenTimes(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const pubkey = yield parameters_1.ParametersService.getPubkeyP(req);
            const from = yield parameters_1.ParametersService.getFromP(req);
            const to = yield parameters_1.ParametersService.getToP(req);
            return this.getFilteredHistory(pubkey, (res) => {
                const histo = res.history;
                histo.sent = _.filter(histo.sent, function (tx) { return tx && tx.time >= from && tx.time <= to; });
                histo.received = _.filter(histo.received, function (tx) { return tx && tx.time >= from && tx.time <= to; });
                _.extend(histo, { sending: [], receiving: [] });
                return res;
            });
        });
    }
    getPendingForPubkey(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const pubkey = yield parameters_1.ParametersService.getPubkeyP(req);
            return this.getFilteredHistory(pubkey, function (res) {
                const histo = res.history;
                _.extend(histo, { sent: [], received: [] });
                return res;
            });
        });
    }
    getPending() {
        return __awaiter(this, void 0, void 0, function* () {
            const pending = yield this.server.dal.getTransactionsPending();
            const res = {
                "currency": this.conf.currency,
                "pending": pending
            };
            pending.map(function (tx, index) {
                pending[index] = _.omit(TransactionDTO_1.TransactionDTO.fromJSONObject(tx).json(), 'currency', 'raw');
            });
            return res;
        });
    }
    getFilteredHistory(pubkey, filter) {
        return __awaiter(this, void 0, void 0, function* () {
            let history = yield this.server.dal.getTransactionsHistory(pubkey);
            let result = {
                "currency": this.conf.currency,
                "pubkey": pubkey,
                "history": {
                    sending: history.sending.map(dbtx2HttpTxOfHistory),
                    received: history.received.map(dbtx2HttpTxOfHistory),
                    receiving: history.receiving.map(dbtx2HttpTxOfHistory),
                    sent: history.sent.map(dbtx2HttpTxOfHistory),
                    pending: history.pending.map(dbtx2HttpTxOfHistory)
                }
            };
            return filter(result);
        });
    }
}
exports.TransactionBinding = TransactionBinding;
function dbtx2HttpTxOfHistory(tx) {
    return {
        version: tx.version,
        locktime: tx.locktime,
        blockstamp: tx.blockstamp,
        blockstampTime: tx.blockstampTime,
        issuers: tx.issuers,
        inputs: tx.inputs,
        outputs: tx.outputs,
        unlocks: tx.unlocks,
        signatures: tx.signatures,
        comment: tx.comment,
        hash: tx.hash,
        time: tx.time,
        block_number: tx.block_number,
        received: tx.received
    };
}
//# sourceMappingURL=transactions.js.map