"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const AbstractController_1 = require("./AbstractController");
const constants_1 = require("../constants");
const IdentityDAL_1 = require("../../../../lib/dal/sqliteDAL/IdentityDAL");
const _ = require('underscore');
const http2raw = require('../http2raw');
const ParametersService = require('../parameters').ParametersService;
class WOTBinding extends AbstractController_1.AbstractController {
    lookup(req) {
        return __awaiter(this, void 0, void 0, function* () {
            // Get the search parameter from HTTP query
            const search = yield ParametersService.getSearchP(req);
            // Make the research
            const identities = yield this.IdentityService.searchIdentities(search);
            // Entitify each result
            identities.forEach((idty, index) => identities[index] = IdentityDAL_1.DBIdentity.copyFromExisting(idty));
            // Prepare some data to avoid displaying expired certifications
            for (const idty of identities) {
                const certs = yield this.server.dal.certsToTarget(idty.pubkey, idty.getTargetHash());
                const validCerts = [];
                for (const cert of certs) {
                    const member = yield this.IdentityService.getWrittenByPubkey(cert.from);
                    if (member) {
                        cert.uids = [member.uid];
                        cert.isMember = member.member;
                        cert.wasMember = member.wasMember;
                    }
                    else {
                        const potentials = yield this.IdentityService.getPendingFromPubkey(cert.from);
                        cert.uids = _(potentials).pluck('uid');
                        cert.isMember = false;
                        cert.wasMember = false;
                    }
                    validCerts.push(cert);
                }
                idty.certs = validCerts;
                const signed = yield this.server.dal.certsFrom(idty.pubkey);
                const validSigned = [];
                for (let j = 0; j < signed.length; j++) {
                    const cert = _.clone(signed[j]);
                    cert.idty = yield this.server.dal.getIdentityByHashOrNull(cert.target);
                    if (cert.idty) {
                        validSigned.push(cert);
                    }
                    else {
                        this.logger.debug('A certification to an unknown identity was found (%s => %s)', cert.from, cert.to);
                    }
                }
                idty.signed = validSigned;
            }
            if (identities.length == 0) {
                throw constants_1.BMAConstants.ERRORS.NO_MATCHING_IDENTITY;
            }
            const resultsByPubkey = {};
            identities.forEach((identity) => {
                const copy = IdentityDAL_1.DBIdentity.copyFromExisting(identity);
                const jsoned = copy.json();
                if (!resultsByPubkey[jsoned.pubkey]) {
                    // Create the first matching identity with this pubkey in the map
                    resultsByPubkey[jsoned.pubkey] = jsoned;
                }
                else {
                    // Merge the identity with the existing(s)
                    const existing = resultsByPubkey[jsoned.pubkey];
                    // We add the UID of the identity to the list of already added UIDs
                    existing.uids = existing.uids.concat(jsoned.uids);
                    // We do not merge the `signed`: every identity with the same pubkey has the same `signed` because it the *pubkey* which signs, not the identity
                }
            });
            return {
                partial: false,
                results: _.values(resultsByPubkey)
            };
        });
    }
    members() {
        return __awaiter(this, void 0, void 0, function* () {
            const identities = yield this.server.dal.getMembers();
            const json = {
                results: []
            };
            identities.forEach((identity) => json.results.push({ pubkey: identity.pubkey, uid: identity.uid }));
            return json;
        });
    }
    certifiersOf(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const search = yield ParametersService.getSearchP(req);
            const idty = yield this.IdentityService.findMemberWithoutMemberships(search);
            const certs = yield this.server.dal.certsToTarget(idty.pubkey, idty.getTargetHash());
            idty.certs = [];
            for (const cert of certs) {
                const certifier = yield this.server.dal.getWrittenIdtyByPubkey(cert.from);
                if (certifier) {
                    cert.uid = certifier.uid;
                    cert.isMember = certifier.member;
                    cert.sigDate = certifier.buid;
                    cert.wasMember = true; // As we checked if(certified)
                    if (!cert.cert_time) {
                        let certBlock = yield this.server.dal.getBlock(cert.block_number);
                        cert.cert_time = {
                            block: certBlock.number,
                            medianTime: certBlock.medianTime
                        };
                    }
                    idty.certs.push(cert);
                }
            }
            const json = {
                pubkey: idty.pubkey,
                uid: idty.uid,
                sigDate: idty.buid,
                isMember: idty.member,
                certifications: []
            };
            idty.certs.forEach(function (cert) {
                json.certifications.push({
                    pubkey: cert.from,
                    uid: cert.uid,
                    isMember: cert.isMember,
                    wasMember: cert.wasMember,
                    cert_time: cert.cert_time,
                    sigDate: cert.sigDate,
                    written: cert.linked ? {
                        number: cert.written_block,
                        hash: cert.written_hash
                    } : null,
                    signature: cert.sig
                });
            });
            return json;
        });
    }
    requirements(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const search = yield ParametersService.getSearchP(req);
            const identities = yield this.IdentityService.searchIdentities(search);
            const all = yield this.BlockchainService.requirementsOfIdentities(identities);
            if (!all || !all.length) {
                throw constants_1.BMAConstants.ERRORS.NO_IDTY_MATCHING_PUB_OR_UID;
            }
            return {
                identities: all
            };
        });
    }
    requirementsOfPending(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const minsig = ParametersService.getMinSig(req);
            let identities = yield this.server.dal.idtyDAL.query('SELECT i.*, count(c.sig) as nbSig ' +
                'FROM idty i, cert c ' +
                'WHERE c.target = i.hash group by i.hash having nbSig >= ?', minsig);
            const members = (yield this.server.dal.idtyDAL.query('SELECT i.*, count(c.sig) as nbSig ' +
                'FROM i_index i, cert c ' +
                'WHERE c.`to` = i.pub group by i.pub having nbSig >= ?', minsig)).map((i) => {
                return {
                    hash: i.hash || "",
                    member: i.member || false,
                    wasMember: i.wasMember || false,
                    pubkey: i.pub,
                    uid: i.uid || "",
                    buid: i.created_on || "",
                    sig: i.sig || "",
                    revocation_sig: "",
                    revoked: false,
                    revoked_on: 0
                };
            });
            identities = identities.concat(members);
            const all = yield this.BlockchainService.requirementsOfIdentities(identities, false);
            if (!all || !all.length) {
                throw constants_1.BMAConstants.ERRORS.NO_IDTY_MATCHING_PUB_OR_UID;
            }
            return {
                identities: all
            };
        });
    }
    certifiedBy(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const search = yield ParametersService.getSearchP(req);
            const idty = yield this.IdentityService.findMemberWithoutMemberships(search);
            const certs = yield this.server.dal.certsFrom(idty.pubkey);
            idty.certs = [];
            for (const cert of certs) {
                const certified = yield this.server.dal.getWrittenIdtyByPubkey(cert.to);
                if (certified) {
                    cert.uid = certified.uid;
                    cert.isMember = certified.member;
                    cert.sigDate = certified.buid;
                    cert.wasMember = true; // As we checked if(certified)
                    if (!cert.cert_time) {
                        let certBlock = yield this.server.dal.getBlock(cert.block_number);
                        cert.cert_time = {
                            block: certBlock.number,
                            medianTime: certBlock.medianTime
                        };
                    }
                    idty.certs.push(cert);
                }
            }
            const json = {
                pubkey: idty.pubkey,
                uid: idty.uid,
                sigDate: idty.buid,
                isMember: idty.member,
                certifications: []
            };
            idty.certs.forEach((cert) => json.certifications.push({
                pubkey: cert.to,
                uid: cert.uid,
                isMember: cert.isMember,
                wasMember: cert.wasMember,
                cert_time: cert.cert_time,
                sigDate: cert.sigDate,
                written: cert.linked ? {
                    number: cert.written_block,
                    hash: cert.written_hash
                } : null,
                signature: cert.sig
            }));
            return json;
        });
    }
    identityOf(req) {
        return __awaiter(this, void 0, void 0, function* () {
            let search = yield ParametersService.getSearchP(req);
            let idty = yield this.IdentityService.findMemberWithoutMemberships(search);
            if (!idty) {
                throw 'Identity not found';
            }
            if (!idty.member) {
                throw 'Not a member';
            }
            return {
                pubkey: idty.pubkey,
                uid: idty.uid,
                sigDate: idty.buid
            };
        });
    }
    add(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.pushEntity(req, http2raw.identity, (raw) => this.server.writeRawIdentity(raw));
            return {
                pubkey: res.pubkey,
                uids: [],
                signed: []
            };
        });
    }
    certify(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.pushEntity(req, http2raw.certification, (raw) => this.server.writeRawCertification(raw));
            const target = {
                issuer: res.idty_issuer,
                uid: res.idty_uid,
                timestamp: res.idty_buid,
                sig: res.idty_sig
            };
            return {
                issuer: res.issuer,
                timestamp: res.buid,
                sig: res.sig,
                target
            };
        });
    }
    revoke(req) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.pushEntity(req, http2raw.revocation, (raw) => this.server.writeRawRevocation(raw));
            return {
                result: true
            };
        });
    }
    pendingMemberships() {
        return __awaiter(this, void 0, void 0, function* () {
            const memberships = yield this.server.dal.findNewcomers();
            const json = {
                memberships: memberships.map((ms) => {
                    return {
                        pubkey: ms.issuer,
                        uid: ms.userid,
                        version: ms.version || 0,
                        currency: this.server.conf.currency,
                        membership: ms.membership,
                        blockNumber: parseInt(ms.blockNumber),
                        blockHash: ms.blockHash,
                        written: (!ms.written_number && ms.written_number !== 0) ? null : ms.written_number
                    };
                })
            };
            json.memberships = _.sortBy(json.memberships, 'blockNumber');
            json.memberships.reverse();
            return json;
        });
    }
}
exports.WOTBinding = WOTBinding;
//# sourceMappingURL=wot.js.map