"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const stream = require("stream");
const PeerDTO_1 = require("../../../lib/dto/PeerDTO");
const pulling_1 = require("./pulling");
const tx_cleaner_1 = require("./tx_cleaner");
const connect_1 = require("./connect");
const constants_1 = require("./constants");
const sandbox_1 = require("./sandbox");
const garbager_1 = require("./garbager");
const _ = require('underscore');
const async = require('async');
const querablep = require('querablep');
/**
 * Service which triggers the server's peering generation (actualization of the Peer document).
 * @constructor
 */
class Crawler extends stream.Transform {
    constructor(server, conf, logger) {
        super({ objectMode: true });
        this.server = server;
        this.conf = conf;
        this.logger = logger;
        this.peerCrawler = new PeerCrawler(server, conf, logger);
        this.peerTester = new PeerTester(server, conf, logger);
        this.blockCrawler = new BlockCrawler(server, logger);
        this.sandboxCrawler = new SandboxCrawler(server, conf, logger);
    }
    pullBlocks(server, pubkey) {
        return this.blockCrawler.pullBlocks(server, pubkey);
    }
    sandboxPull(server) {
        return this.sandboxCrawler.sandboxPull(server);
    }
    startService() {
        if (this.conf.nobma || !this.conf.bmaWithCrawler) {
            return Promise.resolve();
        }
        return Promise.all([
            this.peerCrawler.startService(),
            this.peerTester.startService(),
            this.blockCrawler.startService(),
            this.sandboxCrawler.startService()
        ]);
    }
    stopService() {
        if (this.conf.nobma || !this.conf.bmaWithCrawler) {
            return Promise.resolve();
        }
        return Promise.all([
            this.peerCrawler.stopService(),
            this.peerTester.stopService(),
            this.blockCrawler.stopService(),
            this.sandboxCrawler.stopService()
        ]);
    }
    // Unused
    _write(str, enc, done) {
        done && done();
    }
    ;
}
exports.Crawler = Crawler;
class PeerCrawler {
    constructor(server, conf, logger) {
        this.server = server;
        this.conf = conf;
        this.logger = logger;
        this.DONT_IF_MORE_THAN_FOUR_PEERS = true;
        this.crawlPeersFifo = async.queue((task, callback) => task(callback), 1);
    }
    startService() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.crawlPeersInterval)
                clearInterval(this.crawlPeersInterval);
            this.crawlPeersInterval = setInterval(() => this.crawlPeersFifo.push((cb) => this.crawlPeers(this.server, this.conf).then(cb).catch(cb)), 1000 * this.conf.avgGenTime * constants_1.CrawlerConstants.SYNC_PEERS_INTERVAL);
            yield this.crawlPeers(this.server, this.conf, this.DONT_IF_MORE_THAN_FOUR_PEERS);
        });
    }
    stopService() {
        return __awaiter(this, void 0, void 0, function* () {
            this.crawlPeersFifo.kill();
            clearInterval(this.crawlPeersInterval);
        });
    }
    crawlPeers(server, conf, dontCrawlIfEnoughPeers = false) {
        return __awaiter(this, void 0, void 0, function* () {
            this.logger.info('Crawling the network...');
            const peers = yield server.dal.listAllPeersWithStatusNewUPWithtout(conf.pair.pub);
            if (peers.length > constants_1.CrawlerConstants.COUNT_FOR_ENOUGH_PEERS && dontCrawlIfEnoughPeers == this.DONT_IF_MORE_THAN_FOUR_PEERS) {
                return;
            }
            let peersToTest = peers.slice().map((p) => PeerDTO_1.PeerDTO.fromJSONObject(p));
            let tested = [];
            const found = [];
            while (peersToTest.length > 0) {
                const results = yield Promise.all(peersToTest.map((p) => this.crawlPeer(server, p)));
                tested = tested.concat(peersToTest.map((p) => p.pubkey));
                // End loop condition
                peersToTest.splice(0);
                // Eventually continue the loop
                for (let i = 0, len = results.length; i < len; i++) {
                    const res = results[i];
                    for (let j = 0, len2 = res.length; j < len2; j++) {
                        try {
                            const subpeer = res[j].leaf.value;
                            if (subpeer.currency && tested.indexOf(subpeer.pubkey) === -1) {
                                const p = PeerDTO_1.PeerDTO.fromJSONObject(subpeer);
                                peersToTest.push(p);
                                found.push(p);
                            }
                        }
                        catch (e) {
                            this.logger.warn('Invalid peer %s', res[j]);
                        }
                    }
                }
                // Make unique list
                peersToTest = _.uniq(peersToTest, false, (p) => p.pubkey);
            }
            this.logger.info('Crawling done.');
            for (let i = 0, len = found.length; i < len; i++) {
                let p = found[i];
                try {
                    // Try to write it
                    yield server.writePeer(p);
                }
                catch (e) {
                    // Silent error
                }
            }
            yield garbager_1.cleanLongDownPeers(server, Date.now());
        });
    }
    crawlPeer(server, aPeer) {
        return __awaiter(this, void 0, void 0, function* () {
            let subpeers = [];
            try {
                this.logger.debug('Crawling peers of %s %s', aPeer.pubkey.substr(0, 6), aPeer.getNamedURL());
                const node = yield connect_1.connect(aPeer);
                yield checkPeerValidity(server, aPeer, node);
                const json = yield node.getPeers.bind(node)({ leaves: true });
                for (let i = 0, len = json.leaves.length; i < len; i++) {
                    let leaf = json.leaves[i];
                    let subpeer = yield node.getPeers.bind(node)({ leaf: leaf });
                    subpeers.push(subpeer);
                }
                return subpeers;
            }
            catch (e) {
                return subpeers;
            }
        });
    }
}
exports.PeerCrawler = PeerCrawler;
class SandboxCrawler {
    constructor(server, conf, logger) {
        this.server = server;
        this.conf = conf;
        this.logger = logger;
        this.pullFifo = async.queue((task, callback) => task(callback), 1);
    }
    startService() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.pullInterval)
                clearInterval(this.pullInterval);
            this.pullInterval = setInterval(() => this.pullFifo.push((cb) => this.sandboxPull(this.server).then(cb).catch(cb)), 1000 * this.conf.avgGenTime * constants_1.CrawlerConstants.SANDBOX_CHECK_INTERVAL);
            setTimeout(() => {
                this.pullFifo.push((cb) => this.sandboxPull(this.server).then(cb).catch(cb));
            }, constants_1.CrawlerConstants.SANDBOX_FIRST_PULL_DELAY);
        });
    }
    stopService() {
        return __awaiter(this, void 0, void 0, function* () {
            this.pullFifo.kill();
            clearInterval(this.pullInterval);
        });
    }
    sandboxPull(server) {
        return __awaiter(this, void 0, void 0, function* () {
            this.logger && this.logger.info('Sandbox pulling started...');
            const peers = yield server.dal.getRandomlyUPsWithout([this.conf.pair.pub]);
            const randoms = chooseXin(peers, constants_1.CrawlerConstants.SANDBOX_PEERS_COUNT);
            let peersToTest = randoms.slice().map((p) => PeerDTO_1.PeerDTO.fromJSONObject(p));
            for (const peer of peersToTest) {
                const fromHost = yield connect_1.connect(peer);
                yield sandbox_1.pullSandboxToLocalServer(server.conf.currency, fromHost, server, this.logger);
            }
            this.logger && this.logger.info('Sandbox pulling done.');
        });
    }
}
exports.SandboxCrawler = SandboxCrawler;
class PeerTester {
    constructor(server, conf, logger) {
        this.server = server;
        this.conf = conf;
        this.logger = logger;
        this.FIRST_CALL = true;
        this.testPeerFifo = async.queue((task, callback) => task(callback), 1);
    }
    startService() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.testPeerFifoInterval)
                clearInterval(this.testPeerFifoInterval);
            this.testPeerFifoInterval = setInterval(() => this.testPeerFifo.push((cb) => this.testPeers.bind(this, this.server, this.conf, !this.FIRST_CALL)().then(cb).catch(cb)), 1000 * constants_1.CrawlerConstants.TEST_PEERS_INTERVAL);
            yield this.testPeers(this.server, this.conf, this.FIRST_CALL);
        });
    }
    stopService() {
        return __awaiter(this, void 0, void 0, function* () {
            clearInterval(this.testPeerFifoInterval);
            this.testPeerFifo.kill();
        });
    }
    testPeers(server, conf, displayDelays) {
        return __awaiter(this, void 0, void 0, function* () {
            let peers = yield server.dal.listAllPeers();
            let now = (new Date().getTime());
            peers = _.filter(peers, (p) => p.pubkey != conf.pair.pub);
            yield Promise.all(peers.map((thePeer) => __awaiter(this, void 0, void 0, function* () {
                let p = PeerDTO_1.PeerDTO.fromJSONObject(thePeer);
                if (thePeer.status == 'DOWN') {
                    let shouldDisplayDelays = displayDelays;
                    let downAt = thePeer.first_down || now;
                    let waitRemaining = this.getWaitRemaining(now, downAt, thePeer.last_try);
                    let nextWaitRemaining = this.getWaitRemaining(now, downAt, now);
                    let testIt = waitRemaining <= 0;
                    if (testIt) {
                        // We try to reconnect only with peers marked as DOWN
                        try {
                            this.logger.trace('Checking if node %s is UP... (%s:%s) ', p.pubkey.substr(0, 6), p.getHostPreferDNS(), p.getPort());
                            // We register the try anyway
                            yield server.dal.setPeerDown(p.pubkey);
                            // Now we test
                            let node = yield connect_1.connect(p);
                            let peering = yield node.getPeer();
                            yield checkPeerValidity(server, p, node);
                            // The node answered, it is no more DOWN!
                            this.logger.info('Node %s (%s:%s) is UP!', p.pubkey.substr(0, 6), p.getHostPreferDNS(), p.getPort());
                            yield server.dal.setPeerUP(p.pubkey);
                            // We try to forward its peering entry
                            let sp1 = peering.block.split('-');
                            let currentBlockNumber = sp1[0];
                            let currentBlockHash = sp1[1];
                            let sp2 = peering.block.split('-');
                            let blockNumber = sp2[0];
                            let blockHash = sp2[1];
                            if (!(currentBlockNumber == blockNumber && currentBlockHash == blockHash)) {
                                // The peering changed
                                yield server.PeeringService.submitP(peering);
                            }
                            // Do not need to display when next check will occur: the node is now UP
                            shouldDisplayDelays = false;
                        }
                        catch (err) {
                            if (!err) {
                                err = "NO_REASON";
                            }
                            // Error: we set the peer as DOWN
                            this.logger.trace("Peer %s is DOWN (%s)", p.pubkey, (err.httpCode && 'HTTP ' + err.httpCode) || err.code || err.message || err);
                            yield server.dal.setPeerDown(p.pubkey);
                            shouldDisplayDelays = true;
                        }
                    }
                    if (shouldDisplayDelays) {
                        this.logger.debug('Will check that node %s (%s:%s) is UP in %s min...', p.pubkey.substr(0, 6), p.getHostPreferDNS(), p.getPort(), (nextWaitRemaining / 60).toFixed(0));
                    }
                }
            })));
        });
    }
    getWaitRemaining(now, downAt, last_try) {
        let downDelay = Math.floor((now - downAt) / 1000);
        let waitedSinceLastTest = Math.floor((now - (last_try || now)) / 1000);
        let waitRemaining = 1;
        if (downDelay <= constants_1.CrawlerConstants.DURATIONS.A_MINUTE) {
            waitRemaining = constants_1.CrawlerConstants.DURATIONS.TEN_SECONDS - waitedSinceLastTest;
        }
        else if (downDelay <= constants_1.CrawlerConstants.DURATIONS.TEN_MINUTES) {
            waitRemaining = constants_1.CrawlerConstants.DURATIONS.A_MINUTE - waitedSinceLastTest;
        }
        else if (downDelay <= constants_1.CrawlerConstants.DURATIONS.AN_HOUR) {
            waitRemaining = constants_1.CrawlerConstants.DURATIONS.TEN_MINUTES - waitedSinceLastTest;
        }
        else if (downDelay <= constants_1.CrawlerConstants.DURATIONS.A_DAY) {
            waitRemaining = constants_1.CrawlerConstants.DURATIONS.AN_HOUR - waitedSinceLastTest;
        }
        else if (downDelay <= constants_1.CrawlerConstants.DURATIONS.A_WEEK) {
            waitRemaining = constants_1.CrawlerConstants.DURATIONS.A_DAY - waitedSinceLastTest;
        }
        else if (downDelay <= constants_1.CrawlerConstants.DURATIONS.A_MONTH) {
            waitRemaining = constants_1.CrawlerConstants.DURATIONS.A_WEEK - waitedSinceLastTest;
        }
        // Else do not check it, DOWN for too long
        return waitRemaining;
    }
}
exports.PeerTester = PeerTester;
class BlockCrawler {
    constructor(server, logger) {
        this.server = server;
        this.logger = logger;
        this.CONST_BLOCKS_CHUNK = constants_1.CrawlerConstants.CRAWL_BLOCK_CHUNK;
        this.pullingActualIntervalDuration = constants_1.CrawlerConstants.PULLING_MINIMAL_DELAY;
        this.programStart = Date.now();
        this.syncBlockFifo = async.queue((task, callback) => task(callback), 1);
    }
    startService() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.syncBlockInterval)
                clearInterval(this.syncBlockInterval);
            this.syncBlockInterval = setInterval(() => this.syncBlockFifo.push((cb) => this.syncBlock(this.server).then(cb).catch(cb)), 1000 * this.pullingActualIntervalDuration);
            this.syncBlock(this.server);
        });
    }
    stopService() {
        return __awaiter(this, void 0, void 0, function* () {
            clearInterval(this.syncBlockInterval);
            this.syncBlockFifo.kill();
        });
    }
    pullBlocks(server, pubkey) {
        return this.syncBlock(server, pubkey);
    }
    syncBlock(server, pubkey = "") {
        return __awaiter(this, void 0, void 0, function* () {
            // Eventually change the interval duration
            const minutesElapsed = Math.ceil((Date.now() - this.programStart) / (60 * 1000));
            const FACTOR = Math.sin((minutesElapsed / constants_1.CrawlerConstants.PULLING_INTERVAL_TARGET) * (Math.PI / 2));
            // Make the interval always higher than before
            const pullingTheoreticalIntervalNow = Math.max(Math.max(FACTOR * constants_1.CrawlerConstants.PULLING_INTERVAL_TARGET, constants_1.CrawlerConstants.PULLING_MINIMAL_DELAY), this.pullingActualIntervalDuration);
            if (pullingTheoreticalIntervalNow !== this.pullingActualIntervalDuration) {
                this.pullingActualIntervalDuration = pullingTheoreticalIntervalNow;
                // Change the interval
                if (this.syncBlockInterval)
                    clearInterval(this.syncBlockInterval);
                this.syncBlockInterval = setInterval(() => this.syncBlockFifo.push((cb) => this.syncBlock(server).then(cb).catch(cb)), 1000 * this.pullingActualIntervalDuration);
            }
            try {
                let current = yield server.dal.getCurrentBlockOrNull();
                if (current) {
                    this.pullingEvent(server, 'start', current.number);
                    this.logger && this.logger.info("Pulling blocks from the network...");
                    let peers = yield server.dal.findAllPeersNEWUPBut([server.conf.pair.pub]);
                    peers = _.shuffle(peers);
                    if (pubkey) {
                        _(peers).filter((p) => p.pubkey == pubkey);
                    }
                    // Shuffle the peers
                    peers = _.shuffle(peers);
                    // Only take at max X of them
                    peers = peers.slice(0, constants_1.CrawlerConstants.MAX_NUMBER_OF_PEERS_FOR_PULLING);
                    yield Promise.all(peers.map((thePeer, i) => __awaiter(this, void 0, void 0, function* () {
                        let p = PeerDTO_1.PeerDTO.fromJSONObject(thePeer);
                        this.pullingEvent(server, 'peer', _.extend({ number: i, length: peers.length }, p));
                        this.logger && this.logger.trace("Try with %s %s", p.getURL(), p.pubkey.substr(0, 6));
                        try {
                            let node = yield connect_1.connect(p);
                            let nodeCurrent = null;
                            node.pubkey = p.pubkey;
                            yield checkPeerValidity(server, p, node);
                            let dao = new (class extends pulling_1.AbstractDAO {
                                constructor(crawler) {
                                    super();
                                    this.crawler = crawler;
                                }
                                localCurrent() {
                                    return __awaiter(this, void 0, void 0, function* () {
                                        return server.dal.getCurrentBlockOrNull();
                                    });
                                }
                                remoteCurrent(source) {
                                    return __awaiter(this, void 0, void 0, function* () {
                                        nodeCurrent = yield source.getCurrent();
                                        return nodeCurrent;
                                    });
                                }
                                remotePeers(source) {
                                    return __awaiter(this, void 0, void 0, function* () {
                                        return Promise.resolve([node]);
                                    });
                                }
                                getLocalBlock(number) {
                                    return __awaiter(this, void 0, void 0, function* () {
                                        return server.dal.getBlock(number);
                                    });
                                }
                                getRemoteBlock(thePeer, number) {
                                    return __awaiter(this, void 0, void 0, function* () {
                                        let block = null;
                                        try {
                                            block = yield thePeer.getBlock(number);
                                            tx_cleaner_1.tx_cleaner(block.transactions);
                                        }
                                        catch (e) {
                                            if (e.httpCode != 404) {
                                                throw e;
                                            }
                                        }
                                        return block;
                                    });
                                }
                                applyMainBranch(block) {
                                    return __awaiter(this, void 0, void 0, function* () {
                                        const existing = yield server.dal.getAbsoluteBlockByNumberAndHash(block.number, block.hash);
                                        if (!existing) {
                                            let addedBlock = yield server.writeBlock(block, false, true);
                                            if (!this.lastDownloaded) {
                                                this.lastDownloaded = yield dao.remoteCurrent(node);
                                            }
                                            this.crawler.pullingEvent(server, 'applying', { number: block.number, last: this.lastDownloaded && this.lastDownloaded.number });
                                            if (addedBlock) {
                                                current = addedBlock;
                                                // Emit block events (for sharing with the network) only in forkWindowSize
                                                if (nodeCurrent && nodeCurrent.number - addedBlock.number < server.conf.forksize) {
                                                    server.streamPush(addedBlock);
                                                }
                                            }
                                        }
                                        return true;
                                    });
                                }
                                removeForks() {
                                    return __awaiter(this, void 0, void 0, function* () {
                                        return true;
                                    });
                                }
                                isMemberPeer(thePeer) {
                                    return __awaiter(this, void 0, void 0, function* () {
                                        return true;
                                    });
                                }
                                downloadBlocks(thePeer, fromNumber, count) {
                                    return __awaiter(this, void 0, void 0, function* () {
                                        if (!count) {
                                            count = this.crawler.CONST_BLOCKS_CHUNK;
                                        }
                                        let blocks = yield thePeer.getBlocks(count, fromNumber);
                                        // Fix for #734
                                        for (const block of blocks) {
                                            for (const tx of block.transactions) {
                                                tx.version = constants_1.CrawlerConstants.TRANSACTION_VERSION;
                                            }
                                        }
                                        return blocks;
                                    });
                                }
                            })(this);
                            yield dao.pull(server.conf, server.logger);
                        }
                        catch (e) {
                            if (this.isConnectionError(e)) {
                                this.logger && this.logger.info("Peer %s unreachable: now considered as DOWN.", p.pubkey);
                                yield server.dal.setPeerDown(p.pubkey);
                            }
                            else if (e.httpCode == 404) {
                                this.logger && this.logger.trace("No new block from %s %s", p.pubkey.substr(0, 6), p.getURL());
                            }
                            else {
                                this.logger && this.logger.warn(e);
                            }
                        }
                    })));
                    yield this.server.BlockchainService.pushFIFO("crawlerResolution", () => __awaiter(this, void 0, void 0, function* () {
                        yield server.BlockchainService.blockResolution();
                        yield server.BlockchainService.forkResolution();
                    }));
                    this.pullingEvent(server, 'end', current.number);
                }
                this.logger && this.logger.info('Will pull blocks from the network in %s min %s sec', Math.floor(this.pullingActualIntervalDuration / 60), Math.floor(this.pullingActualIntervalDuration % 60));
            }
            catch (err) {
                this.pullingEvent(server, 'error');
                this.logger && this.logger.warn(err.code || err.stack || err.message || err);
            }
        });
    }
    pullingEvent(server, type, number = null) {
        server.pullingEvent(type, number);
    }
    isConnectionError(err) {
        return err && (err.code == "E_DUNITER_PEER_CHANGED"
            || err.code == "EINVAL"
            || err.code == "ECONNREFUSED"
            || err.code == "ETIMEDOUT"
            || (err.httpCode !== undefined && err.httpCode !== 404));
    }
}
exports.BlockCrawler = BlockCrawler;
function chooseXin(peers, max) {
    const chosen = [];
    const nbPeers = peers.length;
    for (let i = 0; i < Math.min(nbPeers, max); i++) {
        const randIndex = Math.max(Math.floor(Math.random() * 10) - (10 - nbPeers) - i, 0);
        chosen.push(peers[randIndex]);
        peers.splice(randIndex, 1);
    }
    return chosen;
}
const checkPeerValidity = (server, p, node) => __awaiter(this, void 0, void 0, function* () {
    try {
        let document = yield node.getPeer();
        let thePeer = PeerDTO_1.PeerDTO.fromJSONObject(document);
        let goodSignature = server.PeeringService.checkPeerSignature(thePeer);
        if (!goodSignature) {
            throw 'Signature from a peer must match';
        }
        if (p.currency !== thePeer.currency) {
            throw 'Currency has changed from ' + p.currency + ' to ' + thePeer.currency;
        }
        if (p.pubkey !== thePeer.pubkey) {
            throw 'Public key of the peer has changed from ' + p.pubkey + ' to ' + thePeer.pubkey;
        }
        let sp1 = p.block.split('-');
        let sp2 = thePeer.block.split('-');
        let blockNumber1 = parseInt(sp1[0]);
        let blockNumber2 = parseInt(sp2[0]);
        if (blockNumber2 < blockNumber1) {
            throw 'Signature date has changed from block ' + blockNumber1 + ' to older block ' + blockNumber2;
        }
    }
    catch (e) {
        throw { code: "E_DUNITER_PEER_CHANGED" };
    }
});
//# sourceMappingURL=crawler.js.map