"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const DBBlock_1 = require("../../../lib/db/DBBlock");
const _ = require('underscore');
class PullingDao {
}
exports.PullingDao = PullingDao;
class AbstractDAO extends PullingDao {
    /**
     * Sugar function. Apply a bunch of blocks instead of one.
     * @param blocks
     */
    applyBranch(blocks) {
        return __awaiter(this, void 0, void 0, function* () {
            for (const block of blocks) {
                yield this.applyMainBranch(block);
            }
            return true;
        });
    }
    /**
     * Binary search algorithm to find the common root block between a local and a remote blockchain.
     * @param fork An object containing a peer, its current block and top fork block
     * @param forksize The maximum length we can look at to find common root block.
     * @returns {*|Promise}
     */
    findCommonRoot(fork, forksize) {
        return __awaiter(this, void 0, void 0, function* () {
            let commonRoot = null;
            let localCurrent = yield this.localCurrent();
            if (!localCurrent) {
                throw Error('Local blockchain is empty, cannot find a common root');
            }
            // We look between the top block that is known as fork ...
            let topBlock = fork.block;
            // ... and the bottom which is bounded by `forksize`
            let bottomBlock = yield this.getRemoteBlock(fork.peer, Math.max(0, localCurrent.number - forksize));
            let lookBlock = bottomBlock;
            let localEquivalent = yield this.getLocalBlock(bottomBlock.number);
            let isCommonBlock = lookBlock.hash == localEquivalent.hash;
            if (isCommonBlock) {
                // Then common root can be found between top and bottom. We process.
                let position = -1, wrongRemotechain = false;
                do {
                    isCommonBlock = lookBlock.hash == localEquivalent.hash;
                    if (!isCommonBlock) {
                        // Too high, look downward
                        topBlock = lookBlock;
                        position = middle(topBlock.number, bottomBlock.number);
                    }
                    else {
                        let upperBlock = yield this.getRemoteBlock(fork.peer, lookBlock.number + 1);
                        let localUpper = yield this.getLocalBlock(upperBlock.number);
                        let isCommonUpper = upperBlock.hash == localUpper.hash;
                        if (isCommonUpper) {
                            // Too low, look upward
                            bottomBlock = lookBlock;
                            position = middle(topBlock.number, bottomBlock.number);
                        }
                        else {
                            // Spotted!
                            commonRoot = lookBlock;
                        }
                    }
                    let noSpace = topBlock.number == bottomBlock.number + 1;
                    if (!commonRoot && noSpace) {
                        // Remote node have inconsistency blockchain, stop search
                        wrongRemotechain = true;
                    }
                    if (!wrongRemotechain) {
                        lookBlock = yield this.getRemoteBlock(fork.peer, position);
                        localEquivalent = yield this.getLocalBlock(position);
                    }
                } while (!commonRoot && !wrongRemotechain);
            }
            // Otherwise common root is unreachable
            return commonRoot;
        });
    }
    static defaultLocalBlock() {
        const localCurrent = new DBBlock_1.DBBlock();
        localCurrent.number = -1;
        return localCurrent;
    }
    /**
     * Pull algorithm. Look at given peers' blockchain and try to pull blocks from it.
     * May lead local blockchain to fork.
     * @param conf The local node configuration
     * @param dao An abstract layer to retrieve peers data (blocks).
     * @param logger Logger of the main application.
     */
    pull(conf, logger) {
        return __awaiter(this, void 0, void 0, function* () {
            let localCurrent = (yield this.localCurrent()) || AbstractDAO.defaultLocalBlock();
            const forks = [];
            if (!localCurrent) {
                localCurrent = new DBBlock_1.DBBlock();
                localCurrent.number = -1;
            }
            const applyCoroutine = (peer, blocks) => __awaiter(this, void 0, void 0, function* () {
                if (blocks.length > 0) {
                    let isFork = localCurrent
                        && !(blocks[0].previousHash == localCurrent.hash
                            && blocks[0].number == localCurrent.number + 1);
                    if (!isFork) {
                        yield this.applyBranch(blocks);
                        const newLocalCurrent = yield this.localCurrent();
                        localCurrent = newLocalCurrent || AbstractDAO.defaultLocalBlock();
                        const appliedSuccessfully = localCurrent.number == blocks[blocks.length - 1].number
                            && localCurrent.hash == blocks[blocks.length - 1].hash;
                        return appliedSuccessfully;
                    }
                    else {
                        let remoteCurrent = yield this.remoteCurrent(peer);
                        forks.push({
                            peer: peer,
                            block: blocks[0],
                            current: remoteCurrent
                        });
                        return false;
                    }
                }
                return true;
            });
            const downloadCoroutine = (peer, number) => __awaiter(this, void 0, void 0, function* () {
                return yield this.downloadBlocks(peer, number);
            });
            const downloadChuncks = (peer) => __awaiter(this, void 0, void 0, function* () {
                let blocksToApply = [];
                const currentBlock = yield this.localCurrent();
                let currentChunckStart;
                if (currentBlock) {
                    currentChunckStart = currentBlock.number + 1;
                }
                else {
                    currentChunckStart = 0;
                }
                let res = { applied: {}, downloaded: [] };
                do {
                    let [applied, downloaded] = yield Promise.all([
                        applyCoroutine(peer, blocksToApply),
                        downloadCoroutine(peer, currentChunckStart)
                    ]);
                    res.applied = applied;
                    res.downloaded = downloaded;
                    blocksToApply = downloaded;
                    currentChunckStart += downloaded.length;
                    if (!applied) {
                        logger && logger.info("Blocks were not applied.");
                    }
                } while (res.downloaded.length > 0 && res.applied);
            });
            let peers = yield this.remotePeers();
            // Try to get new legit blocks for local blockchain
            const downloadChuncksTasks = [];
            for (const peer of peers) {
                downloadChuncksTasks.push(downloadChuncks(peer));
            }
            yield Promise.all(downloadChuncksTasks);
            // Filter forks: do not include mirror peers (non-member peers)
            let memberForks = [];
            for (const fork of forks) {
                let isMember = yield this.isMemberPeer(fork.peer);
                if (isMember) {
                    memberForks.push(fork);
                }
            }
            memberForks = memberForks.sort((f1, f2) => {
                let result = compare(f1, f2, "number");
                if (result == 0) {
                    result = compare(f1, f2, "medianTime");
                }
                return result;
            });
            memberForks = _.filter(memberForks, (fork) => {
                let blockDistanceInBlocks = (fork.current.number - localCurrent.number);
                let timeDistanceInBlocks = (fork.current.medianTime - localCurrent.medianTime) / conf.avgGenTime;
                const requiredTimeAdvance = conf.switchOnHeadAdvance;
                logger && logger.debug('Fork of %s has blockDistance %s ; timeDistance %s ; required is >= %s for both values to try to follow the fork', fork.peer.pubkey.substr(0, 6), blockDistanceInBlocks.toFixed(2), timeDistanceInBlocks.toFixed(2), requiredTimeAdvance);
                return blockDistanceInBlocks >= requiredTimeAdvance
                    && timeDistanceInBlocks >= requiredTimeAdvance;
            });
            // Remove any previous fork block
            yield this.removeForks();
            // Find the common root block
            let j = 0, successFork = false;
            while (!successFork && j < memberForks.length) {
                let fork = memberForks[j];
                let commonRootBlock = yield this.findCommonRoot(fork, conf.forksize);
                if (commonRootBlock) {
                    let blocksToApply = yield this.downloadBlocks(fork.peer, commonRootBlock.number + 1, conf.forksize);
                    successFork = yield this.applyBranch(blocksToApply);
                }
                else {
                    logger && logger.debug('No common root block with peer %s', fork.peer.pubkey.substr(0, 6));
                }
                j++;
            }
            return this.localCurrent();
        });
    }
}
exports.AbstractDAO = AbstractDAO;
function compare(f1, f2, field) {
    if (f1[field] > f2[field]) {
        return 1;
    }
    if (f1[field] < f2[field]) {
        return -1;
    }
    return 0;
}
function middle(top, bottom) {
    let difference = top - bottom;
    if (difference % 2 == 1) {
        // We look one step below to not forget any block
        difference++;
    }
    return bottom + (difference / 2);
}
//# sourceMappingURL=pulling.js.map