"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const contacter_1 = require("./contacter");
const index_1 = require("../../../lib/common-libs/index");
const index_2 = require("../../../lib/common-libs/parsers/index");
exports.pullSandbox = (currency, fromHost, fromPort, toHost, toPort, logger) => __awaiter(this, void 0, void 0, function* () {
    const from = new contacter_1.Contacter(fromHost, fromPort);
    const to = new contacter_1.Contacter(toHost, toPort);
    let res;
    try {
        res = yield from.getRequirementsPending(1);
    }
    catch (e) {
        // Silent error
        logger && logger.trace('Sandbox pulling: could not fetch requirements on %s', [fromHost, fromPort].join(':'));
    }
    if (res) {
        const docs = getDocumentsTree(currency, res);
        for (const identity of docs.identities) {
            yield submitIdentity(identity, to);
        }
        for (const certification of docs.certifications) {
            yield submitCertification(certification, to);
        }
        for (const membership of docs.memberships) {
            yield submitMembership(membership, to);
        }
    }
});
exports.pullSandboxToLocalServer = (currency, fromHost, toServer, logger, watcher = null, nbCertsMin = 1, notify = true) => __awaiter(this, void 0, void 0, function* () {
    let res;
    try {
        res = yield fromHost.getRequirementsPending(nbCertsMin || 1);
    }
    catch (e) {
        watcher && watcher.writeStatus('Sandbox pulling: could not fetch requirements on %s', [fromHost.host, fromHost.port].join(':'));
    }
    if (res) {
        const docs = getDocumentsTree(currency, res);
        for (let i = 0; i < docs.identities.length; i++) {
            const idty = docs.identities[i];
            watcher && watcher.writeStatus('Identity ' + (i + 1) + '/' + docs.identities.length);
            yield submitIdentityToServer(idty, toServer, notify, logger);
        }
        for (let i = 0; i < docs.revocations.length; i++) {
            const idty = docs.revocations[i];
            watcher && watcher.writeStatus('Revocation ' + (i + 1) + '/' + docs.revocations.length);
            yield submitRevocationToServer(idty, toServer, notify, logger);
        }
        for (let i = 0; i < docs.certifications.length; i++) {
            const cert = docs.certifications[i];
            watcher && watcher.writeStatus('Certification ' + (i + 1) + '/' + docs.certifications.length);
            yield submitCertificationToServer(cert, toServer, notify, logger);
        }
        for (let i = 0; i < docs.memberships.length; i++) {
            const ms = docs.memberships[i];
            watcher && watcher.writeStatus('Membership ' + (i + 1) + '/' + docs.memberships.length);
            yield submitMembershipToServer(ms, toServer, notify, logger);
        }
    }
});
function getDocumentsTree(currency, res) {
    const documents = {
        identities: [],
        certifications: [],
        memberships: [],
        revocations: []
    };
    for (const idty of res.identities) {
        const identity = index_1.rawer.getOfficialIdentity({
            currency,
            uid: idty.uid,
            pubkey: idty.pubkey,
            buid: idty.meta.timestamp,
            sig: idty.sig
        });
        if (idty.revocation_sig) {
            const revocation = index_1.rawer.getOfficialRevocation({
                currency,
                uid: idty.uid,
                issuer: idty.pubkey,
                buid: idty.meta.timestamp,
                sig: idty.sig,
                revocation: idty.revocation_sig
            });
            documents.revocations.push(revocation);
        }
        documents.identities.push(identity);
        for (const cert of idty.pendingCerts) {
            const certification = index_1.rawer.getOfficialCertification({
                currency,
                idty_issuer: idty.pubkey,
                idty_uid: idty.uid,
                idty_buid: idty.meta.timestamp,
                idty_sig: idty.sig,
                issuer: cert.from,
                buid: cert.blockstamp,
                sig: cert.sig
            });
            documents.certifications.push(certification);
        }
        for (const ms of idty.pendingMemberships) {
            const membership = index_1.rawer.getMembership({
                currency,
                userid: idty.uid,
                issuer: idty.pubkey,
                certts: idty.meta.timestamp,
                membership: ms.type,
                block: ms.blockstamp,
                signature: ms.sig
            });
            documents.memberships.push(membership);
        }
    }
    return documents;
}
function submitIdentity(idty, to, logger = null) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            yield to.postIdentity(idty);
            logger && logger.trace('Sandbox pulling: success with identity \'%s\'', idty.uid);
        }
        catch (e) {
            // Silent error
        }
    });
}
function submitCertification(cert, to, logger = null) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            yield to.postCert(cert);
            logger && logger.trace('Sandbox pulling: success with cert key %s => %s', cert.from.substr(0, 6), cert.idty_uid);
        }
        catch (e) {
            // Silent error
        }
    });
}
function submitMembership(ms, to, logger = null) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            yield to.postRenew(ms);
            logger && logger.trace('Sandbox pulling: success with membership \'%s\'', ms.uid);
        }
        catch (e) {
            // Silent error
        }
    });
}
function submitIdentityToServer(idty, toServer, notify, logger) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            const obj = index_2.parsers.parseIdentity.syncWrite(idty);
            yield toServer.writeIdentity(obj, notify);
            logger && logger.trace('Sandbox pulling: success with identity \'%s\'', obj.uid);
        }
        catch (e) {
            // Silent error
        }
    });
}
function submitRevocationToServer(revocation, toServer, notify, logger) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            const obj = index_2.parsers.parseRevocation.syncWrite(revocation);
            yield toServer.writeRevocation(obj, notify);
            logger && logger.trace('Sandbox pulling: success with revocation \'%s\'', obj.uid);
        }
        catch (e) {
            // Silent error
        }
    });
}
function submitCertificationToServer(cert, toServer, notify, logger) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            const obj = index_2.parsers.parseCertification.syncWrite(cert);
            yield toServer.writeCertification(obj, notify);
            logger && logger.trace('Sandbox pulling: success with cert key %s => %s', cert.from.substr(0, 6), cert.idty_uid);
        }
        catch (e) {
            // Silent error
        }
    });
}
function submitMembershipToServer(ms, toServer, notify, logger) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            const obj = index_2.parsers.parseMembership.syncWrite(ms);
            yield toServer.writeMembership(obj, notify);
            logger && logger.trace('Sandbox pulling: success with membership \'%s\'', ms.uid);
        }
        catch (e) {
            // Silent error
        }
    });
}
//# sourceMappingURL=sandbox.js.map