"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const blockGenerator_1 = require("./blockGenerator");
const blockProver_1 = require("./blockProver");
const dos2unix_1 = require("../../../lib/common-libs/dos2unix");
const index_1 = require("../../../lib/common-libs/parsers/index");
const querablep = require('querablep');
class PermanentProver {
    constructor(server) {
        this.server = server;
        this.permanencePromise = null;
        this.blockchainChangedResolver = null;
        this.promiseOfWaitingBetween2BlocksOfOurs = null;
        this.lastComputedBlock = null;
        this.resolveContinuePromise = null;
        this.continuePromise = null;
        this.logger = server.logger;
        this.conf = server.conf;
        this.prover = new blockProver_1.BlockProver(server);
        this.generator = new blockGenerator_1.BlockGeneratorWhichProves(server, this.prover);
        // Promises triggering the prooving lopp
        this.resolveContinuePromise = null;
        this.continuePromise = new Promise((resolve) => this.resolveContinuePromise = resolve);
        this.loops = 0;
    }
    allowedToStart() {
        if (!this.permanencePromise || this.permanencePromise.isFulfilled()) {
            this.startPermanence();
        }
        this.resolveContinuePromise(true);
    }
    startPermanence() {
        return __awaiter(this, void 0, void 0, function* () {
            let permanenceResolve = () => { };
            this.permanencePromise = querablep(new Promise(res => {
                permanenceResolve = res;
            }));
            /******************
             * Main proof loop
             *****************/
            while (yield this.continuePromise) {
                try {
                    const waitingRaces = [];
                    // By default, we do not make a new proof
                    let doProof = false;
                    try {
                        const selfPubkey = this.server.keyPair.publicKey;
                        const dal = this.server.dal;
                        const theConf = this.server.conf;
                        if (!selfPubkey) {
                            throw 'No self pubkey found.';
                        }
                        let current;
                        const isMember = yield dal.isMember(selfPubkey);
                        if (!isMember) {
                            throw 'Local node is not a member. Waiting to be a member before computing a block.';
                        }
                        current = yield dal.getCurrentBlockOrNull();
                        if (!current) {
                            throw 'Waiting for a root block before computing new blocks';
                        }
                        const trial = yield this.server.getBcContext().getIssuerPersonalizedDifficulty(selfPubkey);
                        this.checkTrialIsNotTooHigh(trial, current, selfPubkey);
                        const lastIssuedByUs = current.issuer == selfPubkey;
                        if (lastIssuedByUs && !this.promiseOfWaitingBetween2BlocksOfOurs) {
                            this.promiseOfWaitingBetween2BlocksOfOurs = new Promise((resolve) => setTimeout(resolve, theConf.powDelay));
                            this.logger.warn('Waiting ' + theConf.powDelay + 'ms before starting to compute next block...');
                        }
                        else {
                            // We have waited enough
                            this.promiseOfWaitingBetween2BlocksOfOurs = null;
                            // But under some conditions, we can make one
                            doProof = true;
                        }
                    }
                    catch (e) {
                        this.logger.warn(e);
                    }
                    if (doProof) {
                        /*******************
                         * COMPUTING A BLOCK
                         ******************/
                        try {
                            let cancelAlreadyTriggered = false;
                            // The canceller
                            (() => __awaiter(this, void 0, void 0, function* () {
                                // If the blockchain changes
                                yield new Promise((resolve) => this.blockchainChangedResolver = resolve);
                                cancelAlreadyTriggered = true;
                                // Then cancel the generation
                                yield this.prover.cancel();
                            }))();
                            let unsignedBlock = null, trial2 = 0;
                            if (!cancelAlreadyTriggered) {
                                // The pushFIFO is here to get the difficulty level while excluding any new block to be resolved.
                                // Without it, a new block could be added meanwhile and would make the difficulty wrongly computed.
                                yield this.server.BlockchainService.pushFIFO('generatingNextBlock', () => __awaiter(this, void 0, void 0, function* () {
                                    const current = yield this.server.dal.getCurrentBlockOrNull();
                                    const selfPubkey = this.server.keyPair.publicKey;
                                    if (!cancelAlreadyTriggered) {
                                        trial2 = yield this.server.getBcContext().getIssuerPersonalizedDifficulty(selfPubkey);
                                    }
                                    this.checkTrialIsNotTooHigh(trial2, current, selfPubkey);
                                    if (!cancelAlreadyTriggered) {
                                        unsignedBlock = yield this.generator.nextBlock();
                                    }
                                }));
                                if (!cancelAlreadyTriggered) {
                                    this.lastComputedBlock = yield this.prover.prove(unsignedBlock, trial2, null);
                                }
                                try {
                                    const obj = index_1.parsers.parseBlock.syncWrite(dos2unix_1.dos2unix(this.lastComputedBlock.getRawSigned()));
                                    yield this.server.writeBlock(obj);
                                    yield new Promise(res => {
                                        this.server.once('bcEvent', () => res());
                                    });
                                }
                                catch (err) {
                                    this.logger.warn('Proof-of-work self-submission: %s', err.message || err);
                                }
                            }
                        }
                        catch (e) {
                            this.logger.warn('The proof-of-work generation was canceled: %s', (e && e.message) || (e && e.uerr && e.uerr.message) || e || 'unkonwn reason');
                        }
                    }
                    else {
                        /*******************
                         * OR WAITING PHASE
                         ******************/
                        if (this.promiseOfWaitingBetween2BlocksOfOurs) {
                            waitingRaces.push(this.promiseOfWaitingBetween2BlocksOfOurs);
                        }
                        let raceDone = false;
                        yield Promise.race(waitingRaces.concat([
                            // The blockchain has changed! We or someone else found a proof, we must make a gnu one
                            new Promise((resolve) => this.blockchainChangedResolver = () => {
                                this.logger.warn('Blockchain changed!');
                                resolve();
                            }),
                            // Security: if nothing happens for a while, trigger the whole process again
                            new Promise((resolve) => setTimeout(() => {
                                if (!raceDone) {
                                    this.logger.warn('Security trigger: proof-of-work process seems stuck');
                                    resolve();
                                }
                            }, this.conf.powSecurityRetryDelay))
                        ]));
                        raceDone = true;
                    }
                }
                catch (e) {
                    this.logger.warn(e);
                }
                this.loops++;
                // Informative variable
                this.logger.trace('PoW loops = %s', this.loops);
            }
            permanenceResolve();
        });
    }
    blockchainChanged(gottenBlock) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.server && (!gottenBlock || !this.lastComputedBlock || gottenBlock.hash !== this.lastComputedBlock.hash)) {
                // Cancel any processing proof
                yield this.prover.cancel();
                // If we were waiting, stop it and process the continuous generation
                this.blockchainChangedResolver && this.blockchainChangedResolver();
            }
        });
    }
    stopEveryting() {
        return __awaiter(this, void 0, void 0, function* () {
            // First: avoid continuing the main loop
            this.resolveContinuePromise(true);
            this.continuePromise = new Promise((resolve) => this.resolveContinuePromise = resolve);
            // Second: stop any started proof
            yield this.prover.cancel();
            // If we were waiting, stop it and process the continuous generation
            this.blockchainChangedResolver && this.blockchainChangedResolver();
        });
    }
    checkTrialIsNotTooHigh(trial, current, selfPubkey) {
        if (trial > (current.powMin + this.conf.powMaxHandicap)) {
            this.logger.debug('Trial = %s, powMin = %s, pubkey = %s', trial, current.powMin, selfPubkey.slice(0, 6));
            throw 'Too high difficulty: waiting for other members to write next block';
        }
    }
}
exports.PermanentProver = PermanentProver;
//# sourceMappingURL=permanentProver.js.map