import { Key } from "../../../lib/common-libs/crypto/keyring";
import { WS2PMessageHandler } from "./impl/WS2PMessageHandler";
import { BlockDTO } from "../../../lib/dto/BlockDTO";
import { IdentityDTO } from "../../../lib/dto/IdentityDTO";
import { CertificationDTO } from "../../../lib/dto/CertificationDTO";
import { MembershipDTO } from "../../../lib/dto/MembershipDTO";
import { TransactionDTO } from "../../../lib/dto/TransactionDTO";
import { PeerDTO } from "../../../lib/dto/PeerDTO";
export declare enum WS2P_PUSH {
    PEER = 0,
    TRANSACTION = 1,
    MEMBERSHIP = 2,
    CERTIFICATION = 3,
    IDENTITY = 4,
    BLOCK = 5,
    HEAD = 6,
}
export interface WS2PAuth {
    authenticationIsDone(): Promise<void>;
}
export interface WS2PRemoteAuth extends WS2PAuth {
    registerCONNECT(challenge: string, sig: string, pub: string): Promise<boolean>;
    sendACK(ws: any): Promise<void>;
    registerOK(sig: string): Promise<boolean>;
    isAuthenticatedByRemote(): boolean;
    getPubkey(): string;
}
export interface WS2PLocalAuth extends WS2PAuth {
    sendCONNECT(ws: any): Promise<void>;
    registerACK(sig: string, pub: string): Promise<boolean>;
    sendOK(ws: any): Promise<void>;
    isRemoteAuthenticated(): boolean;
}
/**
 * A passive authenticator based on our keyring.
 */
export declare class WS2PPubkeyRemoteAuth implements WS2PRemoteAuth {
    protected currency: string;
    protected pair: Key;
    protected tellIsAuthorizedPubkey: (pub: string) => Promise<boolean>;
    protected challenge: string;
    protected authenticatedByRemote: boolean;
    protected remotePub: string;
    protected serverAuth: Promise<void>;
    protected serverAuthResolve: () => void;
    protected serverAuthReject: (err: any) => void;
    constructor(currency: string, pair: Key, tellIsAuthorizedPubkey?: (pub: string) => Promise<boolean>);
    getPubkey(): string;
    sendACK(ws: any): Promise<void>;
    registerCONNECT(challenge: string, sig: string, pub: string): Promise<boolean>;
    registerOK(sig: string): Promise<boolean>;
    isAuthenticatedByRemote(): boolean;
    authenticationIsDone(): Promise<void>;
}
/**
 * An connecting authenticator based on our keyring.
 */
export declare class WS2PPubkeyLocalAuth implements WS2PLocalAuth {
    protected currency: string;
    protected pair: Key;
    protected tellIsAuthorizedPubkey: (pub: string) => Promise<boolean>;
    protected challenge: string;
    protected authenticated: boolean;
    protected serverAuth: Promise<void>;
    protected serverAuthResolve: () => void;
    protected serverAuthReject: (err: any) => void;
    constructor(currency: string, pair: Key, tellIsAuthorizedPubkey?: (pub: string) => Promise<boolean>);
    sendCONNECT(ws: any): Promise<void>;
    registerACK(sig: string, pub: string): Promise<boolean>;
    sendOK(ws: any): Promise<void>;
    isRemoteAuthenticated(): boolean;
    authenticationIsDone(): Promise<void>;
}
export interface WS2PRequest {
    name: string;
    params?: any;
}
/**
 * The handler of a WS2P connection.
 *
 * Goal: operating an authenticated bidirectionnal communication over a WebSocket connection.
 * Requires an established WebSocket connection in order to work.
 */
export declare class WS2PConnection {
    private ws;
    private onWsOpened;
    private onWsClosed;
    private messageHandler;
    private localAuth;
    private remoteAuth;
    private options;
    private expectedPub;
    private connectp;
    private connectedp;
    private connectedResolve;
    private connectedReject;
    private nbErrors;
    private nbRequestsCount;
    private nbResponsesCount;
    private nbPushsToRemoteCount;
    private nbPushsByRemoteCount;
    private exchanges;
    constructor(ws: any, onWsOpened: Promise<void>, onWsClosed: Promise<void>, messageHandler: WS2PMessageHandler, localAuth: WS2PLocalAuth, remoteAuth: WS2PRemoteAuth, options?: {
        connectionTimeout: number;
        requestTimeout: number;
    }, expectedPub?: string);
    static newConnectionToAddress(address: string, messageHandler: WS2PMessageHandler, localAuth: WS2PLocalAuth, remoteAuth: WS2PRemoteAuth, proxySocksAddress?: string | undefined, options?: {
        connectionTimeout: number;
        requestTimeout: number;
    }, expectedPub?: string): WS2PConnection;
    static newConnectionFromWebSocketServer(websocket: any, messageHandler: WS2PMessageHandler, localAuth: WS2PLocalAuth, remoteAuth: WS2PRemoteAuth, options?: {
        connectionTimeout: number;
        requestTimeout: number;
    }, expectedPub?: string): WS2PConnection;
    readonly pubkey: string;
    readonly nbRequests: number;
    readonly nbResponses: number;
    readonly nbPushsToRemote: number;
    readonly nbPushsByRemote: number;
    readonly connected: Promise<string>;
    readonly closed: Promise<void>;
    close(): any;
    connect(): Promise<any>;
    request(body: WS2PRequest): Promise<{}>;
    pushBlock(block: BlockDTO): Promise<{}>;
    pushIdentity(idty: IdentityDTO): Promise<{}>;
    pushCertification(cert: CertificationDTO): Promise<{}>;
    pushMembership(ms: MembershipDTO): Promise<{}>;
    pushTransaction(tx: TransactionDTO): Promise<{}>;
    pushPeer(peer: PeerDTO): Promise<{}>;
    pushHeads(heads: {
        message: string;
        sig: string;
    }[]): Promise<{}>;
    pushData(type: WS2P_PUSH, key: string, data: any): Promise<{}>;
    private errorDetected(cause);
}
