"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const WS2PConnection_1 = require("./WS2PConnection");
const keyring_1 = require("../../../lib/common-libs/crypto/keyring");
const events = require("events");
const constants_1 = require("./constants");
const WS2PStreamer_1 = require("./WS2PStreamer");
const WS2PSingleWriteStream_1 = require("./WS2PSingleWriteStream");
const WebSocketServer = require('ws').Server;
class WS2PServer extends events.EventEmitter {
    constructor(server, host, port, fifo, shouldAcceptConnection) {
        super();
        this.server = server;
        this.host = host;
        this.port = port;
        this.fifo = fifo;
        this.shouldAcceptConnection = shouldAcceptConnection;
        this.connections = [];
        this.maxLevel2Size = constants_1.WS2PConstants.MAX_LEVEL_2_PEERS;
        // Conf: max public connections
        if (this.server.conf.ws2p && this.server.conf.ws2p.maxPublic !== undefined) {
            this.maxLevel2Size = this.server.conf.ws2p.maxPublic;
        }
    }
    get maxLevel2Peers() {
        return this.maxLevel2Size || 0;
    }
    set maxLevel2Peers(newValue) {
        this.maxLevel2Size = Math.max(newValue, 0);
    }
    getConnexions() {
        return this.connections.slice();
    }
    listenToWebSocketConnections(messageHandler) {
        const key = new keyring_1.Key(this.server.conf.pair.pub, this.server.conf.pair.sec);
        this.wss = new WebSocketServer({ host: this.host, port: this.port });
        this.wss.on('connection', (ws) => __awaiter(this, void 0, void 0, function* () {
            this.server.logger.info('WS2P %s: new incoming connection from %s:%s!', this.server.conf.pair.pub, ws._sender._socket._handle.owner.remoteAddress, ws._sender._socket._handle.owner.remotePort);
            /******************
             * A NEW CONNECTION
             ******************/
            let saidPubkey = "";
            const acceptPubkey = (pub) => __awaiter(this, void 0, void 0, function* () {
                if (!saidPubkey) {
                    saidPubkey = pub;
                }
                if (saidPubkey !== pub) {
                    // The key must be identical
                    return false;
                }
                return yield this.shouldAcceptConnection(pub, this.getConnexions().map(c => c.pubkey));
            });
            let timeout = {
                connectionTimeout: constants_1.WS2PConstants.CONNEXION_TIMEOUT,
                requestTimeout: constants_1.WS2PConstants.REQUEST_TIMEOUT
            };
            if (this.server.conf.ws2p && this.server.conf.ws2p.remotehost && this.server.conf.ws2p.remotehost.match(constants_1.WS2PConstants.HOST_ONION_REGEX)) {
                timeout = {
                    connectionTimeout: constants_1.WS2PConstants.CONNEXION_TOR_TIMEOUT,
                    requestTimeout: constants_1.WS2PConstants.REQUEST_TOR_TIMEOUT
                };
            }
            const c = WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, messageHandler, new WS2PConnection_1.WS2PPubkeyLocalAuth(this.server.conf.currency, key, acceptPubkey), new WS2PConnection_1.WS2PPubkeyRemoteAuth(this.server.conf.currency, key, acceptPubkey), timeout);
            try {
                yield c.connect();
                const host = ws._sender._socket._handle.owner.remoteAddress;
                const port = ws._sender._socket._handle.owner.remotePort;
                this.server.push({
                    ws2p: 'connected',
                    to: { host, port, pubkey: c.pubkey }
                });
                this.connections.push(c);
                this.emit('newConnection', c);
                this.server.logger.info('WS2P: established incoming connection from %s:%s', host, port);
                // Broadcasting
                const singleWriteProtection = new WS2PSingleWriteStream_1.WS2PSingleWriteStream();
                const ws2pStreamer = new WS2PStreamer_1.WS2PStreamer(c);
                this.server
                    .pipe(singleWriteProtection)
                    .pipe(ws2pStreamer);
                ws.on('error', (e) => {
                    this.server.logger.error(e);
                });
                ws.on('close', () => {
                    this.server.unpipe(singleWriteProtection);
                    singleWriteProtection.unpipe(ws2pStreamer);
                    this.removeConnection(c);
                    this.server.push({
                        ws2p: 'disconnected',
                        peer: {
                            pub: c.pubkey
                        }
                    });
                });
                yield this.trimConnections();
                yield this.server.dal.setPeerUP(c.pubkey);
            }
            catch (e) {
                ws.close();
                this.server.logger.warn('WS2P: cannot connect to incoming WebSocket connection: %s', e);
            }
        }));
    }
    trimConnections() {
        return __awaiter(this, void 0, void 0, function* () {
            /*** OVERFLOW TRIMMING ***/
            let disconnectedOne = true;
            // Disconnect non-members
            while (disconnectedOne && this.connections.length > this.maxLevel2Size) {
                disconnectedOne = false;
                for (const c of this.connections) {
                    const isMember = yield this.server.dal.isMember(c.pubkey);
                    if (!isMember && !disconnectedOne) {
                        c.close();
                        this.removeConnection(c);
                        disconnectedOne = true;
                    }
                }
            }
            // Disconnect members
            while (this.connections.length > this.maxLevel2Size) {
                for (const c of this.connections) {
                    c.close();
                    this.removeConnection(c);
                }
            }
            /*** DUPLICATES TRIMMING ***/
            disconnectedOne = true;
            while (disconnectedOne) {
                disconnectedOne = false;
                const pubkeysFound = [];
                for (const c of this.connections) {
                    if (pubkeysFound.indexOf(c.pubkey) !== -1) {
                        c.close();
                        this.removeConnection(c);
                        disconnectedOne = true;
                    }
                    else if (c.pubkey !== this.server.conf.pair.pub) {
                        pubkeysFound.push(c.pubkey);
                    }
                }
            }
        });
    }
    removeConnection(c) {
        const index = this.connections.indexOf(c);
        if (index !== -1) {
            // Remove the connection
            this.connections.splice(index, 1);
        }
    }
    close() {
        return __awaiter(this, void 0, void 0, function* () {
            yield Promise.all(this.connections.map(c => c.close()));
            return new Promise((res, rej) => {
                this.wss.close((err) => {
                    if (err)
                        return rej(err);
                    res();
                });
            });
        });
    }
    getConnection(pubkeyOfConnection) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.connections.length === 0) {
                throw Error("No connections to look into.");
            }
            return Promise.race(this.connections.map((c) => __awaiter(this, void 0, void 0, function* () {
                yield c.connected;
                if (c.pubkey === pubkeyOfConnection) {
                    return c;
                }
                else {
                    yield new Promise(resolve => setTimeout(resolve, constants_1.WS2PConstants.CONNEXION_TIMEOUT));
                    throw Error("Pubkey not matching or too long to be obtained");
                }
            })));
        });
    }
    static bindOn(server, host, port, fifo, shouldAcceptConnection, messageHandler) {
        return __awaiter(this, void 0, void 0, function* () {
            const ws2ps = new WS2PServer(server, host, port, fifo, shouldAcceptConnection);
            yield ws2ps.listenToWebSocketConnections(messageHandler);
            server.logger.info('WS2P server %s listening on %s:%s', server.conf.pair.pub, host, port);
            return ws2ps;
        });
    }
}
exports.WS2PServer = WS2PServer;
//# sourceMappingURL=WS2PServer.js.map