"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const constants_1 = require("./constants");
const upnp = require('nnupnp');
class WS2PUpnp {
    constructor(logger) {
        this.logger = logger;
        this.client = upnp.createClient();
    }
    checkUPnPisAvailable() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield new Promise((resolve, reject) => {
                    this.client.externalIp((err, res) => {
                        if (err || !res) {
                            reject();
                        }
                        else {
                            resolve();
                        }
                    });
                });
                return true;
            }
            catch (err) {
                return false;
            }
        });
    }
    getCurrentConfig() {
        return this.currentConfig;
    }
    /**
     * Always open the same port during an execution of Duniter.
     * @returns { host:string, port:number }
     */
    openPort() {
        return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
            if (!this.currentConfig) {
                this.currentConfig = yield WS2PUpnp.getAvailablePort(this.client);
            }
            this.logger.trace('WS2P: mapping external port %s to local %s using UPnP...', this.currentConfig.port, [this.currentConfig.host, this.currentConfig.port].join(':'));
            const client = upnp.createClient();
            client.portMapping({
                'public': this.currentConfig.port,
                'private': this.currentConfig.port,
                'ttl': constants_1.WS2PConstants.WS2P_UPNP_TTL,
                'description': 'duniter:ws2p:upnp'
            }, (err) => {
                client.close();
                if (err) {
                    this.logger.warn(err);
                    return reject(err);
                }
                resolve(this.currentConfig);
            });
        }));
    }
    startRegular() {
        return __awaiter(this, void 0, void 0, function* () {
            this.stopRegular();
            const available = yield this.checkUPnPisAvailable();
            if (available) {
                // Update UPnP IGD every INTERVAL seconds
                this.interval = setInterval(() => this.openPort(), 1000 * constants_1.WS2PConstants.WS2P_UPNP_INTERVAL);
                const { host, port } = yield this.openPort();
                return { host, port, available };
            }
            return { host: '', port: 0, available: false };
        });
    }
    stopRegular() {
        if (this.interval) {
            clearInterval(this.interval);
        }
    }
    static getLocalIP(client) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield new Promise((resolve, reject) => {
                client.findGateway((err, res, localIP) => {
                    if (err)
                        return reject(err);
                    resolve(localIP);
                });
            });
        });
    }
    static getRemoteIP(client) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield new Promise((resolve, reject) => {
                client.externalIp((err, externalIP) => {
                    if (err)
                        return reject(err);
                    resolve(externalIP);
                });
            });
        });
    }
    static getAvailablePort(client) {
        return __awaiter(this, void 0, void 0, function* () {
            const localIP = yield WS2PUpnp.getLocalIP(client);
            const remoteIP = yield WS2PUpnp.getRemoteIP(client);
            const mappings = yield WS2PUpnp.getUPnPMappings(client);
            const externalPortsUsed = mappings.map((m) => {
                return m.public.port;
            });
            let availablePort = constants_1.WS2PConstants.WS2P_PORTS_START;
            while (externalPortsUsed.indexOf(availablePort) !== -1
                && availablePort <= constants_1.WS2PConstants.WS2P_PORTS_END) {
                availablePort++;
            }
            if (availablePort > constants_1.WS2PConstants.WS2P_PORTS_END) {
                throw "No port available for UPnP";
            }
            return {
                remotehost: remoteIP,
                host: localIP,
                port: availablePort
            };
        });
    }
    static getUPnPMappings(client) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                client.getMappings((err, res) => {
                    if (err) {
                        reject(err);
                    }
                    else {
                        resolve(res);
                    }
                });
            });
        });
    }
}
exports.WS2PUpnp = WS2PUpnp;
//# sourceMappingURL=ws2p-upnp.js.map