"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const BlockchainContext_1 = require("../lib/computation/BlockchainContext");
const QuickSync_1 = require("../lib/computation/QuickSync");
const BlockDTO_1 = require("../lib/dto/BlockDTO");
const global_rules_1 = require("../lib/rules/global_rules");
const index_1 = require("../lib/common-libs/parsers/index");
const FIFOService_1 = require("./FIFOService");
const constants_1 = require("../lib/common-libs/constants");
const local_rules_1 = require("../lib/rules/local_rules");
const Switcher_1 = require("../lib/blockchain/Switcher");
const other_constants_1 = require("../lib/other_constants");
const _ = require('underscore');
const constants = require('../lib/constants');
class BlockchainService extends FIFOService_1.FIFOService {
    constructor(server, fifoPromiseHandler) {
        super(fifoPromiseHandler);
        this.server = server;
        this.invalidForks = [];
        this.mainContext = new BlockchainContext_1.BlockchainContext();
        this.switcherDao = new (class ForkDao {
            constructor(bcService) {
                this.bcService = bcService;
            }
            getCurrent() {
                return this.bcService.current();
            }
            getPotentials(numberStart, timeStart, maxNumber) {
                return __awaiter(this, void 0, void 0, function* () {
                    const blocks = yield this.bcService.dal.getPotentialForkBlocks(numberStart, timeStart, maxNumber);
                    return blocks.map((b) => BlockDTO_1.BlockDTO.fromJSONObject(b));
                });
            }
            getBlockchainBlock(number, hash) {
                return __awaiter(this, void 0, void 0, function* () {
                    try {
                        return BlockDTO_1.BlockDTO.fromJSONObject(yield this.bcService.dal.getBlockByNumberAndHash(number, hash));
                    }
                    catch (e) {
                        return null;
                    }
                });
            }
            getSandboxBlock(number, hash) {
                return __awaiter(this, void 0, void 0, function* () {
                    const block = yield this.bcService.dal.getAbsoluteBlockByNumberAndHash(number, hash);
                    if (block && block.fork) {
                        return BlockDTO_1.BlockDTO.fromJSONObject(block);
                    }
                    else {
                        return null;
                    }
                });
            }
            revertTo(number) {
                return __awaiter(this, void 0, void 0, function* () {
                    const blocks = [];
                    const current = yield this.bcService.current();
                    for (let i = 0, count = current.number - number; i < count; i++) {
                        const reverted = yield this.bcService.mainContext.revertCurrentBlock();
                        blocks.push(BlockDTO_1.BlockDTO.fromJSONObject(reverted));
                    }
                    if (current.number < number) {
                        throw "Already below this number";
                    }
                    return blocks;
                });
            }
            addBlock(block) {
                return __awaiter(this, void 0, void 0, function* () {
                    return yield this.bcService.mainContext.checkAndAddBlock(block);
                });
            }
        })(this);
    }
    /**
     * Mandatory with stream.Readable
     * @private
     */
    _read() { }
    getContext() {
        return this.mainContext;
    }
    setConfDAL(newConf, newDAL, newKeyPair) {
        this.dal = newDAL;
        this.conf = newConf;
        this.logger = require('../lib/logger').NewLogger(this.dal.profile);
        this.quickSynchronizer = new QuickSync_1.QuickSynchronizer(this.server.blockchain, this.conf, this.dal, this.logger);
        this.mainContext.setConfDAL(this.conf, this.dal, this.server.blockchain, this.quickSynchronizer);
        this.selfPubkey = newKeyPair.publicKey;
    }
    current() {
        return this.dal.getCurrentBlockOrNull();
    }
    promoted(number) {
        return __awaiter(this, void 0, void 0, function* () {
            const bb = yield this.dal.getPromoted(number);
            if (!bb)
                throw constants.ERRORS.BLOCK_NOT_FOUND;
            return bb;
        });
    }
    checkBlock(block, withPoWAndSignature = true) {
        const dto = BlockDTO_1.BlockDTO.fromJSONObject(block);
        return this.mainContext.checkBlock(dto, withPoWAndSignature);
    }
    branches() {
        return __awaiter(this, void 0, void 0, function* () {
            const current = yield this.current();
            const switcher = new Switcher_1.Switcher(this.switcherDao, this.invalidForks, this.conf.avgGenTime, this.conf.forksize, this.conf.switchOnHeadAdvance, this.logger);
            const heads = yield switcher.findPotentialSuitesHeads(current);
            return heads.concat([current]);
        });
    }
    submitBlock(blockToAdd, noResolution = false) {
        const obj = index_1.parsers.parseBlock.syncWrite(BlockDTO_1.BlockDTO.fromJSONObject(blockToAdd).getRawSigned());
        const dto = BlockDTO_1.BlockDTO.fromJSONObject(obj);
        const hash = dto.getHash();
        return this.pushFIFO(hash, () => __awaiter(this, void 0, void 0, function* () {
            // Check basic fields:
            // * currency relatively to conf
            if (this.conf && this.conf.currency && this.conf.currency !== dto.currency) {
                throw constants_1.CommonConstants.ERRORS.WRONG_CURRENCY;
            }
            // * hash relatively to powMin
            if (!local_rules_1.LOCAL_RULES_FUNCTIONS.isProofOfWorkCorrect(dto)) {
                throw constants_1.CommonConstants.ERRORS.WRONG_POW;
            }
            // * number relatively to fork window and current block
            if (this.conf && this.conf.forksize !== undefined) {
                const current = yield this.current();
                if (current && dto.number < current.number - this.conf.forksize) {
                    throw constants_1.CommonConstants.ERRORS.OUT_OF_FORK_WINDOW;
                }
            }
            const absolute = yield this.dal.getAbsoluteBlockByNumberAndHash(obj.number, obj.hash);
            if (!absolute) {
                // Save the block in the sandbox
                yield this.mainContext.addSideBlock(dto);
                // Trigger the save + block resolution in an async way: this allows to answer immediately that the submission
                // was accepted, and that the document can be rerouted and is under treatment.
                // This will enhence the block propagation on the network, thus will avoid potential forks to emerge.
                if (!noResolution) {
                    (() => {
                        return this.pushFIFO('resolution_' + dto.getHash(), () => __awaiter(this, void 0, void 0, function* () {
                            // Resolve the potential new HEAD
                            yield this.blockResolution();
                            // Resolve the potential forks
                            yield this.forkResolution();
                            const current = yield this.current();
                            this.push({
                                bcEvent: other_constants_1.OtherConstants.BC_EVENT.RESOLUTION_DONE,
                                block: current
                            });
                        }));
                    })();
                }
            }
            else {
                throw "Block already known";
            }
            return dto;
        }));
    }
    blockResolution() {
        return __awaiter(this, void 0, void 0, function* () {
            let added = true;
            while (added) {
                const current = yield this.current();
                let potentials = [];
                if (current) {
                    potentials = yield this.dal.getForkBlocksFollowing(current);
                    this.logger.info('Block resolution: %s potential blocks after current#%s...', potentials.length, current.number);
                }
                else {
                    potentials = yield this.dal.getPotentialRootBlocks();
                    this.logger.info('Block resolution: %s potential blocks for root block...', potentials.length);
                }
                added = false;
                let i = 0;
                while (!added && i < potentials.length) {
                    const dto = BlockDTO_1.BlockDTO.fromJSONObject(potentials[i]);
                    try {
                        const addedBlock = yield this.mainContext.checkAndAddBlock(dto);
                        added = true;
                        this.push({
                            bcEvent: other_constants_1.OtherConstants.BC_EVENT.HEAD_CHANGED,
                            block: addedBlock
                        });
                        // Clear invalid forks' cache
                        this.invalidForks.splice(0, this.invalidForks.length);
                    }
                    catch (e) {
                        this.logger.error(e);
                        added = false;
                        this.push({
                            blockResolutionError: e && e.message
                        });
                    }
                    i++;
                }
            }
        });
    }
    forkResolution() {
        return __awaiter(this, void 0, void 0, function* () {
            const switcher = new Switcher_1.Switcher(this.switcherDao, this.invalidForks, this.conf.avgGenTime, this.conf.forksize, this.conf.switchOnHeadAdvance, this.logger);
            const newCurrent = yield switcher.tryToFork();
            if (newCurrent) {
                this.push({
                    bcEvent: other_constants_1.OtherConstants.BC_EVENT.SWITCHED,
                    block: newCurrent
                });
            }
        });
    }
    revertCurrentBlock() {
        return this.pushFIFO("revertCurrentBlock", () => this.mainContext.revertCurrentBlock());
    }
    applyNextAvailableFork() {
        return this.pushFIFO("applyNextAvailableFork", () => this.mainContext.applyNextAvailableFork());
    }
    requirementsOfIdentities(identities, computeDistance = true) {
        return __awaiter(this, void 0, void 0, function* () {
            let all = [];
            let current = yield this.dal.getCurrentBlockOrNull();
            for (const obj of identities) {
                try {
                    let reqs = yield this.requirementsOfIdentity(obj, current, computeDistance);
                    all.push(reqs);
                }
                catch (e) {
                    this.logger.warn(e);
                }
            }
            return all;
        });
    }
    requirementsOfIdentity(idty, current, computeDistance = true) {
        return __awaiter(this, void 0, void 0, function* () {
            // TODO: this is not clear
            let expired = false;
            let outdistanced = false;
            let isSentry = false;
            let wasMember = false;
            let expiresMS = 0;
            let expiresPending = 0;
            let certs = [];
            let certsPending = [];
            let mssPending = [];
            try {
                const join = yield this.server.generatorGetJoinData(current, idty.hash, 'a');
                const pubkey = join.identity.pubkey;
                // Check WoT stability
                const someNewcomers = join.identity.wasMember ? [] : [join.identity.pubkey];
                const nextBlockNumber = current ? current.number + 1 : 0;
                const joinData = {};
                joinData[join.identity.pubkey] = join;
                const updates = {};
                certsPending = yield this.dal.certDAL.getToTarget(idty.hash);
                certsPending = certsPending.map((c) => {
                    c.blockstamp = [c.block_number, c.block_hash].join('-');
                    return c;
                });
                mssPending = yield this.dal.msDAL.getPendingINOfTarget(idty.hash);
                mssPending = mssPending.map((ms) => {
                    ms.blockstamp = ms.block;
                    ms.sig = ms.signature;
                    ms.type = ms.membership;
                    return ms;
                });
                const newCerts = yield this.server.generatorComputeNewCerts(nextBlockNumber, [join.identity.pubkey], joinData, updates);
                const newLinks = yield this.server.generatorNewCertsToLinks(newCerts, updates);
                const currentTime = current ? current.medianTime : 0;
                certs = yield this.getValidCerts(pubkey, newCerts);
                if (computeDistance) {
                    outdistanced = yield global_rules_1.GLOBAL_RULES_HELPERS.isOver3Hops(pubkey, newLinks, someNewcomers, current, this.conf, this.dal);
                }
                // Expiration of current membershship
                const currentMembership = yield this.dal.mindexDAL.getReducedMS(pubkey);
                const currentMSN = currentMembership ? parseInt(currentMembership.created_on) : -1;
                if (currentMSN >= 0) {
                    if (join.identity.member) {
                        const msBlock = yield this.dal.getBlock(currentMSN);
                        if (msBlock && msBlock.medianTime) {
                            expiresMS = Math.max(0, (msBlock.medianTime + this.conf.msValidity - currentTime));
                        }
                        else {
                            expiresMS = this.conf.msValidity;
                        }
                    }
                    else {
                        expiresMS = 0;
                    }
                }
                // Expiration of pending membership
                const lastJoin = yield this.dal.lastJoinOfIdentity(idty.hash);
                if (lastJoin) {
                    const msBlock = yield this.dal.getBlock(lastJoin.blockNumber);
                    if (msBlock && msBlock.medianTime) {
                        expiresPending = Math.max(0, (msBlock.medianTime + this.conf.msValidity - currentTime));
                    }
                    else {
                        expiresPending = this.conf.msValidity;
                    }
                }
                wasMember = idty.wasMember;
                isSentry = idty.member && (yield this.dal.isSentry(idty.pubkey, this.conf));
                // Expiration of certifications
                for (const cert of certs) {
                    cert.expiresIn = Math.max(0, cert.timestamp + this.conf.sigValidity - currentTime);
                }
            }
            catch (e) {
                // We throw whatever isn't "Too old identity" error
                if (!(e && e.uerr && e.uerr.ucode == constants.ERRORS.TOO_OLD_IDENTITY.uerr.ucode)) {
                    throw e;
                }
                else {
                    expired = true;
                }
            }
            return {
                pubkey: idty.pubkey,
                uid: idty.uid,
                sig: idty.sig,
                meta: {
                    timestamp: idty.buid
                },
                revocation_sig: idty.revocation_sig,
                revoked: idty.revoked,
                revoked_on: idty.revoked_on,
                expired: expired,
                outdistanced: outdistanced,
                isSentry: isSentry,
                wasMember: wasMember,
                certifications: certs,
                pendingCerts: certsPending,
                pendingMemberships: mssPending,
                membershipPendingExpiresIn: expiresPending,
                membershipExpiresIn: expiresMS
            };
        });
    }
    getValidCerts(newcomer, newCerts) {
        return __awaiter(this, void 0, void 0, function* () {
            const links = yield this.dal.getValidLinksTo(newcomer);
            const certsFromLinks = links.map((lnk) => { return { from: lnk.issuer, to: lnk.receiver, timestamp: lnk.expires_on - this.conf.sigValidity }; });
            const certsFromCerts = [];
            const certs = newCerts[newcomer] || [];
            for (const cert of certs) {
                const block = yield this.dal.getBlock(cert.block_number);
                certsFromCerts.push({
                    from: cert.from,
                    to: cert.to,
                    sig: cert.sig,
                    timestamp: block.medianTime
                });
            }
            return certsFromLinks.concat(certsFromCerts);
        });
    }
    isMember() {
        return this.dal.isMember(this.selfPubkey);
    }
    getCountOfSelfMadePoW() {
        return this.dal.getCountOfPoW(this.selfPubkey);
    }
    // This method is called by duniter-crawler 1.3.x
    saveParametersForRootBlock(block) {
        return this.server.blockchain.saveParametersForRoot(block, this.conf, this.dal);
    }
    blocksBetween(from, count) {
        return __awaiter(this, void 0, void 0, function* () {
            if (count > 5000) {
                throw 'Count is too high';
            }
            const current = yield this.current();
            count = Math.min(current.number - from + 1, count);
            if (!current || current.number < from) {
                return [];
            }
            return this.dal.getBlocksBetween(from, from + count - 1);
        });
    }
    /**
     * Allows to quickly insert a bunch of blocks. To reach such speed, this method skips global rules and buffers changes.
     *
     * **This method should be used ONLY when a node is really far away from current blockchain HEAD (i.e several hundreds of blocks late).
     *
     * This method is called by duniter-crawler 1.3.x.
     *
     * @param blocks An array of blocks to insert.
     * @param to The final block number of the fast insertion.
     */
    fastBlockInsertions(blocks, to) {
        return this.mainContext.quickApplyBlocks(blocks, to);
    }
}
exports.BlockchainService = BlockchainService;
//# sourceMappingURL=BlockchainService.js.map