"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const IdentityDAL_1 = require("../lib/dal/sqliteDAL/IdentityDAL");
const global_rules_1 = require("../lib/rules/global_rules");
const BlockDTO_1 = require("../lib/dto/BlockDTO");
const RevocationDTO_1 = require("../lib/dto/RevocationDTO");
const IdentityDTO_1 = require("../lib/dto/IdentityDTO");
const CertificationDTO_1 = require("../lib/dto/CertificationDTO");
const keyring_1 = require("../lib/common-libs/crypto/keyring");
const FIFOService_1 = require("./FIFOService");
"use strict";
const constants = require('../lib/constants');
const BY_ABSORPTION = true;
class IdentityService extends FIFOService_1.FIFOService {
    constructor(fifoPromiseHandler) {
        super(fifoPromiseHandler);
    }
    setConfDAL(newConf, newDAL) {
        this.dal = newDAL;
        this.conf = newConf;
        this.logger = require('../lib/logger').NewLogger(this.dal.profile);
    }
    searchIdentities(search) {
        return this.dal.searchJustIdentities(search);
    }
    findMember(search) {
        return __awaiter(this, void 0, void 0, function* () {
            let idty = null;
            if (search.match(constants.PUBLIC_KEY)) {
                idty = yield this.dal.getWrittenIdtyByPubkey(search);
            }
            else {
                idty = yield this.dal.getWrittenIdtyByUID(search);
            }
            if (!idty) {
                throw constants.ERRORS.NO_MEMBER_MATCHING_PUB_OR_UID;
            }
            const obj = IdentityDAL_1.DBIdentity.copyFromExisting(idty);
            yield this.dal.fillInMembershipsOfIdentity(Promise.resolve(obj));
            return obj;
        });
    }
    findMemberWithoutMemberships(search) {
        return __awaiter(this, void 0, void 0, function* () {
            let idty = null;
            if (search.match(constants.PUBLIC_KEY)) {
                idty = yield this.dal.getWrittenIdtyByPubkey(search);
            }
            else {
                idty = yield this.dal.getWrittenIdtyByUID(search);
            }
            if (!idty) {
                throw constants.ERRORS.NO_MEMBER_MATCHING_PUB_OR_UID;
            }
            return IdentityDAL_1.DBIdentity.copyFromExisting(idty);
        });
    }
    getWrittenByPubkey(pubkey) {
        return this.dal.getWrittenIdtyByPubkey(pubkey);
    }
    getPendingFromPubkey(pubkey) {
        return this.dal.getNonWritten(pubkey);
    }
    submitIdentity(idty, byAbsorption = false) {
        const idtyObj = IdentityDTO_1.IdentityDTO.fromJSONObject(idty);
        const toSave = IdentityDTO_1.IdentityDTO.fromBasicIdentity(idty);
        // Force usage of local currency name, do not accept other currencies documents
        idtyObj.currency = this.conf.currency;
        const createIdentity = idtyObj.rawWithoutSig();
        const hash = idtyObj.getHash();
        return this.pushFIFO(hash, () => __awaiter(this, void 0, void 0, function* () {
            this.logger.info('⬇ IDTY %s %s', idty.pubkey, idty.uid);
            // Check signature's validity
            let verified = keyring_1.verify(createIdentity, idty.sig, idty.pubkey);
            if (!verified) {
                throw constants.ERRORS.SIGNATURE_DOES_NOT_MATCH;
            }
            let existing = yield this.dal.getIdentityByHashOrNull(toSave.hash);
            if (existing) {
                throw constants.ERRORS.ALREADY_UP_TO_DATE;
            }
            else {
                // Create if not already written uid/pubkey
                let used = yield this.dal.getWrittenIdtyByPubkey(idty.pubkey);
                if (used) {
                    throw constants.ERRORS.PUBKEY_ALREADY_USED;
                }
                used = yield this.dal.getWrittenIdtyByUID(idty.uid);
                if (used) {
                    throw constants.ERRORS.UID_ALREADY_USED;
                }
                const current = yield this.dal.getCurrentBlockOrNull();
                if (idty.buid == '0-E3B0C44298FC1C149AFBF4C8996FB92427AE41E4649B934CA495991B7852B855' && current) {
                    throw constants.ERRORS.BLOCKSTAMP_DOES_NOT_MATCH_A_BLOCK;
                }
                else if (current) {
                    let basedBlock = yield this.dal.getBlockByBlockstamp(idty.buid);
                    if (!basedBlock) {
                        throw constants.ERRORS.BLOCKSTAMP_DOES_NOT_MATCH_A_BLOCK;
                    }
                    toSave.expires_on = basedBlock.medianTime + this.conf.idtyWindow;
                }
                yield global_rules_1.GLOBAL_RULES_FUNCTIONS.checkIdentitiesAreWritable({ identities: [idtyObj.inline()], version: (current && current.version) || constants.BLOCK_GENERATED_VERSION }, this.conf, this.dal);
                if (byAbsorption !== BY_ABSORPTION) {
                    if (!(yield this.dal.idtyDAL.sandbox.acceptNewSandBoxEntry({
                        pubkey: idty.pubkey,
                        ref_block: parseInt(idty.buid.split('-')[0])
                    }, this.conf.pair && this.conf.pair.pub))) {
                        throw constants.ERRORS.SANDBOX_FOR_IDENTITY_IS_FULL;
                    }
                }
                yield this.dal.savePendingIdentity(toSave);
                this.logger.info('✔ IDTY %s %s', idty.pubkey, idty.uid);
                return toSave;
            }
        }));
    }
    submitCertification(obj) {
        return __awaiter(this, void 0, void 0, function* () {
            const current = yield this.dal.getCurrentBlockOrNull();
            // Prepare validator for certifications
            const potentialNext = BlockDTO_1.BlockDTO.fromJSONObject({ currency: this.conf.currency, identities: [], number: current ? current.number + 1 : 0 });
            // Force usage of local currency name, do not accept other currencies documents
            obj.currency = this.conf.currency || obj.currency;
            const cert = CertificationDTO_1.CertificationDTO.fromJSONObject(obj);
            const targetHash = cert.getTargetHash();
            let idty = yield this.dal.getIdentityByHashOrNull(targetHash);
            let idtyAbsorbed = false;
            if (!idty) {
                idtyAbsorbed = true;
                idty = yield this.submitIdentity({
                    pubkey: cert.idty_issuer,
                    uid: cert.idty_uid,
                    buid: cert.idty_buid,
                    sig: cert.idty_sig
                }, BY_ABSORPTION);
            }
            let anErr;
            const hash = cert.getHash();
            return this.pushFIFO(hash, () => __awaiter(this, void 0, void 0, function* () {
                this.logger.info('⬇ CERT %s block#%s -> %s', cert.from, cert.block_number, idty.uid);
                try {
                    yield global_rules_1.GLOBAL_RULES_HELPERS.checkCertificationIsValid(cert, potentialNext, () => Promise.resolve(idty), this.conf, this.dal);
                }
                catch (e) {
                    anErr = e;
                }
                if (!anErr) {
                    try {
                        let basedBlock = yield this.dal.getBlock(cert.block_number);
                        if (cert.block_number == 0 && !basedBlock) {
                            basedBlock = {
                                number: 0,
                                hash: 'E3B0C44298FC1C149AFBF4C8996FB92427AE41E4649B934CA495991B7852B855'
                            };
                        }
                        const mCert = {
                            from: cert.from,
                            sig: cert.sig,
                            block_number: cert.block_number,
                            block_hash: basedBlock.hash,
                            target: targetHash,
                            to: idty.pubkey,
                            expires_on: basedBlock.medianTime + this.conf.sigWindow,
                            linked: false,
                            written: false,
                            expired: false,
                            written_block: null,
                            written_hash: null,
                            block: cert.block_number
                        };
                        let existingCert = yield this.dal.existsCert(mCert);
                        if (!existingCert) {
                            if (!(yield this.dal.certDAL.getSandboxForKey(cert.from).acceptNewSandBoxEntry(mCert, this.conf.pair && this.conf.pair.pub))) {
                                throw constants.ERRORS.SANDBOX_FOR_CERT_IS_FULL;
                            }
                            yield this.dal.registerNewCertification(mCert);
                            this.logger.info('✔ CERT %s', mCert.from);
                        }
                        else {
                            throw constants.ERRORS.ALREADY_UP_TO_DATE;
                        }
                    }
                    catch (e) {
                        anErr = e;
                    }
                }
                if (anErr) {
                    if (idtyAbsorbed) {
                        yield this.dal.idtyDAL.deleteByHash(targetHash);
                    }
                    const err = anErr;
                    const errMessage = (err.uerr && err.uerr.message) || err.message || err;
                    this.logger.info('✘ CERT %s %s', cert.from, errMessage);
                    throw anErr;
                }
                return cert;
            }));
        });
    }
    submitRevocation(obj) {
        // Force usage of local currency name, do not accept other currencies documents
        obj.currency = this.conf.currency || obj.currency;
        const revoc = RevocationDTO_1.RevocationDTO.fromJSONObject(obj);
        const raw = revoc.rawWithoutSig();
        const hash = revoc.getHash();
        return this.pushFIFO(hash, () => __awaiter(this, void 0, void 0, function* () {
            try {
                this.logger.info('⬇ REVOCATION %s %s', revoc.pubkey, revoc.idty_uid);
                let verified = keyring_1.verify(raw, revoc.revocation, revoc.pubkey);
                if (!verified) {
                    throw 'Wrong signature for revocation';
                }
                const existing = yield this.dal.getIdentityByHashOrNull(obj.hash);
                if (existing) {
                    // Modify
                    if (existing.revoked) {
                        throw 'Already revoked';
                    }
                    else if (existing.revocation_sig) {
                        throw 'Revocation already registered';
                    }
                    else {
                        yield this.dal.setRevocating(existing, revoc.revocation);
                        this.logger.info('✔ REVOCATION %s %s', revoc.pubkey, revoc.idty_uid);
                        return revoc;
                    }
                }
                else {
                    // Create identity given by the revocation
                    const idty = IdentityDTO_1.IdentityDTO.fromRevocation(revoc);
                    idty.revocation_sig = revoc.revocation;
                    if (!(yield this.dal.idtyDAL.sandbox.acceptNewSandBoxEntry({
                        pubkey: idty.pubkey,
                        ref_block: parseInt(idty.buid.split('-')[0]),
                        certsCount: 0
                    }, this.conf.pair && this.conf.pair.pub))) {
                        throw constants.ERRORS.SANDBOX_FOR_IDENTITY_IS_FULL;
                    }
                    yield this.dal.savePendingIdentity(idty);
                    this.logger.info('✔ REVOCATION %s %s', revoc.pubkey, revoc.idty_uid);
                    return revoc;
                }
            }
            catch (e) {
                this.logger.info('✘ REVOCATION %s %s', revoc.pubkey, revoc.idty_uid);
                throw e;
            }
        }));
    }
}
exports.IdentityService = IdentityService;
//# sourceMappingURL=IdentityService.js.map