"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const assert = require("assert");
const Switcher_1 = require("../../app/lib/blockchain/Switcher");
const logger_1 = require("../../app/lib/logger");
const logger = logger_1.NewLogger();
const avgGenTime = 5 * 60;
const forkWindowSize = 5;
const switchOnHeadAdvance = 3;
describe("Fork resolution 3-3 algo", () => {
    it('should switch on a valid fork', () => __awaiter(this, void 0, void 0, function* () {
        // B10 -- B11 -- B12 -- B13
        //            `- C12 -- C13 -- C14 -- C15 -- C16
        const bc = new Blockchain(Block.from("B10"));
        bc.add(Block.from("B11"));
        bc.add(Block.from("B12"));
        bc.add(Block.from("B13"));
        assert.equal(bc.current.number, 13);
        const sbx = new BlockSandbox([
            Block.from("C12", "B11"),
            Block.from("C13"),
            Block.from("C14"),
            Block.from("C15"),
            Block.from("C16")
        ]);
        const switcher = new Switcher_1.Switcher(new TestingSwitcherDao(bc, sbx), [], avgGenTime, forkWindowSize, switchOnHeadAdvance, logger);
        yield switcher.tryToFork();
        assert.equal(bc.current.number, 16);
        assert.equal(bc.current.hash, "C16");
    }));
    it('should not switch if no fork block 3-3 exist', () => __awaiter(this, void 0, void 0, function* () {
        // B10 -- B11 -- B12 -- B13
        //            `- C12 -- C13 -- C14 -- C15
        const bc = new Blockchain(Block.from("B10"));
        bc.add(Block.from("B11"));
        bc.add(Block.from("B12"));
        bc.add(Block.from("B13"));
        assert.equal(bc.current.number, 13);
        const sbx = new BlockSandbox([
            Block.from("C12", "B11"),
            Block.from("C13"),
            Block.from("C14"),
            Block.from("C15")
        ]);
        const switcher = new Switcher_1.Switcher(new TestingSwitcherDao(bc, sbx), [], avgGenTime, forkWindowSize, switchOnHeadAdvance);
        yield switcher.tryToFork();
        assert.equal(bc.current.number, 13);
        assert.equal(bc.current.hash, "B13");
    }));
    it('should eliminate a fork with missing blocks', () => __awaiter(this, void 0, void 0, function* () {
        // B10 -- B11 -- B12 -- B13
        //            `- C12 -- C13 -- C14 -- C15 -- C16
        const bc = new Blockchain(Block.from("B10"));
        bc.add(Block.from("B11"));
        bc.add(Block.from("B12"));
        bc.add(Block.from("B13"));
        assert.equal(bc.current.number, 13);
        const sbx = new BlockSandbox([
            Block.from("C14"),
            Block.from("C15")
        ]);
        const switcher = new Switcher_1.Switcher(new TestingSwitcherDao(bc, sbx), [], avgGenTime, forkWindowSize, switchOnHeadAdvance);
        yield switcher.tryToFork();
        assert.equal(bc.current.number, 13);
        assert.equal(bc.current.hash, "B13");
    }));
    it('should eliminate a fork out of fork window', () => __awaiter(this, void 0, void 0, function* () {
        // B10 -- B11 -- B12 -- B13
        //     `- C11 -- C12 -- C13 -- C14 -- C15 -- C16
        const bc = new Blockchain(Block.from("B10"));
        bc.add(Block.from("B11"));
        bc.add(Block.from("B12"));
        bc.add(Block.from("B13"));
        assert.equal(bc.current.number, 13);
        const sbx = new BlockSandbox([
            Block.from("C10"),
            Block.from("C11"),
            Block.from("C12"),
            Block.from("C13"),
            Block.from("C14"),
            Block.from("C15"),
            Block.from("C16")
        ]);
        const switcher = new Switcher_1.Switcher(new TestingSwitcherDao(bc, sbx), [], avgGenTime, forkWindowSize, switchOnHeadAdvance);
        yield switcher.tryToFork();
        assert.equal(bc.current.number, 13);
        assert.equal(bc.current.hash, "B13");
    }));
    it('should refuse a fork right on the limit of the fork window', () => __awaiter(this, void 0, void 0, function* () {
        // B10 -- B11 -- B12 -- B13
        //     `- C11 -- C12 -- C13 -- C14 -- C15 -- C16
        const bc = new Blockchain(Block.from("B10"));
        bc.add(Block.from("B11"));
        bc.add(Block.from("B12"));
        bc.add(Block.from("B13"));
        assert.equal(bc.current.number, 13);
        const sbx = new BlockSandbox([
            Block.from("C11", "B10"),
            Block.from("C12"),
            Block.from("C13"),
            Block.from("C14"),
            Block.from("C15"),
            Block.from("C16")
        ]);
        const switcher = new Switcher_1.Switcher(new TestingSwitcherDao(bc, sbx), [], avgGenTime, forkWindowSize, switchOnHeadAdvance);
        yield switcher.tryToFork();
        assert.equal(bc.current.number, 13);
        assert.equal(bc.current.hash, "B13");
    }));
    it('should eliminate a fork whose 2nd block is invalid', () => __awaiter(this, void 0, void 0, function* () {
        // B10 -- B11 -- B12 -- B13
        //            `- C12 -- C13 -- C14 -- C15 -- C16
        const bc = new Blockchain(Block.from("B10"));
        bc.add(Block.from("B11"));
        bc.add(Block.from("B12"));
        bc.add(Block.from("B13"));
        assert.equal(bc.current.number, 13);
        const sbx = new BlockSandbox([
            Block.from("C12", "B11"),
            Block.from("C13"),
            Block.from("C14", "", () => false),
            Block.from("C15"),
            Block.from("C16")
        ]);
        const switcher = new Switcher_1.Switcher(new TestingSwitcherDao(bc, sbx), [], avgGenTime, forkWindowSize, switchOnHeadAdvance);
        yield switcher.tryToFork();
        assert.equal(bc.current.number, 13);
        assert.equal(bc.current.hash, "B13");
    }));
    it('should select the longest fork', () => __awaiter(this, void 0, void 0, function* () {
        // B10 -- B11 -- B12 -- B13 -- B14
        //           `              `- C14 -- C15          <-- "length" 2
        //           |                     `- D15 -- D16   <-- "length" 3 (should be selected)
        //            `- E12 -- E13 -- E14 -- E15          <-- "length" 4
        const bc = new Blockchain(Block.from("B10"));
        bc.add(Block.from("B11"));
        bc.add(Block.from("B12"));
        bc.add(Block.from("B13"));
        bc.add(Block.from("B14"));
        assert.equal(bc.current.number, 14);
        const sbx = new BlockSandbox([
            Block.from("C14", "B13"),
            Block.from("C15"),
            Block.from("D15", "C14"),
            Block.from("D16"),
            Block.from("E12", "B11"),
            Block.from("E13"),
            Block.from("E14"),
            Block.from("E15")
        ]);
        const switcher = new Switcher_1.Switcher(new TestingSwitcherDao(bc, sbx), [], avgGenTime, forkWindowSize, 1);
        yield switcher.tryToFork();
        assert.equal(16, bc.current.number);
        assert.equal("D16", bc.current.hash);
    }));
});
class TestingSwitcherDao {
    constructor(bc, sbx) {
        this.bc = bc;
        this.sbx = sbx;
    }
    getCurrent() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.bc.current;
        });
    }
    getPotentials(numberStart, timeStart) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.sbx.getPotentials(numberStart, timeStart);
        });
    }
    getBlockchainBlock(number, hash) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.bc.getBlock(number, hash);
        });
    }
    getSandboxBlock(number, hash) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.sbx.getBlock(number, hash);
        });
    }
    revertTo(number) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.bc.revertTo(number);
        });
    }
    addBlock(block) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.bc.add(block);
        });
    }
}
/**
 * A super simple sandbox for new blocks.
 */
class BlockSandbox {
    constructor(blocks = []) {
        this.blocks = blocks;
    }
    /**
     * Gets a particular block.
     * @param number The block number.
     * @param {hash} hash The block hash.
     * @returns The block or null if it was not found.
     */
    getBlock(number, hash) {
        for (const b of this.blocks) {
            if (b.number === number && b.hash === hash) {
                return b;
            }
        }
        return null;
    }
    /**
     * Retrieves all the candidate blocks for the switch.
     * @param numberStart Will pick blocks whose number >= numberStart
     * @param timeStart
     * @returns The candidate blocks.
     */
    getPotentials(numberStart, timeStart) {
        const potentials = [];
        for (const b of this.blocks) {
            if (b.number >= numberStart && b.medianTime >= timeStart) {
                potentials.push(b);
            }
        }
        return potentials;
    }
}
/**
 * A super simple blockchain
 */
class Blockchain {
    /**
     * The root block of the blockchain (does not need to have number `0`).
     * @param {Block} rootBlock
     */
    constructor(rootBlock) {
        this.blocks = [];
        this.blocks.push(rootBlock);
    }
    /**
     * Returns the current block (HEAD) of the blockchain.
     * @returns {Block}
     */
    get current() {
        return this.blocks[this.blocks.length - 1];
    }
    /**
     * Adds a block on top of HEAD.
     * @param {Block} block
     * @returns {Block}
     */
    add(block) {
        if (!block.chainsOn(this.current)) {
            throw "Unchainable";
        }
        this.blocks.push(block);
        return block;
    }
    /**
     * Gets a particular block.
     * @param number The block number.
     * @param hash The block hash.
     * @returns The block or null if it was not found.
     */
    getBlock(number, hash) {
        for (const b of this.blocks) {
            if (b.number === number && b.hash === hash) {
                return b;
            }
        }
        return null;
    }
    /**
     * Pops blocks from HEAD to HEAD - number.
     * @param number The block number that will be our new HEAD.
     * @returns {Block[]}
     */
    revertTo(number) {
        const reverted = [];
        if (this.current.number < number) {
            throw "Already below this number";
        }
        while (this.current.number > number) {
            const poped = this.blocks.pop();
            reverted.push(poped);
        }
        return reverted;
    }
}
class Block {
    constructor(chain, number, thePreviousHash, chainsOnHook = () => true) {
        this.chain = chain;
        this.number = number;
        this.thePreviousHash = thePreviousHash;
        this.chainsOnHook = chainsOnHook;
    }
    get hash() {
        return [this.chain, this.number].join('');
    }
    get medianTime() {
        return this.number * avgGenTime;
    }
    get previousHash() {
        return this.thePreviousHash || [this.chain, this.number - 1].join('');
    }
    chainsOn(previous) {
        return this.number === previous.number + 1 && (this.chain === previous.chain || this.previousHash === previous.hash) && this.chainsOnHook(previous);
    }
    static from(hash, previousHash = "", chainsOnHook = undefined) {
        const match = hash.match(/([A-Z])(\d+)/);
        const chain = match && match[1] || "";
        const number = parseInt(match && match[2] || "0");
        return new Block(chain, number, previousHash, chainsOnHook);
    }
}
//# sourceMappingURL=fork-resolution-3-3.js.map