"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const pulling_1 = require("../../../../app/modules/crawler/lib/pulling");
const logger_1 = require("../../../../app/lib/logger");
const should = require('should');
const _ = require('underscore');
let commonConf = {
    switchOnHeadAdvance: 1,
    avgGenTime: 30 * 60,
    forksize: 100
};
describe('Pulling blocks', () => {
    it('from genesis with good sidechain should work', pullinTest({
        blockchain: [
            newBlock(0, 'A')
        ],
        sidechains: [
            [
                newBlock(0, 'A'),
                newBlock(1, 'A') // <-- 1) checks this block: is good, we add it
            ]
        ],
        expectHash: 'A1'
    }));
    it('from genesis with fork sidechain should not work', pullinTest({
        blockchain: [
            newBlock(0, 'A')
        ],
        sidechains: [
            [
                newBlock(0, 'B'),
                newBlock(1, 'B') // <-- 1) checks this block: ah, a fork! let's find common root ...
            ]
        ],
        expectHash: 'A0'
    }));
    it('from genesis with multiple good sidechains should work', pullinTest({
        blockchain: [
            newBlock(0, 'A')
        ],
        sidechains: [
            [
                newBlock(0, 'A'),
                newBlock(1, 'A'),
                newBlock(2, 'A') // <-- 2) checks this block: is good, we add it
            ],
            [
                newBlock(0, 'A'),
                newBlock(1, 'A') // <-- 3) you are a bit late ... we are on A2 yet!
            ],
            [
                newBlock(0, 'A'),
                newBlock(1, 'A'),
                newBlock(2, 'A'),
                newBlock(3, 'A') // <-- 4) checks this block: is good, we add it
            ],
            [
                newBlock(0, 'A'),
                newBlock(1, 'A') // <-- 5 really too late
            ]
        ],
        expectHash: 'A3'
    }));
    it('sync with a single fork', pullinTest({
        blockchain: [
            newBlock(0, 'A'),
            newBlock(1, 'A'),
            newBlock(2, 'A'),
            newBlock(3, 'A')
        ],
        sidechains: [
            [
                newBlock(0, 'A'),
                newBlock(1, 'A'),
                newBlock(2, 'B'),
                newBlock(3, 'B'),
                newBlock(4, 'B'),
                newBlock(5, 'B')
            ]
        ],
        expectHash: 'B5'
    }));
    it('sync with multiple forks', pullinTest({
        blockchain: [
            newBlock(0, 'A'),
            newBlock(1, 'A'),
            newBlock(2, 'A'),
            newBlock(3, 'A')
        ],
        sidechains: [
            [
                newBlock(0, 'A'),
                newBlock(1, 'A'),
                newBlock(2, 'B'),
                newBlock(3, 'B'),
                newBlock(4, 'B'),
                newBlock(5, 'B')
            ],
            // This fork should not be followed because we switch only one time per pulling, and B5 is already OK
            [
                newBlock(0, 'A'),
                newBlock(1, 'A'),
                newBlock(2, 'B'),
                newBlock(3, 'B'),
                newBlock(4, 'B'),
                newBlock(5, 'B'),
                newBlock(6, 'B')
            ]
        ],
        expectHash: 'B5'
    }));
    it('sync with inconsistant fork should skip it', pullinTest({
        blockchain: [
            newBlock(0, 'A'),
            newBlock(1, 'A'),
            newBlock(2, 'A'),
            newBlock(3, 'A')
        ],
        sidechains: [
            [
                newBlock(0, 'A'),
                qwaBlock(1, 'A'),
                newBlock(2, 'C'),
                newBlock(3, 'C'),
                newBlock(4, 'C'),
                newBlock(5, 'C')
            ]
        ],
        expectHash: 'A3'
    }));
});
function newBlock(number, branch, rootBranch = null, quantum = false) {
    let previousNumber = number - 1;
    let previousBranch = rootBranch || branch;
    let previousHash = previousNumber >= 0 ? previousBranch + previousNumber : '';
    return {
        number: number,
        medianTime: number * 30 * 60,
        hash: branch + number,
        previousHash: previousHash,
        // this is not a real field, just here for the sake of demonstration: a quantum block changes itself
        // when we consult it, making the chain inconsistent
        quantum: quantum
    };
}
function qwaBlock(number, branch, rootBranch = null) {
    return newBlock(number, branch, rootBranch, true);
}
function pullinTest(testConfiguration) {
    return () => __awaiter(this, void 0, void 0, function* () {
        // The blockchains we are testing against
        let blockchain = testConfiguration.blockchain;
        let sidechains = testConfiguration.sidechains;
        // The data access object simulating network access
        let dao = new mockDao(blockchain, sidechains);
        // The very last block of a blockchain should have the good number
        const local = yield dao.localCurrent();
        local.should.have.property('number').equal(blockchain[blockchain.length - 1].number);
        // And after we make a pulling...
        yield dao.pull(commonConf, logger_1.NewLogger());
        // We test the new local blockchain current block (it should have changed in case of successful pull)
        let localCurrent = yield dao.localCurrent();
        if (testConfiguration.expectHash !== undefined && testConfiguration.expectHash !== null) {
            localCurrent.should.have.property('hash').equal(testConfiguration.expectHash);
        }
        if (testConfiguration.expectFunc !== undefined && testConfiguration.expectFunc !== null) {
            testConfiguration.expectFunc(dao);
        }
    });
}
/**
 * Network mocker
 * @param blockchain
 * @param sideChains
 * @returns DAO
 */
class mockDao extends pulling_1.AbstractDAO {
    constructor(blockchain, sideChains) {
        super();
        this.blockchain = blockchain;
        this.sideChains = sideChains;
    }
    // Get the local blockchain current block
    localCurrent() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.blockchain[this.blockchain.length - 1];
        });
    }
    // Get the remote blockchain (bc) current block
    remoteCurrent(bc) {
        return __awaiter(this, void 0, void 0, function* () {
            return bc[bc.length - 1];
        });
    }
    // Get the remote peers to be pulled
    remotePeers() {
        return Promise.resolve(this.sideChains.map((sc, index) => {
            sc.pubkey = 'PUBK' + index;
            return sc;
        }));
    }
    // Get block of given peer with given block number
    getLocalBlock(number) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.blockchain[number] || null;
        });
    }
    // Get block of given peer with given block number
    getRemoteBlock(bc, number) {
        return __awaiter(this, void 0, void 0, function* () {
            let block = bc[number] || null;
            // Quantum block implementation
            if (block && block.quantum) {
                bc[number] = _.clone(block);
                bc[number].hash = 'Q' + block.hash;
            }
            return block;
        });
    }
    // Simulate the adding of a single new block on local blockchain
    applyMainBranch(block) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.blockchain.push(block);
        });
    }
    // Clean the eventual fork blocks already registered in DB (since real fork mechanism uses them, so we want
    // every old fork block to be removed)
    removeForks() {
        return __awaiter(this, void 0, void 0, function* () {
            return true;
        });
    }
    // Tells wether given peer is a member peer
    isMemberPeer() {
        return __awaiter(this, void 0, void 0, function* () {
            return true;
        });
    }
    // Simulates the downloading of blocks from a peer
    downloadBlocks(bc, fromNumber, count) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!count) {
                const block = yield this.getRemoteBlock(bc, fromNumber);
                if (block) {
                    return [block];
                }
                else {
                    return [];
                }
            }
            return bc.slice(fromNumber, fromNumber + count);
        });
    }
}
//# sourceMappingURL=block_pulling.js.map