"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const WS2PConnection_1 = require("../../app/modules/ws2p/lib/WS2PConnection");
const keyring_1 = require("../../app/lib/common-libs/crypto/keyring");
const toolbox_1 = require("./tools/toolbox");
const constants_1 = require("../../app/modules/ws2p/lib/constants");
const assert = require('assert');
const WebSocketServer = require('ws').Server;
const logger = require('../../app/lib/logger').NewLogger('ws2p');
const gtest = "gtest";
describe('WS2P', () => {
    constants_1.WS2PConstants.CONNEXION_TIMEOUT = 100;
    constants_1.WS2PConstants.REQUEST_TIMEOUT = 100;
    describe("WS2P client connection", function () {
        describe("no auth", () => {
            let wss, portA;
            before(() => __awaiter(this, void 0, void 0, function* () {
                portA = toolbox_1.getNewTestingPort();
                wss = new WebSocketServer({ port: portA });
                wss.on('connection', (ws) => {
                    ws.on('message', (data) => {
                        const obj = JSON.parse(data);
                        if (obj.reqId) {
                            ws.send(JSON.stringify({ resId: obj.reqId, body: { bla: 'aa' } }));
                        }
                    });
                });
            }));
            after((done) => {
                wss.close(done);
            });
            it('should be able to create a connection', () => __awaiter(this, void 0, void 0, function* () {
                const ws2p = WS2PConnection_1.WS2PConnection.newConnectionToAddress('ws://localhost:' + portA, new WS2PMutedHandler(), new WS2PNoLocalAuth(), new WS2PNoRemoteAuth());
                const res = yield ws2p.request({ name: 'head' });
                assert.deepEqual({ bla: 'aa' }, res);
            }));
        });
        describe("pubkey auth", () => {
            let wss, clientAskError = "";
            before(() => __awaiter(this, void 0, void 0, function* () {
                const serverKeypair = new keyring_1.Key('DKpQPUL4ckzXYdnDRvCRKAm1gNvSdmAXnTrJZ7LvM5Qo', '64EYRvdPpTfLGGmaX5nijLXRqWXaVz8r1Z1GtaahXwVSJGQRn7tqkxLb288zwSYzELMEG5ZhXSBYSxsTsz1m9y8F');
                let nbAsk = 0;
                wss = new WebSocketServer({ port: 20903 });
                wss.on('connection', (ws) => {
                    ws.on('message', (data) => {
                        const obj = JSON.parse(data);
                        if (obj.reqId) {
                            ws.send(JSON.stringify({ resId: obj.reqId, body: { bla: 'aa' } }));
                        }
                        if (obj.auth) {
                            if (nbAsk == 1 || nbAsk == 3) {
                                const challengeMessage = `WS2P:ACK:gtest:${serverKeypair.pub}:${obj.challenge}`;
                                const sig = serverKeypair.signSync(challengeMessage);
                                if (nbAsk == 1) {
                                    ws.send(JSON.stringify({ auth: 'ACK', pub: serverKeypair.pub, sig: 'hiohoihio' }));
                                }
                                if (nbAsk == 3) {
                                    ws.send(JSON.stringify({ auth: 'ACK', pub: serverKeypair.pub, sig }));
                                }
                            }
                            if (nbAsk == 2) {
                                // We do like if the key was wrong
                                const clientPub = 'GgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd';
                                const challengeMessage = `WS2P:CONNECT:${clientPub}:${obj.challenge}`;
                                if (!keyring_1.verify(challengeMessage, obj.sig, clientPub)) {
                                    clientAskError = 'Wrong signature from client CONNECT';
                                }
                            }
                            nbAsk++;
                        }
                    });
                });
            }));
            after((done) => {
                wss.close(done);
            });
            it('should refuse the connection if the server does not answer', () => __awaiter(this, void 0, void 0, function* () {
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const ws2p = WS2PConnection_1.WS2PConnection.newConnectionToAddress('ws://localhost:20903', new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, keypair), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, keypair), undefined, {
                    connectionTimeout: 100,
                    requestTimeout: 100
                });
                yield toolbox_1.assertThrows(ws2p.request({ name: 'a' }), "WS2P connection timeout");
            }));
            it('should refuse the connection if the server answers with a wrong signature', () => __awaiter(this, void 0, void 0, function* () {
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const ws2p = WS2PConnection_1.WS2PConnection.newConnectionToAddress('ws://localhost:20903', new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, keypair), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, keypair), undefined, {
                    connectionTimeout: 100,
                    requestTimeout: 100
                });
                yield toolbox_1.assertThrows(ws2p.request({ name: 'a' }), "Wrong signature from server ACK");
            }));
            it('should refuse the connection if the server refuses our signature', () => __awaiter(this, void 0, void 0, function* () {
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const ws2p = WS2PConnection_1.WS2PConnection.newConnectionToAddress('ws://localhost:20903', new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, keypair), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, keypair), undefined, {
                    connectionTimeout: 100,
                    requestTimeout: 100
                });
                yield toolbox_1.assertThrows(ws2p.request({ name: 'a' }), "WS2P connection timeout");
                assert.equal('Wrong signature from client CONNECT', clientAskError);
            }));
            it('should accept the connection if the server answers with a good signature', () => __awaiter(this, void 0, void 0, function* () {
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const ws2p = WS2PConnection_1.WS2PConnection.newConnectionToAddress('ws://localhost:20903', new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, keypair), new WS2PNoRemoteAuth(), undefined, {
                    connectionTimeout: 1000,
                    requestTimeout: 1000
                });
                const res = yield ws2p.request({ name: 'head' });
                assert.deepEqual({ bla: 'aa' }, res);
            }));
        });
    });
    describe("WS2P server connection", function () {
        describe("no auth", () => {
            let wss;
            let s1;
            let s2;
            let portB;
            before(() => __awaiter(this, void 0, void 0, function* () {
                let i = 0;
                portB = toolbox_1.getNewTestingPort();
                wss = new WebSocketServer({ port: portB });
                wss.on('connection', (ws) => {
                    switch (i) {
                        case 0:
                            s1 = WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, new (class TmpHandler {
                                handlePushMessage(json) {
                                    return __awaiter(this, void 0, void 0, function* () {
                                    });
                                }
                                answerToRequest(json) {
                                    return __awaiter(this, void 0, void 0, function* () {
                                        return { answer: 'world' };
                                    });
                                }
                            }), new WS2PNoLocalAuth(), new WS2PNoRemoteAuth());
                            s1.connect().catch(e => logger.error('WS2P: newConnectionFromWebSocketServer connection error'));
                            break;
                        case 1:
                            let j = 0;
                            s2 = WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, new (class TmpHandler {
                                handlePushMessage(json) {
                                    return __awaiter(this, void 0, void 0, function* () {
                                    });
                                }
                                answerToRequest(json) {
                                    return __awaiter(this, void 0, void 0, function* () {
                                        return { answer: 'this is s2![j = ' + (j++) + ']' };
                                    });
                                }
                            }), new WS2PNoLocalAuth(), new WS2PNoRemoteAuth());
                            s2.connect().catch(e => logger.error('WS2P: newConnectionFromWebSocketServer connection error'));
                            break;
                    }
                    i++;
                });
            }));
            after((done) => {
                wss.close(done);
            });
            it('should be able to create connections and make several requests', () => __awaiter(this, void 0, void 0, function* () {
                // connection 1
                const c1 = WS2PConnection_1.WS2PConnection.newConnectionToAddress('ws://localhost:' + portB, new WS2PMutedHandler(), new WS2PNoLocalAuth(), new WS2PNoRemoteAuth());
                assert.deepEqual({ answer: 'world' }, yield c1.request({ name: 'hello!' }));
                assert.deepEqual({ answer: 'world' }, yield c1.request({ name: 'hello2!' }));
                assert.equal(s1.nbRequests, 0);
                assert.equal(c1.nbRequests, 2);
                assert.equal(s1.nbResponses, 0);
                assert.equal(c1.nbResponses, 2);
                assert.equal(s1.nbPushsToRemote, 0);
                assert.equal(c1.nbPushsToRemote, 0);
                assert.equal(s1.nbPushsByRemote, 0);
                assert.equal(c1.nbPushsByRemote, 0);
                // connection 2
                const c2 = WS2PConnection_1.WS2PConnection.newConnectionToAddress('ws://localhost:' + portB, new WS2PMutedHandler(), new WS2PNoLocalAuth(), new WS2PNoRemoteAuth());
                assert.deepEqual({ answer: 'this is s2![j = 0]' }, yield c2.request({ name: 'test?' }));
                assert.deepEqual({ answer: 'this is s2![j = 1]' }, yield c2.request({ name: 'test!' }));
                assert.deepEqual({ answer: 'this is s2![j = 2]' }, yield c2.request({ name: 'test!!!' }));
                assert.equal(s1.nbRequests, 0);
                assert.equal(c2.nbRequests, 3);
                assert.equal(s1.nbResponses, 0);
                assert.equal(c2.nbResponses, 3);
                assert.equal(s1.nbPushsToRemote, 0);
                assert.equal(c2.nbPushsToRemote, 0);
                assert.equal(s1.nbPushsByRemote, 0);
                assert.equal(c2.nbPushsByRemote, 0);
            }));
        });
        describe("pubkey auth", () => {
            let wss;
            let resolveS1;
            let resolveS2;
            let resolveS3;
            let resolveS4;
            let resolveS5;
            let resolveS6;
            let s1p = new Promise(res => resolveS1 = res);
            let s2p = new Promise(res => resolveS2 = res);
            let s3p = new Promise(res => resolveS3 = res);
            let s4p = new Promise(res => resolveS4 = res);
            let s5p = new Promise(res => resolveS5 = res);
            let s6p = new Promise(res => resolveS6 = res);
            before(() => __awaiter(this, void 0, void 0, function* () {
                let i = 1;
                wss = new WebSocketServer({ port: 20903 });
                const serverKeypair = new keyring_1.Key('DKpQPUL4ckzXYdnDRvCRKAm1gNvSdmAXnTrJZ7LvM5Qo', '64EYRvdPpTfLGGmaX5nijLXRqWXaVz8r1Z1GtaahXwVSJGQRn7tqkxLb288zwSYzELMEG5ZhXSBYSxsTsz1m9y8F');
                wss.on('connection', (ws) => __awaiter(this, void 0, void 0, function* () {
                    switch (i) {
                        case 1:
                            resolveS1(WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, serverKeypair), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, serverKeypair), {
                                connectionTimeout: 100,
                                requestTimeout: 100
                            }));
                            (yield s1p).connect().catch((e) => logger.error('WS2P: newConnectionFromWebSocketServer connection error'));
                            break;
                        case 2:
                            class WS2PPubkeyNotAnsweringWithOKAuth extends WS2PConnection_1.WS2PPubkeyRemoteAuth {
                                registerOK(sig) {
                                    return __awaiter(this, void 0, void 0, function* () {
                                        return Promise.resolve(true);
                                    });
                                }
                            }
                            resolveS2(WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, serverKeypair), new WS2PPubkeyNotAnsweringWithOKAuth(gtest, serverKeypair), {
                                connectionTimeout: 100,
                                requestTimeout: 100
                            }));
                            (yield s2p).connect().catch((e) => logger.error('WS2P: newConnectionFromWebSocketServer connection error'));
                            break;
                        case 3:
                            resolveS3(WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, serverKeypair), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, serverKeypair), {
                                connectionTimeout: 100,
                                requestTimeout: 100
                            }));
                            (yield s3p).connect().catch((e) => logger.error('WS2P: newConnectionFromWebSocketServer connection error'));
                            break;
                        case 4:
                            resolveS4(WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, serverKeypair), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, serverKeypair), {
                                connectionTimeout: 100,
                                requestTimeout: 100
                            }));
                            (yield s4p).connect().catch((e) => logger.error('WS2P: newConnectionFromWebSocketServer connection error'));
                            break;
                        case 5:
                            resolveS5(WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, serverKeypair), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, serverKeypair)));
                            (yield s5p).connect().catch((e) => logger.error('WS2P: newConnectionFromWebSocketServer connection error'));
                            break;
                        case 6:
                            resolveS6(WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, serverKeypair), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, serverKeypair), {
                                connectionTimeout: 100,
                                requestTimeout: 100
                            }));
                            (yield s6p).connect().catch((e) => logger.error('WS2P: newConnectionFromWebSocketServer connection error'));
                            break;
                    }
                    i++;
                }));
            }));
            after((done) => {
                wss.close(done);
            });
            it('should refuse the connection if the client does not send ACK', () => __awaiter(this, void 0, void 0, function* () {
                class WS2PPubkeyNotAnsweringWithACKAuth extends WS2PConnection_1.WS2PPubkeyRemoteAuth {
                    sendACK(ws) {
                        return __awaiter(this, void 0, void 0, function* () {
                            return Promise.resolve();
                        });
                    }
                }
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const c1 = WS2PConnection_1.WS2PConnection.newConnectionToAddress('ws://localhost:20903', new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, keypair), new WS2PPubkeyNotAnsweringWithACKAuth(gtest, keypair));
                c1.connect().catch((e) => logger.error('WS2P: connection error'));
                const s1 = yield s1p;
                yield toolbox_1.assertThrows(s1.request({ name: 'something' }), "WS2P connection timeout");
            }));
            it('should refuse the connection if the client not confirm with OK', () => __awaiter(this, void 0, void 0, function* () {
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                WS2PConnection_1.WS2PConnection.newConnectionToAddress('ws://localhost:20903', new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, keypair), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, keypair));
                const s2 = yield s2p;
                yield toolbox_1.assertThrows(s2.request({ name: 'something' }), "WS2P connection timeout");
            }));
            it('should refuse the connection if the client answers with a wrong signature', () => __awaiter(this, void 0, void 0, function* () {
                class WS2PPubkeyAnsweringWithWrongSigForACK extends WS2PConnection_1.WS2PPubkeyRemoteAuth {
                    sendACK(ws) {
                        return __awaiter(this, void 0, void 0, function* () {
                            const challengeMessage = `WS2P:WRONG:${this.pair.pub}:${this.challenge}`;
                            const sig = this.pair.signSync(challengeMessage);
                            yield ws.send(JSON.stringify({
                                auth: 'ACK',
                                pub: this.pair.pub,
                                sig
                            }));
                        });
                    }
                }
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const c3 = WS2PConnection_1.WS2PConnection.newConnectionToAddress('ws://localhost:20903', new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, keypair), new WS2PPubkeyAnsweringWithWrongSigForACK(gtest, keypair));
                c3.connect().catch((e) => logger.error('WS2P: connection error'));
                const s3 = yield s3p;
                yield toolbox_1.assertThrows(s3.request({ name: 'something' }), "Wrong signature from server ACK");
            }));
            it('should refuse the connection if the client refuses our signature', () => __awaiter(this, void 0, void 0, function* () {
                class WS2PPubkeyRefusingACKSignature extends WS2PConnection_1.WS2PPubkeyLocalAuth {
                    registerACK(sig, pub) {
                        return __awaiter(this, void 0, void 0, function* () {
                            const challengeMessage = `WS2P:BLABLA:${pub}:${this.challenge}`;
                            this.authenticated = keyring_1.verify(challengeMessage, sig, pub);
                            if (!this.authenticated) {
                                this.serverAuthReject("Wrong signature from server ACK");
                            }
                            else {
                                this.serverAuthResolve();
                            }
                            return this.authenticated;
                        });
                    }
                }
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const c4 = WS2PConnection_1.WS2PConnection.newConnectionToAddress('ws://localhost:20903', new WS2PMutedHandler(), new WS2PPubkeyRefusingACKSignature(gtest, keypair), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, keypair));
                const s4 = yield s4p;
                yield toolbox_1.assertThrows(c4.connect(), "Wrong signature from server ACK");
            }));
            it('should accept the connection if everything is OK on both side', () => __awaiter(this, void 0, void 0, function* () {
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const c5 = WS2PConnection_1.WS2PConnection.newConnectionToAddress('ws://localhost:20903', new (class TmpHandler {
                    handlePushMessage(json) {
                        return __awaiter(this, void 0, void 0, function* () {
                        });
                    }
                    answerToRequest(json) {
                        return __awaiter(this, void 0, void 0, function* () {
                            return { answer: 'success!' };
                        });
                    }
                }), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, keypair), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, keypair));
                yield c5.connect().catch((e) => logger.error('WS2P: connection error'));
                const s5 = yield s5p;
                assert.deepEqual({ answer: 'success!' }, yield s5.request({ name: 'connection?' }));
            }));
            it('should refuse the connection if the client does not send OK', () => __awaiter(this, void 0, void 0, function* () {
                class WS2PPubkeyNotAnsweringWithOKAuth extends WS2PConnection_1.WS2PPubkeyLocalAuth {
                    sendOK(ws) {
                        return __awaiter(this, void 0, void 0, function* () {
                        });
                    }
                }
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const c6 = WS2PConnection_1.WS2PConnection.newConnectionToAddress('ws://localhost:20903', new WS2PMutedHandler(), new WS2PPubkeyNotAnsweringWithOKAuth(gtest, keypair), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, keypair));
                c6.connect().catch((e) => logger.error('WS2P: connection error'));
                const s6 = yield s6p;
                yield toolbox_1.assertThrows(s6.request({ name: 'something' }), "WS2P connection timeout");
            }));
        });
    });
});
/**************
 * TEST CLASSES
 *************/
class WS2PNoLocalAuth {
    sendCONNECT(ws) {
        return __awaiter(this, void 0, void 0, function* () {
        });
    }
    registerACK(sig, pub) {
        return __awaiter(this, void 0, void 0, function* () {
            return true;
        });
    }
    isRemoteAuthenticated() {
        return true;
    }
    isAuthorizedPubkey(pub) {
        return __awaiter(this, void 0, void 0, function* () {
            return true;
        });
    }
    sendOK(ws) {
        return __awaiter(this, void 0, void 0, function* () {
        });
    }
    authenticationIsDone() {
        return __awaiter(this, void 0, void 0, function* () {
        });
    }
}
class WS2PNoRemoteAuth {
    getPubkey() {
        return "";
    }
    sendACK(ws) {
        return __awaiter(this, void 0, void 0, function* () {
        });
    }
    registerCONNECT(challenge, sig, pub) {
        return __awaiter(this, void 0, void 0, function* () {
            return true;
        });
    }
    registerOK(sig) {
        return __awaiter(this, void 0, void 0, function* () {
            return true;
        });
    }
    isAuthenticatedByRemote() {
        return true;
    }
    isAuthorizedPubkey(pub) {
        return __awaiter(this, void 0, void 0, function* () {
            return true;
        });
    }
    authenticationIsDone() {
        return __awaiter(this, void 0, void 0, function* () {
        });
    }
}
class WS2PMutedHandler {
    handlePushMessage(json) {
        return __awaiter(this, void 0, void 0, function* () {
        });
    }
    answerToRequest(json) {
        return __awaiter(this, void 0, void 0, function* () {
            throw "Does not answer";
        });
    }
}
//# sourceMappingURL=ws2p_connection.js.map